//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.agent.conversation.content;

//----------------------------------------------------------------------

// These are the import statement(s) required for this class. 
import java.util.*;

    /**
   * This class encapsulates content information in the form: <p>
   * <b>this</b> has an <b>attribute</b> with a value of <b>value</b><p>
   *
   * In general, the information contained within a Content object is
   * either: a) Other Content objects (forming a tree like structure);
   * b) Double objects; c) Long objects; or d) String objects (anything
   * which doesn't fit catergories a-c).
   *
   * <br>
   * <br>
   * @deprecated Content objects should be avoided - please use XML Databinding instead
   * @see fipaos.util.XMLDataBinding*/

public class Content
{
	//----------------------------------------------------------------------
  // CONSTANTS
  //----------------------------------------------------------------------

  //----------------------------------------------------------------------
  // STATIC VARIABLES
  //----------------------------------------------------------------------
		
  //----------------------------------------------------------------------
  // INSTANCE VARIABLES
  //----------------------------------------------------------------------

	// This is a reference to an instance of a Map. 
	private Map _map = null;
	
	// This is an instance variable that holds the name.
	private String _name = null;
	
	
	//----------------------------------------------------------------------
  // CONSTRUCTOR
  //----------------------------------------------------------------------

  /**
	 * Creates a completely blank Content object with no attributes and no name
	 */
	public Content() 
	{
		_map = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
	}
	
	/**
	 * Creates a completely blank Content object with the given name
   *
   * @param name The name of this particular Content object
	 */
	public Content( String name ) 
	{
		_name = name;
		_map = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
	}		


	/**
	 * Constructs a Content object using the given Map as a basis for the internal
	 * data structure.  DO NOT USE THIS CONSTRUCTOR - The internal representation of
	 * the Content object may change at a later date
	 *
	 * @param content The Map upon which the internal data structure should be based
	 */
	public Content( Map content )
	{
		_name = "Root";
		_map = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
		
		// Get all keys in the Map
		Set keys = content.keySet();
		
		// Iterate through the keys & create our internal map representation
		Iterator i = keys.iterator();
		while( i.hasNext () )
		{
			String next = (String) i.next();
			addAttribute( next, content.get( next ) );
		}
	}


	/**
	 * This Constructor will clone the structure of the Content object
	 * given in a manner which completely decouples this from the given
	 * Content object (i.e. it is recursively cloned - modifiying one object
	 * referenced directly or indirectly by this will have no effect on the
	 * objects directly or indirectly referenced by the given Content object).
	 *
	 * @param content The Content object to "clone" 
	 */
	public Content( Content c )
	{
		// Create a new Map for this
		_map = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
		
		// Duplicate the name
		_name = c.getName();
		
		// Get a list of the keys in the other Content object
		Iterator i = c.getAttributeNames().iterator();
		
		// Iterate through keys and clone the values it points to
		while( i.hasNext() )
		{
			String key = (String) i.next();

			// Get the Object that the key points to
			Object value = c.getAttribute( key );
			
			// Attempt to clone it
			Object clone = cloneObject( ( (Wrapper) value ).obj );

			// Add the clone to this
			_map.put( key, new Wrapper( clone ) );
		}
	}

	
//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------
		
	/**
	 * This method returns a Set of the attribute names contained in this Content object.
	 *
	 * @return A Set containing all attributes of this Content object
	 */ 
	public Set getAttributeNames()
	{
  	return _map.keySet();
	}
	

	/** 
	 * This method associates the specified value with the specified attribute.
	 * If the attribute already exists, a List will be created to hold both
	 * values, else if a List has already been created, the value will be appended
	 * to the  list.<p>
	 *
	 * addAttribute() does not overwrite attributes.  It is suggested that if this
	 * is the desired operation, setAttribute() is used instead.
	 *
	 * @param name The attribute name to add the value to
	 * @param value The value to associate with the attribute	 
	 */
	public synchronized void addAttribute( String name, Object value )
	{
		List list = null;
		
		// Check if a name has been given
		if( name == null )
		{
			return;
		}
		
		// Check if we're dealing with a List
		if( value instanceof List )
		{
			list = (List) value;
		}
		

		// Check if the key already exists
		Object current_object = getAttribute( name );

		if( current_object != null )
		{
			if( ( current_object instanceof java.util.List ) )
			{
				// Already a List there!  Just add new value
				List existing_list = (List) current_object;
				
				// Check if value to be added is a List
				if( list != null )
				{
					// Value is a List, so add its contents
					existing_list.addAll( list );
				}
				else
				{
					// Value isn't a List, so just add it
					existing_list.add( value );					
				}
			}
			else
			{
				// We're not using a List here yet!
				// Create a new List, add current value and new value to
				// it, and then put the Set into the Map
				List new_list = new LinkedList();
				new_list.add( current_object );

				// Check if value to be added is a List
				if( list != null )
				{
					// Value is a List, so add its contents
					new_list.addAll( list );
				}
				else
				{
					// Value isn't a List, so just add it
					new_list.add( value );					
				}
				
				// Put updated attribute info into Map
				_map.put( name, new Wrapper( new_list ) );
			}
		}
		else if( list != null )
		{
			// Just put List into _map
			_map.put( name, new Wrapper( list ) );
		}
		else
		{
			// Just put the value in the Map
			_map.put(name, new Wrapper( value ) );
		}
	}	
	

	/** 
	 * This method associates the specified values with the specified attribute.
	 * If the attribute already exists, a List will be created to hold the current
	 * values and the new values<p>
	 *
	 * addAttribute() does not overwrite attributes.  It is suggested that if this
	 * is the desired operation, setAttribute() is used instead.
	 *
	 * @param name The attribute name to add the value to
	 * @param values The values to associate with the attribute	 
	 */
	public synchronized void addAttribute( String name, List value )
	{
		addAttribute( name, (Object) value );
	}

	
	/** 
	 * This method associates the specified value with the specified attribute.
	 * If the attribute already exists it will be over-written<p>
	 *
	 * @param name The attribute name to add the value to
	 * @param value The value to associate with the attribute	 
	 */
	public synchronized void setAttribute( String name, Object value )
	{
		// Check if a name has been given
		if( name == null )
		{
			return;
		}
		
		// Remove attribute if it exists		
		removeAttribute( name );
		
		// Add attribute
		addAttribute( name, value );
	}


	/** 
	 * This method associates the specified values with the specified attribute.
	 * If the attribute already exists it will be over-written<p>
	 *
	 * @param name The attribute name to add the value to
	 * @param values The values to associate with the attribute	 
	 */
	public synchronized void setAttribute( String name, List values )
	{
		setAttribute( name, (Object) values );
	}
	
	
	/** 
	 * This method removes the mapping for this attribute from this if it is present. 
	 *
	 * @param name The name of the attribute to remove
	 */
	public synchronized void removeAttribute( String name )
	{
		// Check if a name has been given
		if( name == null )
		{
			return;
		}
		
		_map.remove( name );
	}	


	/** 
	 * This method returns the value to which the given attribute maps to.
	 * In the case that this is a 1..N mapping, the object will be a List instance
	 *
	 * @param name The name of the attribute for which to retreive the associated value
	 *
	 * @return Either the object associated with the attribute, or a List containing
	 *         multiple objects, or null if there is no such attribute.
	 */
	public Object getAttribute( String name ) 
	{
		// Check if a name has been given
		if( name == null )
		{
			return null;
		}
		
		Object o = _map.get( name );
		
		// If the object is not null, un-wrap the enclosed object
		if( ( o != null ) ) 
		{
			o = ( (Wrapper) o ).obj;
		}
		
		return o;
	}
	
	
	/**
	 * Method to get the name of this Content object
	 *
	 * @return The name associated with this
	 */
	public String getName()
	{
		return _name;
	}						
	
	
	/**
	 * This method indicates if this Content object has no attributes
	 * associated with it.
	 *
	 * @return True if this has no attributes
	 */
	public boolean isEmpty()
	{
		return _map.isEmpty();
	}
	

	/**
	 * This is a toString method which returns a String
	 * representation of this, each line pre-fixed with the given String.
	 *
	 * @param pre The String to prefix to each line of the created String
	 *
	 * @return A String representation of this
	 */
	public String toString( String pre )
	{
		String out = "\n" + pre + "ContentName=" + _name + "\n";
		
		Iterator i = getAttributeNames().iterator();
		
		// Iterate through keys and clone the values it points to
		while( i.hasNext() )
		{
			String key = (String) i.next();
			Object ovalue = ( (Wrapper) _map.get( key ) ).obj;
			String value = null;
			
			if( ovalue instanceof Content )
			{
				value = ( (Content) ovalue ).toString( pre + "\t" );
			}
			else if( ovalue instanceof List )
			{
				// Iterate through the set and invoke the toString() method
				value = "";

				List new_set = new LinkedList();
				Iterator j = ((List) ovalue).listIterator();		

				while( j.hasNext() )
				{
					Object ovalue2 = j.next();
					if( ovalue2 instanceof Content )
					{
						value += ( (Content) ovalue2 ).toString( pre + "\t" );
					}
					else
					{
						value += pre + key + "=" + ovalue2.toString() + "\n";
					}
				}
			}
			else
			{
				value = ovalue.toString();
			}
			
			out += pre + key + "=" + value + "\n";
		}
		
		return out;
	}	
	

	/**
	 * This is the public toString method which returns a String.
	 */
	public String toString()
	{
		return toString("");
	}
	

//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------

	/**
	 * Method to clone "completely" the given object.  The object should be one of the
	 * following types: Long, Double, String, List or Content
	 *
	 * @param value The object to clone.
	 *
	 * @return A clone of the given object, which is completely decoupled from it.
	 */
	private Object cloneObject( Object value )
	{
		Object clone = null;
		
		if( value instanceof Long )
		{
			clone = new Long( ( (Long) value ).longValue() );
		}
		else if( value instanceof Double )
		{
			clone = new Double( ( (Double) value ).doubleValue() );
		}
		else if( value instanceof String )
		{
			clone = new String( (String) value );
		}
		else if( value instanceof List )
		{
			// In the case of a List we must clone each individual
			// item within it as well.
			List new_list = new LinkedList();
			Iterator i = ((List) value).listIterator();		

			while( i.hasNext() )
			{
				new_list.add( cloneObject( i.next() ) );
			}

			clone = new_list;
		}
		else if( value instanceof Content )
		{
			clone = new Content( (Content) value );
		}

		return clone;
	}


//----------------------------------------------------------------------
// INNER CLASS
//----------------------------------------------------------------------

	/**
	 * This inner-class provides a simple wrapper so that the TreeMap
	 * implementation thinks that everything within it it of the same type
	 */
	private class Wrapper
	{
		public Object obj;
		
		public Wrapper( Object o )
		{
			obj = o;
		}	

		public boolean equals( Object o )
		{
			if( obj.getClass().equals( o.getClass() ) )
			{
				return obj.equals( o );
			}
			
			return false;	
		}
	}
}