//******************************************************************************
//	Open Source Copyright Notice and License: FIPA-OS
//
//	1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//	2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.ont.fipa;

import fipaos.ont.fipa.fipaman.FIPAMANCONSTANTS;
import fipaos.ont.fipa.fipaman.AgentID;
import fipaos.ont.BASETOKENS;
import fipaos.util.*;
import fipaos.parser.ParserException;
import fipaos.parser.acl.string.ACLParser;
import fipaos.parser.acl.string.ParseException;

import java.util.StringTokenizer;
import java.util.Iterator;
import java.io.Serializable;
import java.util.List;
import java.util.LinkedList;

/**
 * This class provides the representation of an ACL communication within an Agent.
 * <p> Implements XC00061D & XC00070F (except for support for ":encoding" field)
 */ 
public class ACL implements Serializable, Cloneable
{
	/**
	 * Misc. serial ID to allow different versions of this class to be serialised/deserialised
	 * without causing IncompatibleClassChange exceptions
	 */
	private static final long serialVersionUID = 1234567890123456L;
	
	/**
	 * Content encoding constants - see FIPA ACL spec.
	 */
	public final static int STRINGLITERAL_ENCODING = 1;
	public final static int BYTELENGTH_ENCODING = 2;
	
	/**
	 * ACL Encoding name as defined in FIPA00070
	 */
	public final static String FIPA00070 = "fipa.acl.rep.string.std";
	
	/**
	 * ACL Encoding name as defined in FIPA00069
	 */
	public final static String FIPA00069 = "fipa.acl.rep.bitefficient.std";
	
	/**
	 * ACL Encoding name as defined in FIPA00071
	 */
	public final static String FIPA00071 = "fipa.acl.rep.xml.std";	
	
	/**
	 * ACL Encoding using Java Serialisation
	 */
	public final static String FIPA_OS_SERIALIZATION = "fipaos.acl.rep.java.serialization";
	
	/**
	 * Encoding type for content of this ACL message
	 */
	private String _content_encoding = null;
	
	/**
	 * AID of the Agent sending the ACL communication
	 */
	private AgentID _sender;
	
	/**
	 * AID('s) of the recipients of the message
	 */
	private List _receivers;
	
	/**
	 * Reply-to field of the ACL communication
	 */
	private List _reply_to;
	
	/**
	 * Content of this communication - can be any type of object.
	 */
	private Object _content;
	
	/**
	 * Language of the content (only used if this is deparsed into a String form, or
	 * is already in a String form)
	 */
	private String _language;
	
	/**
	 * Ontology of the content of this ACL communication
	 */
	private String _ontology;
	
	/**
	 * Protocol used within the current conversation
	 */
	private String _protocol;
	
	/**
	 * Conversation-id of the conversation that this communication belongs to
	 */
	private String _conversation_id;
	
	/**
	 * In-reply-to field of the ACL communication
	 */
	private String _in_reply_to;
	
	/**
	 * Reply-with field of the ACL communication
	 */
	private String _reply_with;

	/**
	 * Reply-to field of the ACL communication
	 */
	private UTCContainer _reply_by;
	
	/**
	 * Performative associated with this communication
	 */
	private String _performative;
	
	/**
	 * Default encoding mechanism for content, if Stringified
	 */
	private int _encoding = BYTELENGTH_ENCODING;
	
	/**
	 * Indicates if encoding has taken place
	 */
	private boolean _encoded = false;
	
	/**
	 * Indicates if brackets have been stripped from SL content
	 */
	private boolean _brackets_stripped = true;
	
	/**
	 * <p>Duplicates the content of the given ACL object
	 *
	 * @deprecated Please use clone() instead
	 */
	public ACL( ACL acl )
	{
		// Default constructor first
		this();
		
		// Copy the fields from the given ACL object into this
		_conversation_id = acl._conversation_id;
		_protocol = acl._protocol;
		_ontology = acl._ontology;
		_language = acl._language;
		_content = acl._content;
		_in_reply_to = acl._in_reply_to;
		_performative = acl._performative;
		_reply_with = acl._reply_with;
		_encoding = acl._encoding;
		_encoded = acl._encoded;
		_brackets_stripped = acl._brackets_stripped;
		
		if( acl._sender != null )
		{
			_sender = (AgentID) acl._sender.clone();
		}
		
		if( acl._reply_by != null )
		{
			_reply_by = (UTCContainer) acl._reply_by.clone();
		}

		if( acl._reply_to != null )
		{
			_reply_to = new LinkedList();
			Iterator i = acl._reply_to.iterator();
			while( i.hasNext() )
			{
				_reply_to.add( ( (AgentID) i.next() ).clone() );
			}
		}
		
		if( acl._receivers != null )
		{
			_receivers = new LinkedList();
			Iterator i = acl._receivers.iterator();
			while( i.hasNext() )
			{
				_receivers.add( ( (AgentID) i.next() ).clone() );
			}
		}
	}
	
	
	public Object clone()
	{
		try
		{
			ACL acl = (ACL) super.clone();
			
			if( _sender != null )
			{
				acl._sender = (AgentID) _sender.clone();
			}

			if( _reply_by != null )
			{
				acl._reply_by = (UTCContainer) _reply_by.clone();
		}
		
			if( _reply_to != null )
			{
				acl._reply_to = new LinkedList();
				Iterator i = _reply_to.iterator();
				while( i.hasNext() )
				{
					acl._reply_to.add( ( (AgentID) i.next() ).clone() );
				}
			}

			if( _receivers != null )
			{
				acl._receivers = new LinkedList();
				Iterator i = _receivers.iterator();
				while( i.hasNext() )
				{
					acl._receivers.add( ( (AgentID) i.next() ).clone() );
				}
			}
			
			// See if content is clonable
			if( _content instanceof Cloneable )
			{
				// TODO: This can't be done!
//				acl._content = _content.clone();
			}
			else if( _content instanceof String )
			{
				// Do nothing!
			}
			else if( _content instanceof Serializable )
			{
				// Use serialisation (v.slow!!!!!)
				acl._content = Serialize.duplicate( _content );
			}
			
			return acl;
		}
		catch( CloneNotSupportedException e )
		{
			return null;
		}
	}
	
	
	/**
	 * Creates an ACL object from the given ACL string
	 */
	public ACL( String msg ) throws ParserException
	{
		// Invoke default constructor
		this();
		
		ACL acl = null;
		try
		{
			acl = ACLParser.parse( msg );
		}
		catch( ParseException e )
		{
			throw new ParserException( e.toString() );
		}
		
		// Duplicate the content of the ACL object
		_sender = acl._sender;
		_reply_by = acl._reply_by;
		_content = acl._content;
		_conversation_id = acl._conversation_id;
		_in_reply_to = acl._in_reply_to;
		_language = acl._language;
		_ontology = acl._ontology;
		_performative = acl._performative;
		_protocol = acl._protocol;
		_reply_with = acl._reply_with;
		_receivers = acl._receivers;
		_reply_to = acl._reply_to;
		_encoded = acl._encoded;
		_encoding = acl._encoding;
		_content_encoding = acl._content_encoding;
		_brackets_stripped = acl._brackets_stripped;
	}
	
	
	/**
	 * Creates an empty ACL communication
	 */
	public ACL()
	{
		// Create the list of receivers
		_receivers = new LinkedList();
		_reply_to = new LinkedList();
	}
	
	
	public void setSenderAID( AgentID aid )
	{
		_sender = aid;
	}
	
	
	public AgentID getSenderAID()
	{
		return _sender;
	}
	
	
	public void addReceiverAID( AgentID aid )
	{
		_receivers.add( aid );
	}
	
	
	public void setReceiverAID( AgentID aid )
	{
		_receivers.clear();

		if( aid != null )
		{
			_receivers.add( aid );
		}
	}
	
	
	public void setReceiverAIDs( List aids )
	{
		_receivers.clear();
		
		if( aids != null )
		{
			_receivers.addAll( aids );
		}
	}
	
	
	/**
	 * Get just the first AID of the receivers field, if present
	 */
	public AgentID getReceiverAID()
	{
		if( _receivers != null && _receivers.size() > 0 )
		{
			return (AgentID) _receivers.get( 0 );
		}
		else
		{
			return null;
		}
	}

	
	public List getReceiverAIDs()
	{
		return new LinkedList( _receivers );
	}
	

	public void addReplyToAID( AgentID aid )
	{
		_reply_to.add( aid );
	}
	
	
	public void setReplyToAIDs( List aids )
	{
		_reply_to.clear();

		if( aids != null )
		{
			_reply_to.addAll( aids );
		}
	}
	
	
	public List getReplyToAIDs()
	{
		return new LinkedList( _reply_to );
	}
	
	
	/**
	 * Method to be invoked by the ACLParser to set the content -
	 * causes outer brakets to be stripped from SL
	 */
	public void setContentObjectACLParser( String s )
	{
		// Indicate content hasn't had brakets stripped
		_brackets_stripped = false;
		
		_content = s;
		
		// Check for encoding
		if( s.startsWith( "\"" ) && s.endsWith( "\"" ) )
		{
			_encoded = true;
			_encoding = STRINGLITERAL_ENCODING;
		}
		else if( s.startsWith( "#" ) )
		{
			_encoded = true;
			_encoding = BYTELENGTH_ENCODING;				
		}
		else
		{
			// Not encoded already!
			_encoded = false;
			_encoding = STRINGLITERAL_ENCODING;
		}
	}
	
	
	public void setContentObject( Object content )
	{
		_content = content;
		
		// Check content type, and if it is already encoded
		if( _content instanceof String )
		{
			// Check for encoding
			String s = (String) content;
			if( s.startsWith( "\"" ) && s.endsWith( "\"" ) )
			{
				_encoded = true;
				_encoding = STRINGLITERAL_ENCODING;
			}
			else if( s.startsWith( "#" ) )
			{
				_encoded = true;
				_encoding = BYTELENGTH_ENCODING;				
			}
			else
			{
				// Not encoded already!
				_encoded = false;
				_encoding = STRINGLITERAL_ENCODING;
			}
		}
		else
		{
			// Not a String - assume not encoded
			_encoded = false;
			_encoding = BYTELENGTH_ENCODING;
		}
		
		_brackets_stripped = true;
	}
	
	
	public Object getContentObject()
	{
		String content = null;
		
		// Check if content is a String
		if( _content instanceof String )
		{
			// Typecast to correct type
			content = (String) _content;
			content = content.trim();
			
			// Is message encoded yet?
			if( _encoded )
			{
				// Decide what to do in various encoding cases
				// NOTE: Only the first case should only every occur under current
				// circumstances!
				switch( _encoding )
				{
					case BYTELENGTH_ENCODING:
						content = byteLengthDecode( content );
						break;
						
					case STRINGLITERAL_ENCODING:
						content = stringLiteralDecode( content );
						break;
				}
			}
			
			// Strip brackets if we're dealing with SL
			if(	_language != null && _language.toLowerCase().startsWith( "fipa-sl" ) )
			{
				if( !_brackets_stripped )
				{
					content = StringParser.trimBrackets( content );
				}
			}

			// Do nothing if String is in original form
			return content;
		}
		else
		{
			return _content;
		}
	}
	
	
	/**
	 * Sets the content, optionally encoding the content
	 * as the specified type (see FIPA specification), if it is Stringified
	 *
	 * @param content The content of the message
	 * @param encode Indicates the type of encoding to use for the string
	 */
	public void setContentObject( Object content, int encode )
	{
		setContentObject( content );
		_encoding = encode;
		_encoded = false;
		_brackets_stripped = true;
	}
	
	
	public void setContentEncoding( String encoding )
	{
		_content_encoding = encoding;
	}
	
	public String getContentEncoding()
	{
		return _content_encoding;
	}
	
	public void setConversationID( String conv_id )
	{
		_conversation_id = conv_id;
	}
	
	
	public String getConversationID()
	{
		return _conversation_id;
	}
	
	
	public void setInReplyTo( String in_reply_to )
	{
		_in_reply_to = in_reply_to;
	}
	
	
	public String getInReplyTo()
	{
		return _in_reply_to;
	}
	
	
	public void setLanguage( String lang )
	{
		_language = lang;
	}
	
	
	public String getLanguage()
	{
		return _language;
	}
	
	
	public void setOntology( String ont )
	{
		_ontology = ont;
	}
	
	
	public String getOntology()
	{
		return _ontology;
	}
	
	
	public void setPerformative( String perf )
	{
		_performative = perf;
	}
	
	
	public String getPerformative()
	{
		return _performative;
	}
	
	
	public void setProtocol( String protocol )
	{
		_protocol = protocol;
	}
	
	
	public String getProtocol()
	{
		return _protocol;
	}
	
	
	public void setReplyByUTC( UTCContainer utc )
	{
		_reply_by = utc;
	}
	
	
	public UTCContainer getReplyByUTC()
	{
		return _reply_by;
	}
	
	
	public void setReplyWith( String reply_with )
	{
		_reply_with = reply_with;
	}
	
	
	public String getReplyWith()
	{
		return _reply_with;
	}
	
	
	/**
	 *	toString returns a string representation of the message in default ACL
	 * encoding
	 *  @return string
	 */
  public String toString()
	{
		StringBuffer tosend = new StringBuffer();
		tosend.append( BASETOKENS.OPENBRACKET );
		
		// Add performative
		tosend.append( _performative );
		tosend.append( BASETOKENS.SPACE );

		// Get the sender (if any) and add
		if ( _sender != null )
		{
			addWithSpaces( tosend, FIPACONSTANTS.SENDER, _sender.toString() );
		}

		// Create the receivers field
		tosend.append( FIPACONSTANTS.RECEIVER );
		tosend.append( BASETOKENS.SPACE );

		// Add all recipients
		tosend.append( BASETOKENS.OPENBRACKETSPACE );
		tosend.append( FIPAMANCONSTANTS.SET );

		// Iterate through recipients
		Iterator i = _receivers.iterator();
		while( i.hasNext() )
		{
			String recipient = i.next().toString();
			tosend.append( BASETOKENS.SPACE );
			tosend.append( recipient );
			tosend.append( BASETOKENS.SPACE );
		}

		tosend.append( BASETOKENS.CLOSEBRACKET );
		tosend.append( BASETOKENS.SPACE );
		
		// Create the reply-to field
		if( _reply_to != null && _reply_to.size() > 0 )
		{
			tosend.append( FIPACONSTANTS.REPLY_TO );
			tosend.append( BASETOKENS.SPACE );

			// Add all recipients
			tosend.append( BASETOKENS.OPENBRACKETSPACE );
			tosend.append( FIPAMANCONSTANTS.SET );

			// Iterate through recipients
			i = _reply_to.iterator();
			while( i.hasNext() )
			{
				String recipient = i.next().toString();
				tosend.append( BASETOKENS.SPACE );
				tosend.append( recipient );
				tosend.append( BASETOKENS.SPACE );
			}

			tosend.append( BASETOKENS.CLOSEBRACKET );
			tosend.append( BASETOKENS.SPACE );
		}

		// Add the content
		if ( _content != null )
		{
			// Check if content needs encoding
			String content = _content.toString();
			
			// Check if we're using SL, and need to add brakets
			if(	_language != null && 
					_language.toLowerCase().startsWith( "fipa-sl" ) &&
					_brackets_stripped )
			{
				content = "(" + content + ")";
			}
			
			// If encoding is specified, do it
			if( !_encoded )
			{
				switch( _encoding )
				{
					case STRINGLITERAL_ENCODING:
						content = stringLiteralEncode( content );
						break;
						
					case BYTELENGTH_ENCODING:
						content = byteLengthEncode( content );
						break;
						
					default:
						break;
				}
			}
			else
			{
				// By default just send "as is"
			}
			
			addWithSpaces( tosend, FIPACONSTANTS.CONTENT, content );
		}

		// Add reply-with
		if ( _reply_with != null )
		{
			addWithSpaces( tosend, FIPACONSTANTS.REPLY_WITH, _reply_with );
		}

		// Add reply-by
		if ( _reply_by != null )
		{
			addWithSpaces( tosend, FIPACONSTANTS.REPLY_BY, _reply_by.toString() );
		}

		// Add in-reply-to
		if ( _in_reply_to != null )
		{
			addWithSpaces( tosend, FIPACONSTANTS.IN_REPLY_TO, _in_reply_to );
		}

		// Add language
		if ( _language != null )
		{
			addWithSpaces( tosend, FIPACONSTANTS.LANGUAGE, _language );
		}

		// Add ontology
		if ( _ontology != null )
		{
			addWithSpaces( tosend, FIPACONSTANTS.ONTOLOGY, _ontology );
		}
		
		// Add protocol
		if ( _protocol != null )
		{
			addWithSpaces( tosend, FIPACONSTANTS.PROTOCOL, _protocol );
		}
		
		// Add conversation ID
		if ( _conversation_id != null )
		{
			addWithSpaces( tosend, FIPACONSTANTS.CONVERSATION_ID, _conversation_id );
		}
		
		tosend.append( BASETOKENS.CLOSEBRACKET );
		
		// Return message
		return tosend.toString();
  }    	  	
  
	/**
	 * Encodes the given String into a StringLiteral format (see FIPA ACL definiton)
	 *
	 * @param in The String to encode
	 * @return The encoded String
	 */
	public static String stringLiteralEncode( String in )
	{
		// StringBuffer to contain String
		StringBuffer buff = new StringBuffer();
		
		// Add first quote
		buff.append( BASETOKENS.QUOTES_CHAR );

		// Iterate through string, removing all " characters
		StringTokenizer st = new StringTokenizer( in, "\\\"", true );
		while( st.hasMoreTokens() )
		{
			// Get next token
			String token = st.nextToken();
			
			// Check if its a "
			if( token.equals( BASETOKENS.QUOTES ) )
			{
				// Append token in encoded form
				buff.append( BASETOKENS.BACKSLASH_CHAR );
				buff.append( BASETOKENS.QUOTES_CHAR );
			}
			else if( token.equals( BASETOKENS.BACKSLASH ) )
			{
				// Append token in encoded form
				buff.append( BASETOKENS.BACKSLASH_CHAR );
				buff.append( BASETOKENS.BACKSLASH_CHAR );				
			}
			else
			{
				// Append token as normal
				buff.append( token );
			}
		}
	
		// Add terminating quote
		buff.append( BASETOKENS.QUOTES_CHAR );
		
		return buff.toString();
	}
  
	
	/**
	 * Decodes the given String from a StringLiteral format (see FIPA ACL definiton)
	 *
	 * @param in The String to decode
	 * @return The decoded String
	 */
	public static String stringLiteralDecode( String in )
	{
		// StringBuffer to contain String
		StringBuffer buff = new StringBuffer();
		
		// Skip first and last quotes
		int index1 = in.indexOf( BASETOKENS.QUOTES_CHAR );
		int index2 = in.lastIndexOf( BASETOKENS.QUOTES_CHAR );
		
		// Check if indexs are equals (implies either no " chars, or only one!)
		if( index1 == index2 )
		{
			return in;
		}
		else
		{
			in = in.substring( index1 + 1, index2 );
		}

		// Iterate through string, removing all " characters
		StringTokenizer st = new StringTokenizer( in, "\\\"", true );
		while( st.hasMoreTokens() )
		{
			// Get next token
			String token = st.nextToken();
			
			// Check if its a "
			if( token.equals( BASETOKENS.BACKSLASH ) )
			{
				// See if next char is a quote char
				String next_token = st.nextToken();
				
				if( next_token.equals( BASETOKENS.QUOTES ) )
				{
					// Append token in decoded form
					buff.append( BASETOKENS.QUOTES_CHAR );
				}
				else if( next_token.equals( BASETOKENS.BACKSLASH ) )
				{
					// Append token in decoded form
					buff.append( BASETOKENS.BACKSLASH_CHAR );					
				}
				else
				{
					// Append tokens as originally intended
					buff.append( token );
					buff.append( next_token );
				}
			}
			else
			{
				// Append token as normal
				buff.append( token );
			}
		}
	
		return buff.toString();
	}
  
	
	/**
	 * Encodes the given String into a ByteLengthEncodedString format (see FIPA ACL definiton)
	 *
	 * @param in The String to encode
	 * @return The encoded String
	 */
	public static String byteLengthEncode( String in )
	{
		// Add #xxx" to front of string
		StringBuffer buffer = new StringBuffer();
		
		buffer.append( BASETOKENS.HASH );
		buffer.append( in.length() );
		buffer.append( BASETOKENS.QUOTES_CHAR );
		buffer.append( in );
		
		return buffer.toString();
	}

	
	/**
	 * Decodes the given String from a ByteLengthEncodedString format (see FIPA ACL definiton)
	 *
	 * @param in The String to decode
	 * @return The decoded String
	 */
	public static String byteLengthDecode( String in )
	{
		// Remove #xxx" from from front of String
		int length = -1;
		int hash_index = in.indexOf( BASETOKENS.HASH );
		int quote_index = in.indexOf( BASETOKENS.QUOTES_CHAR );
		
		try
		{
			length = Integer.parseInt( in.substring( hash_index + 1, quote_index ) ); 
		}
		catch( Throwable t )
		{
		}
		
		String out = in.substring( quote_index + 1 );
		
		// Check length
		if( out.length() > length )
		{
			out = out.substring( 0, length );
		}
		
		return out;
	}
	
//----------------------------------------------------------------------
// SERIALIZATION METHODS
//----------------------------------------------------------------------
  
	/**
	 * Read object - just reads fields from OIS and places into relevant fields -
	 * ignores news fields (hence removing forward-compatibility issues)
	 */
	private void readObject( java.io.ObjectInputStream in ) throws java.io.IOException, ClassNotFoundException
	{
		// Read all fields by default
		in.defaultReadObject();
	}

	/**
	 * Write object - just writes the fields of this object into the OOS
	 * (hence removing forward-compatibility issues)
	 */
	private void writeObject( java.io.ObjectOutputStream out ) throws java.io.IOException
	{
		// Write all fields by default
		out.defaultWriteObject();
	}

//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------
  
  /**
   * Add the keyword and value pair with spaces to the stringbuffer.
   */
  private void addWithSpaces(StringBuffer toAddto, String keyword, String value )
	{
		toAddto.append(keyword);
		toAddto.append(BASETOKENS.SPACE);
		toAddto.append(value);
		toAddto.append(BASETOKENS.SPACE);
	}                                   	


  private String removeParentheses( String str )
  {
	str = str.trim();
	int start = 0;
	int end = 0;
	if( str.startsWith( BASETOKENS.OPENBRACKET ) )
	{
	  start = ( str.indexOf( BASETOKENS.OPENBRACKET ) + 1) ;
	  str = str.substring( start, str.length() );
	}
	if( str.endsWith( BASETOKENS.CLOSEBRACKET ) )
	{
	  end = str.lastIndexOf( BASETOKENS.CLOSEBRACKET );
	  str = str.substring( 0, str.length() - 1);
	}
	return str.trim();
  }  
}