//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.ont.fipa.fipaman;

//----------------------------------------------------------------------

import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;
import java.util.List;
import java.util.Collections;
import java.io.Serializable;

import fipaos.ont.BASETOKENS;
import fipaos.agent.conversation.content.Content;
import fipaos.parser.ParserException;
import fipaos.parser.sl.SLParser;

//----------------------------------------------------------------------

/** The Agent Platform Transport Description forms part of an AP description (see [PC00023]) and is expressed in SL0
 * TODO: there is at present no checking to see if the characters in the fields are valid as per the BNF definitions
 * <br>
 * <br>
 */
public class APTDescription implements Serializable
{

	//----------------------------------------------------------------------
	// CONSTANTS
	//----------------------------------------------------------------------

	//----------------------------------------------------------------------
	// STATIC VARIABLES
	//----------------------------------------------------------------------

	//----------------------------------------------------------------------
	// INSTANCE VARIABLES
	//----------------------------------------------------------------------

	/** A set of the message transport protocols
	 */
	protected Set _mtpd;

	//----------------------------------------------------------------------
	// CONSTRUCTOR(S)
	//----------------------------------------------------------------------

	/** This is the empty constructor.
	 */

	public APTDescription() {
		_mtpd = Collections.synchronizedSet( new TreeSet( fipaos.util.Comparators.OBJECT_COMPARE ) );
	}

	/** Constructs a APTDescription object from a string
	 * @param description string representation of a APTDescription object
	 * @throws ParserException if the String is invalid
	 */

	public APTDescription(String description) throws ParserException {

		try {
			_mtpd = Collections.synchronizedSet( new TreeSet( fipaos.util.Comparators.OBJECT_COMPARE ) );
			parse(SLParser.parseTerm(description));
		}
		catch( fipaos.parser.sl.ParseException e ) {
			throw new ParserException(e.getMessage());
		}
	}

	//----------------------------------------------------------------------
	// PRIVATE METHODS
	//----------------------------------------------------------------------

	private void addKeywordValuePair(StringBuffer buffer, String keyword, String value)
	{
		buffer.append(keyword);
		buffer.append(BASETOKENS.SPACE);
		buffer.append(value);
		buffer.append(BASETOKENS.NEWLINE);
	}

	private String convertMTPDSetIntoString(Set set)
	{
		StringBuffer buffer=new StringBuffer();

		buffer.append(BASETOKENS.OPENBRACKET);
		buffer.append(FIPAMANCONSTANTS.SET);
		buffer.append(BASETOKENS.SPACE);

		MTPDescription mtpd;

		Iterator iterator=set.iterator();

		while (iterator.hasNext())
		{
			mtpd=(MTPDescription) iterator.next();
			buffer.append(mtpd.toString());
			buffer.append(BASETOKENS.SPACE);
		}

		buffer.append(BASETOKENS.CLOSEBRACKET);

		return buffer.toString();
	}

	/**
	 * Creates a Java set from a SL Set content
	 */

	private Set createSet(Content content) throws ParserException {

		Set set= Collections.synchronizedSet( new TreeSet( fipaos.util.Comparators.OBJECT_COMPARE ) );

		Content con=(Content) content.getAttribute("parameterValue");

		con=(Content) con.getAttribute("term");

		con=(Content) con.getAttribute("set");

		if (con==null) {
			throw new IllegalArgumentException("Illegal APT-DESCRIPTION FORMAT: SET");
		}

		if (con.getAttribute("term") instanceof List) {

			List list=(List) con.getAttribute("term");

			int size=list.size();

			String[] array=new String[size];

			for (int i=0;i<size;i++) {

				Content term=(Content) list.get(i);

				set.add(new MTPDescription((String) term.getAttribute("formattedString")));

			}
		}
		else {

			set.add(new MTPDescription((String) ((Content) con.getAttribute("term")).getAttribute("formattedString")));

		}

		return set;
	}

	/**
	 * parse an ACL FIPA-APT-Description into its component objects
	 * e.g.
	 * <b>
	 * (ap-transport-description :available-mtps (set (mtp-description
	 *			     				:profile fipa-alpha
	 *                           			    	:mtp-name fipa-iiop-std
	 *                       			    	:addresses (set iiop://monitor.co.uk:9000/acc http://www.agentworld.co.uk:9000)
	 *					      )))
	 * <br>
	 *
	 * apt_content is traversed to construct a APTDescription object
	 **/

	private void parse(Content apt_content) throws IllegalArgumentException, ParserException {

		if (apt_content!=null) {

			Content functional_content=(Content) apt_content.getAttribute("functionalTerm");

			if (functional_content==null || !functional_content.getAttribute("functionalOperator").equals(FIPAMANCONSTANTS.AP_TRANSPORT_DESCRIPTION)) {
				throw new IllegalArgumentException("Content is not a APT-DESCRIPTION");
			}

			Content parameters=(Content) functional_content.getAttribute("parameter");

			if (parameters!=null) {

				String name=(String) parameters.getAttribute("parameterName");

				if (name.equals(FIPAMANCONSTANTS.AVAILABLE_MTPS)) {

					_mtpd=createSet(parameters);
				}
			}
		}
	}

	//----------------------------------------------------------------------
	// PROTECTED METHODS
	//----------------------------------------------------------------------

	//----------------------------------------------------------------------
	// PUBLIC METHODS
	//----------------------------------------------------------------------

	/** This method sets the message transport protocol set
	 * @param mtpd MTPD to be added to the set
	 */
	public void setMTPSSet(Set mtpd) {
		_mtpd = mtpd;
	}

	/** This method sets the message transport protocol set.
	 * @param mtpd array of MTPDs to be added to the set
	 */
	public void setMTPSSet(MTPDescription[] mtpd) {

		Set set=new TreeSet( fipaos.util.Comparators.OBJECT_COMPARE );

		for (int i=0; i<mtpd.length; i++) {
			set.add(mtpd[i]);
		}

		_mtpd=set;
	}

	/** This method returns the set of message transport protocols
	 * @return a set of message transport protocols
	 */
	public Set getMTPSSet() {
		return _mtpd;
	}

	/** Add a message transport protocol to the list
	 * @return boolean - true if operation successful, false if not
	 * @param mtpd the mtpd to be added to the set
	 */
	public boolean addMTPD(MTPDescription mtpd) {

		synchronized( _mtpd ) {
			if( !_mtpd.contains( mtpd ) ) {
				return _mtpd.add( mtpd );
			}
		}
		return false;
	}

	/** Remove a message transport protocol from the set
	 * @return boolean - true if operation successful, false if not.
	 * @param mtpd remove a MTPD from the set
	 */
	public boolean removeMTPD(MTPDescription mtpd) {

		synchronized( _mtpd ) {
			return _mtpd.remove( mtpd );
		}
	}

	/** Does the set contain a particular message transport protocol. This can be either the profile name or the message transport protocol name
	 * @param name the name of message transport protocol or the mtp profile
	 * @return boolean - true if the protocol is in the set, false if not
	 */

	public boolean contains(String name) {

		MTPDescription mtpd;

		Iterator iterator=_mtpd.iterator();

		while (iterator.hasNext()) {

			mtpd = ( MTPDescription ) iterator.next();

			if( name.equals( mtpd.getProfile() ) ||  name.equals( mtpd.getMTPName() )) {
				return true;
			}
		}

		return false;
	}

	/** A String Representation of this ATPDescription object
	 * @return A string representation of this ATPDescription object
	 */

	public String toString() {

		StringBuffer buffer=new StringBuffer();
		buffer.append(BASETOKENS.OPENBRACKET);
		buffer.append(FIPAMANCONSTANTS.AP_TRANSPORT_DESCRIPTION);
		buffer.append(BASETOKENS.NEWLINE);

		if (_mtpd !=null ) {
			addKeywordValuePair(buffer,FIPAMANCONSTANTS.AVAILABLE_MTPS,convertMTPDSetIntoString(_mtpd));
		}

		buffer.append(BASETOKENS.CLOSEBRACKET);

		return buffer.toString();

	}

	//----------------------------------------------------------------------
	// MAIN/TESTING METHOD
	//----------------------------------------------------------------------

	/** The main method is for testing purposes
	 * @param args Arguments
	 */
	public static void main(String[] args) {

		String apt_string=" (ap-transport-description " +
		" :available-mtps " + "(set " +
		"(mtp-description " +
		":profile fipa-alpha " +
		":mtp-name fipa-iiop-std " +
		":addresses "  + "(set " +
		"iiop://monitor.co.uk:9000/acc " +
		"http://www.agentworld.co.uk:9000 " +
		") " +
		") " +
		") " +
		") ";

		try {
			APTDescription apt_description=new APTDescription(apt_string);
			System.out.println(apt_description.getMTPSSet());
		}
		catch (ParserException e) {
			e.printStackTrace();
		}
	}
} // End class