//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.ont.fipa.fipaman;

import fipaos.parser.acl.ACLMessage;
import fipaos.parser.acl.string.ACLParser;
import fipaos.ont.BASETOKENS;
import fipaos.ont.fipa.FIPACONSTANTS;
import fipaos.parser.aid.AIDParser;
import fipaos.parser.ParserException;
import fipaos.util.URL;
import fipaos.util.DynamicClass;
import fipaos.util.DIAGNOSTICS;

import java.io.*;
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;
import java.util.StringTokenizer;


/** This class represents an AID, as described in the FIPA99 specification
 */
public class AgentID implements Serializable, Cloneable
{
  private String _name;
  private List _addresses;
  private List _resolvers;
	
	/** Constructs a blank AgentID.
	 */
  public AgentID() 
  {
	_name = null;
	_addresses = new LinkedList();
	_resolvers = new LinkedList();
  }  
	
	/** Constructs an AgentID from the given String, which is expected to conform to
	 * the FIPA99 representation for an AID
	 * @param aid The stringified AID to parse
	 * @throws ParserException Thrown if the AID specified is malformed
	 */
  public AgentID( String aid ) throws ParserException
  {
		// Invoke default constructor
		this();
		
		AgentID id = null;
		
		// Create a AgentID from the String
		try
		{
			id = AIDParser.parse( aid );
		}
		catch( fipaos.parser.aid.ParseException e )
		{
			throw new ParserException( e.toString() );
		}
		
		// Copy created AID info into this
		_name = id._name;
		_addresses = id._addresses;
		_resolvers = id._resolvers;
  }  
  
	
	/** Constructs a new AgentID from the given AgentID.
	 * @param aid The AID to copy
	 * @deprecated Please use clone() instead
	 */
	public AgentID( AgentID aid )
	{
		this();
		
		// Copy AID info into this
		_name = aid._name;
		
		// Duplicate addresses (URL's)
		Iterator i = aid._addresses.iterator();
		while( i.hasNext() )
		{
			_addresses.add( ( (URL) i.next() ).clone() );
		}

		// Duplicate resolvers (AgentID's)
		i = aid._resolvers.iterator();
		while( i.hasNext() )
		{
			_resolvers.add( new AgentID( (AgentID) i.next() ) );
		}
	}
	

	public Object clone()
	{
		try
		{
			AgentID aid = (AgentID) super.clone();
			aid._addresses = new LinkedList();
			aid._resolvers = new LinkedList();
			
			// Duplicate addresses (URL's)
			Iterator i = _addresses.iterator();
			while( i.hasNext() )
			{
				aid._addresses.add( ( (URL) i.next() ).clone() );
			}

			// Duplicate resolvers (AgentID's)
			i = _resolvers.iterator();
			while( i.hasNext() )
			{
				aid._resolvers.add( ( (AgentID) i.next() ).clone() );
			}
			
			return aid;
		}
		catch( CloneNotSupportedException e )
		{
			return null;
		}
	}
	
	
	/** Sets the name of the Agent represented by this AgentID
	 * @param name The name to set
	 */
	public void setName( String name )
  {
	_name = name;
  }  
  
	/** Sets the addresses associated with this AgentID
	 * @param addresses String containing the addresses of the Agent in stringified form
	 */
  public void setAddresses( String addresses )
  {
	addresses = removeParentheses( addresses );
		addresses = addresses.substring( addresses.indexOf( FIPAMANCONSTANTS.SEQUENCE ) + 
																	FIPAMANCONSTANTS.SEQUENCE.length() );
	addresses = addresses.trim();
	StringTokenizer strtok = new StringTokenizer( addresses, BASETOKENS.SPACE );
		while( strtok.hasMoreTokens() )
		{
	  synchronized( _addresses )
	  {
		_addresses.add( new URL( strtok.nextToken() ) );
	  }
	}
  }  
	
	/** Sets the addresses (URL's) through which the Agent can be contacted
	 * @param addresses List of fipaos.util.URL objects
	 */
	public void setAddresses( List addresses )
	{
		synchronized( _addresses )
		{
			_addresses.clear();
			_addresses.addAll( addresses );
		}
	}

	
	/** Sets the list of resolvers for this Agent
	 * @param resolvers Stringified version of the list of resolvers
	 */
	public void setResolvers( String resolvers )
  {
		// Set the resolvers
		setResolvers( getAgentIDs( resolvers ) );
  }  
	
	/** Sets tbe resolvers of this Agent
	 * @param resolvers A List of AgentID's, which can resolve the Agent
	 */
	public void setResolvers( List resolvers )
  {
		synchronized( _resolvers )
		{
			_resolvers.clear();
			_resolvers.addAll( resolvers );
		}
	}
		
	/** Gets the name associated with this AgentID
	 * @return A String containing the Agent name
	 */
	public String getName()
  {
	return _name;
  }  
	
	/** Gets a List of addresses (URL's) associated with this AgentID
	 * @return A List of addresses (URL's) in stringified form
	 */
  public List getAddresses()
  {
	return _addresses;
  }  
  
	/** Gets a List of resolvers for the Agent represented by this
	 * @return A List containing AgentID's of resolvers for this
	 */
  public List getResolvers()
  {
	return _resolvers;
  }  
  
	/** Converts this into the FIPA99 AgentID format
	 * @return A String representation of this
	 */
	public String toString()
	{
		// Create StringBuffer for efficiency
		StringBuffer buff = new StringBuffer();
		
		// Create header
		buff.append( BASETOKENS.OPENBRACKET );
		buff.append( FIPAMANCONSTANTS.AGENT_IDENTIFIER );
		buff.append( BASETOKENS.SPACE);
		
		// Add name of Agent
		buff.append( FIPAMANCONSTANTS.NAME );
		buff.append( BASETOKENS.SPACE);
		buff.append( _name );
		
		// Add addresses of Agent
		if( _addresses.size() > 0 )
		{
			buff.append( BASETOKENS.SPACE);
			buff.append( FIPAMANCONSTANTS.ADDRESSES );
			buff.append( BASETOKENS.SPACE);
			buff.append( BASETOKENS.OPENBRACKET );
			buff.append( FIPAMANCONSTANTS.SEQUENCE );
			
			// Iterate through URL's
			Iterator i = _addresses.iterator();
			while( i.hasNext() )
			{
				URL url = (URL) i.next();
				buff.append( BASETOKENS.SPACE);
				buff.append( url.toString() );
			}
			
			// End sequence
			buff.append( BASETOKENS.SPACE);
			buff.append( BASETOKENS.CLOSEBRACKET );
		}
		
		// Add resolvers of Agent
		if( _resolvers.size() > 0 )
		{
			buff.append( BASETOKENS.SPACE);
			buff.append( FIPAMANCONSTANTS.RESOLVERS );
			buff.append( BASETOKENS.SPACE);
			buff.append( BASETOKENS.OPENBRACKET );
			buff.append( FIPAMANCONSTANTS.SEQUENCE );
			
			// Iterate through URL's
			Iterator i = _resolvers.iterator();
			while( i.hasNext() )
			{
				AgentID aid = (AgentID) i.next();
				buff.append( BASETOKENS.SPACE);
				aid.toResolverString( buff );
			}
			
			// End sequence
			buff.append( BASETOKENS.SPACE);
			buff.append( BASETOKENS.CLOSEBRACKET );
		}
		
		// End
		buff.append( BASETOKENS.SPACE);
		buff.append( BASETOKENS.CLOSEBRACKET );
		return buff.toString();
	}
	
	
	/**
	 * Converts a list of stringified AgentID's into a List of AgentID objects
	 */
	public static List getAgentIDs( String aids )
	{
		aids = aids.trim();
		
		// First check to see if we're dealing with a set/sequence
		String seq = removeParentheses( aids );
		if( seq.startsWith( FIPAMANCONSTANTS.SET ) )
		{
			aids = seq.substring( seq.indexOf( FIPAMANCONSTANTS.SET ) + 
																			FIPAMANCONSTANTS.SET.length() ).trim();
		}
		else if( seq.startsWith( FIPAMANCONSTANTS.SEQUENCE ) )
		{
			aids = seq.substring( seq.indexOf( FIPAMANCONSTANTS.SEQUENCE ) + 
																			FIPAMANCONSTANTS.SEQUENCE.length() ).trim();			
		}
		
		// Create a list to contain the new AgentID's to create
		List out = new LinkedList();
		
		// This nest bit will get all of the resolvers in the list
		while( aids.indexOf( FIPAMANCONSTANTS.AGENT_IDENTIFIER ) >= 0 )
		{
			try
			{
				AgentID aid = AIDParser.parse( aids );
				out.add( aid );
			}
			catch(Exception pe)
			{
				DIAGNOSTICS.println( "Exception when parsing AgentID", pe, DIAGNOSTICS.LEVEL_2 );
			}

			// Find next resolver
			aids = removeNextParenthesis( aids );
		}
		
		return out;
	}
	
	
	/**
	 * Compare this with another object
	 */
	public boolean equals( Object other )
	{
		// Default result is false
		boolean result = false;
		
		if( other instanceof AgentID )
		{
			// Type-cast other object
			AgentID other_aid = (AgentID) other;
			
			// Check if agent name is identical
			result = other_aid._name.equalsIgnoreCase( _name );
		}
		
		// Return result
		return result;
	}
	
	/**
	 * Method to provide property hints to DynamicClass
	 */
	public static DynamicClass.PropertyHint getDynamicClassHints( String prop )
	{
		if( prop.equals( "Addresses" ) )
		{
			return new DynamicClass.CollectionHint( new Class[] { URL.class } );
		}
		else if( prop.equals( "Resolvers" ) ) 
		{
			return new DynamicClass.CollectionHint( new Class[] { AgentID.class } );
		}
		else
		{
			return null;
		}
	}
	
//------------------------------------------------------------------------------
// PRIVATE METHODS
//------------------------------------------------------------------------------
	
	private void toResolverString( StringBuffer buff )
	{
		// Create header
		buff.append( BASETOKENS.OPENBRACKET );
		buff.append( FIPAMANCONSTANTS.AGENT_IDENTIFIER );
		buff.append( BASETOKENS.SPACE);
		
		// Add name of Agent
		buff.append( FIPAMANCONSTANTS.NAME );
		buff.append( BASETOKENS.SPACE);
		buff.append( _name );
		
		// Add addresses of Agent
		if( _addresses.size() > 0 )
		{
			buff.append( BASETOKENS.SPACE);
			buff.append( FIPAMANCONSTANTS.ADDRESSES );
			buff.append( BASETOKENS.SPACE);
			buff.append( BASETOKENS.OPENBRACKET );
			buff.append( FIPAMANCONSTANTS.SEQUENCE );
			
			// Iterate through URL's
			Iterator i = _addresses.iterator();
			while( i.hasNext() )
			{
				URL url = (URL) i.next();
				buff.append( BASETOKENS.SPACE);
				buff.append( url.toString() );
			}
			
			// End sequence
			buff.append( BASETOKENS.SPACE);
			buff.append( BASETOKENS.CLOSEBRACKET );
		}
		
		// End
		buff.append( BASETOKENS.SPACE);		
		buff.append( BASETOKENS.CLOSEBRACKET );
	}
	
  private static int getNoOfResolvers( String resolvers )
  {
	int count = 0;
	while( resolvers.indexOf( "agent-identifier" ) >= 0 )
	{
	  resolvers = resolvers.substring( 0, resolvers.lastIndexOf( "agent-identifier") );
	  count++;
	}
	return count;
  }  
  
  private static String removeParentheses( String str )
  {
	str = str.trim();
	int start = 0;
	int end = 0;
	if( str.startsWith( BASETOKENS.OPENBRACKET ) )
	{
	  start = ( str.indexOf( BASETOKENS.OPENBRACKET ) + 1) ;
	  str = str.substring( start, str.length() );
	}
	if( str.endsWith( BASETOKENS.CLOSEBRACKET ) )
	{
	  end = str.lastIndexOf( BASETOKENS.CLOSEBRACKET );
	  str = str.substring( 0, str.length() - 1);
	}
	return str.trim();
  }  
	
	private static String removeNextParenthesis(String str)
	{
		// Count up and down brackets until we either: a) reach 0 again, or b)
		// reach the end of the String
		StringTokenizer st = new StringTokenizer( str, "()", true );
		
		int level = 0;
		boolean start = true;
		while( st.hasMoreTokens() && ( level != 0 || start ) )
		{
			String next = st.nextToken();
			if( next.equals( BASETOKENS.OPENBRACKET ) )
			{
				level++;
				start = false;
			}
			else if( next.equals( BASETOKENS.CLOSEBRACKET ) )
			{
				level--;
			}
		}
		
		if( st.hasMoreTokens() )
		{
			return st.nextToken( "" );
		}
		else
		{
			return "";
		}
	}
	
	/*
	 * Returns the unique hashcode of this AgentID
	 * Uniqueness is guaranteed by the fact that the name component of an
	 * AgentId is required to be globally unique.
	 *
	 * @return int. The hashcode of this AgentID
	 */
	public int hashCode()
	{
		
		return _name.toLowerCase().hashCode();		
	}
  
//------------------------------------------------------------------------------
// TEST HARNESS
//------------------------------------------------------------------------------
	
	public static void main(String[] args) throws Exception
	{
		String aid = "(agent-identifier :name some-agent " +
											":addresses (sequence fipaos-rmi://localhost:1099/ iiop://somehost.com:8888/acc) " +
											":resolvers (sequence (agent-identifier :name ams) (agent-identifier :name acc) ) )";
		
		System.out.println( "Stringified AID == " + aid );
		
		AgentID aid_obj = new AgentID( aid );
		
		System.out.println( "\n\naid_obj.toString() ==  " + aid_obj.toString() );
		
		AgentID aid_obj2 = new AgentID( aid_obj.toString() );
		
		System.out.println( "\n\naid_obj2.toString() == " + aid_obj2.toString() );
		
		if( aid_obj.toString().equals( aid_obj2.toString() ) )
		{
			System.out.println( "Parse and deparse work OK" );
		}
	}  
}