//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.ont.fipa.fipaman;

import fipaos.mts.event.*;
import fipaos.util.*;
import fipaos.ont.fipa.ACL;
import fipaos.ont.fipa.FIPACONSTANTS;

import java.io.Serializable;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;
import java.util.List;
import java.util.ListIterator;
import java.util.LinkedList;
import java.util.Arrays;
import java.util.Set;
import java.util.TreeSet;

/** <p>This class provides a representation of the FIPA99/2000 envelope (implements XC00067C)
 *
 * <p>In conformance with the FIPA99/2000 specification slots cannot be overwriten, just
 * overridden.
 */
public class Envelope implements Serializable, Cloneable
{
	/**
	 * Array of non-user slots
	 */
	private static final String STANDARD_SLOTS[] = { FIPACONSTANTS.ACL_REPRESENTATION,
																							FIPACONSTANTS.COMMENTS,
																							FIPACONSTANTS.DATE,
																							FIPACONSTANTS.ENCRYPTED,
																							FIPACONSTANTS.INTENDED_RECEIVER,
																							FIPACONSTANTS.RECEIVED,
																							FIPACONSTANTS.RECEIVER,
																							FIPACONSTANTS.SENDER,
																							FIPACONSTANTS.TRANSPORT_BEHAVIOUR,
																							FIPACONSTANTS.PAYLOAD_ENCODING,
																							FIPACONSTANTS.PAYLOAD_LENGTH };
	
	
	/**
	 * Slots are contained within this Map, key => value(s).  The values are
	 * represented as a List containing Strings making up entries of the slot 
	 */
	private Map _slots;
	
	/**
	 * List of Envelope's composing this envelope
	 */
	private List _sub_envelopes;
	
	/**
	 * Envelope that this belongs to
	 */
	private Envelope _parent;
	
	/**
	 * Current envelope object (to be added to list of sub-envelopes)
	 */
	private Envelope _current_envelope;
	
	/**
	 * Error state for the message this envelope contains
	 */
	private transient int _error_code = 0;
	
	/**
	 * Indicates if this message should actually be routed via the ACC
	 */
	private transient boolean _via_acc = false;
	
	/**
	 * List of transports "tried" when trying to send this message
	 */
	private transient List _transports_tried;
	
	/**
	 * List of URL's "tried" when trying to send this message, but no MTP could deal with it
	 */
	private transient List _urls_tried;
	
	/**
	 * List of MessageListener's interested in the message this is associated with
	 */
	private transient List _listeners;
	
	/**
	 * Default constructor - creates an empty envelope
	 */
	public Envelope()
	{
		this( false );
	}
	
	
	/** Initialises the Envelope based upon whether it is a sub-envelope or not
	 * @param subenvelope If True, this is a sub-envelope
	 */
	private Envelope( boolean subenvelope )
	{
		// Setup variables
		_slots = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
		_sub_envelopes = new LinkedList();

		// Check if this is not a sub-envelope (i.e. top-level)
		if( !subenvelope )
		{
			// Setup sub-envelope handling instance variables
			_current_envelope = new Envelope( true );
			_current_envelope._parent = this;
			_sub_envelopes.add( _current_envelope );
		}
	}
	
	
	/** Creates a copy of the given Envelope object
	 * @param env Envelope to duplicate
	 *
	 * @deprecated Please use clone() instead
	 */
	public Envelope( Envelope env )
	{
		// Clone parent if necessary
		if( env._parent != env._parent )
		{
			_parent = env._parent;
		}

		if( env._slots != null )
		{
			// Clone slots
			_slots = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
			Iterator i = env._slots.keySet().iterator();
			while( i.hasNext() )
			{
				String next = (String) i.next();
				_slots.put( next, new LinkedList( (List) env._slots.get( next ) ) );
			}
		}

		// Clone sub-envelopes
		if( env._sub_envelopes != null )
		{
			// Clone sub-envelopes
			_sub_envelopes = new LinkedList();
			Iterator i = env._sub_envelopes.iterator();
			while( i.hasNext() )
			{
				_sub_envelopes.add( ( (Envelope) i.next() ).clone() );
			}
		}

		// Clone current envelope
		if( env._current_envelope != null )
		{
			_current_envelope = (Envelope) env._current_envelope.clone();
		}
		
		if( env._listeners != null )
		{
			_listeners = new LinkedList( env._listeners );
		}
	}
	
	
	/** Creates an Envelope based upon the given ACL object
	 * @param msg ACL to base the Envelope on
	 */
	public Envelope( ACL msg )
	{
		// Invoked default constructor
		this( false );
		
		// Copy sender and receiver fields of ACL message
		if( msg.getSenderAID() != null )
		{
			setSender( (AgentID) msg.getSenderAID().clone() );
		}
		
		if( msg.getReceiverAIDs() != null )
		{
			List l = new LinkedList();
			Iterator i = msg.getReceiverAIDs().iterator();
			while( i.hasNext() )
			{
				l.add( ( (AgentID) i.next() ).clone() );
			}
			setReceivers( l );
		}
		
		// Set the creation time of the envelope
		setDate( new UTCTime() );
		
		// Set the default encoding type
		setACLRepresentation( ACL.FIPA00070 );
	}
	

	public Object clone()
	{
		try
		{
			Envelope env = (Envelope) super.clone();
			
			// Setup parent if necessary
			if( _parent != null )
			{
				if( _parent == this )
				{
					env._parent = env;
				}
				else
				{
					// We don't explicitly clone our parent, else we end up recursively
					// cloning our parent, and our parent recursively cloning its children!!!
				}
			}
			
			if( _slots != null )
			{
				// Clone slots
				env._slots = new TreeMap( fipaos.util.Comparators.STRING_COMPARE );
				Iterator i = _slots.keySet().iterator();
				while( i.hasNext() )
				{
					String next = (String) i.next();
					env._slots.put( next, cloneSlot( _slots.get( next ) ) );
				}
			}

			// Clone sub-envelopes
			if( _sub_envelopes != null )
			{
				// Clone sub-envelopes
				env._sub_envelopes = new LinkedList();
				Iterator i = _sub_envelopes.iterator();
				while( i.hasNext() )
				{
					// Get next, clone, setup-parent and add to clones' sub-envelopes
					Envelope next = (Envelope) ( (Envelope) i.next() ).clone();
					next._parent = env;
					env._sub_envelopes.add( next );
				}
				
				// Get clone of current envelope
				if( _current_envelope != null )
				{
					env._current_envelope = (Envelope) env._sub_envelopes.get( 0 );
				}
			}
			
			// Duplicate transport details
			if( _transports_tried != null )
			{
				// Clone sub-envelopes
				env._transports_tried = new LinkedList( _transports_tried );
			}
			
			// Duplicate unsuccesful URL details
			if( _urls_tried != null )
			{
				env._urls_tried = new LinkedList();
				Iterator i = _urls_tried.iterator();
				while( i.hasNext() )
				{
					// Get next, clone, setup-parent and add to clones' sub-envelopes
					env._urls_tried.add( ( (URL) i.next() ).clone() );
				}				
			}
			
			return env;
		}
		catch( CloneNotSupportedException e )
		{
			return null;
		}
	}
	
	/**
	 * Attempt to clone a value that has been inserted into a slot
	 */
	public Object cloneSlot( Object slot )
	{
		// Check the type, and clone appropriately
		if( slot instanceof List )
		{
			// Recursively clone list contents
			List l = new LinkedList();
			Iterator i = ((List) slot).iterator();
			while( i.hasNext() )
			{
				l.add( cloneSlot( i.next() ) );
			}
			return l;
		}
		else if( slot instanceof Set ) 
		{
			// Recursively clone set contents
			Set s = new TreeSet( fipaos.util.Comparators.OBJECT_COMPARE );
			Iterator i = ((Set) slot).iterator();
			while( i.hasNext() )
			{
				s.add( cloneSlot( i.next() ) );
			}
			return s;			
		}
		else if( slot instanceof Map ) 
		{
			// Recursively clone map contents
			Map m = new TreeMap( fipaos.util.Comparators.OBJECT_COMPARE );
			Iterator i = ((Map) slot).keySet().iterator();
			while( i.hasNext() )
			{
				Object key = i.next();
				m.put( cloneSlot( key ), cloneSlot( ((Map) slot).get( key ) ) );
			}
			return m;			
		}
		else if( slot instanceof String )
		{
			return new String( (String) slot );
		}
		else if( slot instanceof AgentID )
		{
			return ( (AgentID) slot ).clone();
		}
		else if( slot instanceof URL )
		{
			return ( (URL) slot ).clone();			
		}
		else if( slot instanceof UTCContainer )
		{
			return ( (UTCContainer) slot ).clone();						
		}
		else if( slot instanceof Received )
		{
			return ( (Received) slot ).clone();
		}
		else if( slot instanceof TransportBehaviour )
		{
			return ( (TransportBehaviour) slot ).clone();			
		}
		else
		{
			return slot.toString();
		}
	}
	
	
	/**
	 * Add a MessageListener to this
	 */
	public void addMessageListener( MessageListener ml )
	{
		if( _listeners == null )
		{
			_listeners = new LinkedList();
		}
		
		if( !_listeners.contains( ml ) )
		{
			_listeners.add( ml );
		}
	}
	
	
	/**
	 * Remove a MessageListener from this
	 */
	public void removeMessageListener( MessageListener ml )
	{
		if( _listeners == null )
		{
			return;
		}
		
		if( _listeners.contains( ml ) )
		{
			_listeners.remove( ml );
		}
	}
	
	
	/**
	 * Get list of MessageListeners
	 */
	public List getMessageListeners()
	{
		List listeners = new LinkedList();
		
		if( _listeners != null )
		{
			listeners.addAll( _listeners );
		}
		
		return listeners;
	}

	
	/** Indicates that a new sub-envelope should be started (i.e. when the
	 * Envelope is received by a new ACC)
	 */
	public synchronized void newSubEnvelope()
	{
		// Check this isn't a sub-envelope - if so just return
		if( _parent != null )
		{
			return;
		}
		
		// Create a new sub-envelope to contain the next ACC's
		// changes
		_current_envelope = new Envelope( true );
		_current_envelope._parent = this;
		
		// Save sub-envelope
		_sub_envelopes.add( _current_envelope );
	}
	
	
	/** Returns a list of the sub-envelopes composing this
	 * @return A List of Envelope objects
	 */
	public synchronized List getSubEnvelopes()
	{
		// Only retreive if we aren't a sub-envelope
		if( _parent == null )
		{
			return new LinkedList( _sub_envelopes );
		}
		
		return null;
	}
	
	
	/** Gets a particular sub-envelope
	 * @param x The sub-envelope number to get
	 * @return A sub-envelope of this
	 */
	public synchronized Envelope getSubEnvelope( int x )
	{
		// Get the envelope at position x
		if( x >= _sub_envelopes.size() || _parent != null )
		{
			return null;
		}
		else
		{
			return (Envelope) _sub_envelopes.get( x );
		}
	}
	
	
	/** Sets the ACL representation of this Envelope
	 * @param type The type of the ACL representation (i.e. RDF)
	 */
	public void setACLRepresentation( String type )
	{
		setSlotValue( FIPACONSTANTS.ACL_REPRESENTATION, type );
	}
	
	
	/** Gets the ACL representation associated with this envelope
	 * @return ACL encoding
	 */
	public String getACLRepresentation()
	{
		return (String) getSlotValue( FIPACONSTANTS.ACL_REPRESENTATION );
	}
	
	
	/** Sets the comments associated with this envelope
	 * @param comments Comment to add
	 */
	public void setComments( String comments )
	{
		setSlotValue( FIPACONSTANTS.COMMENTS, comments );
	}


	/** Gets the comments associated with this Envelope
	 * @return The comments
	 */
	public String getComments()
	{
		return (String) getSlotValue( FIPACONSTANTS.COMMENTS );
	}
	
	
	/** Sets the payload encoding associated with this envelope
	 */
	public void setPayloadEncoding( String encoding )
	{
		setSlotValue( FIPACONSTANTS.PAYLOAD_ENCODING, encoding );
	}


	/** Gets the payload encoding associated with this Envelope
	 */
	public String getPayloadEncoding()
	{
		return (String) getSlotValue( FIPACONSTANTS.PAYLOAD_ENCODING );
	}
	
	
	/** Sets the payload length associated with this envelope
	 * @param length of payload (or -1 for none)
	 */
	public void setPayloadLength( int length )
	{
		if( length > -1 )
		{
			setSlotValue( FIPACONSTANTS.PAYLOAD_ENCODING, new Integer( length ) );
		}
		else
		{
			setSlotValue( FIPACONSTANTS.PAYLOAD_ENCODING, null );
		}
	}


	/** Gets the payload length associated with this Envelope
	 * @return The length, or -1 if this hasn't been set
	 */
	public int getPayloadLength()
	{
		Integer i = (Integer) getSlotValue( FIPACONSTANTS.PAYLOAD_LENGTH );
		
		if( i != null )
		{
			return i.intValue();
		}
		else
		{
			return -1;
		}
	}
	
	
	/** Sets the date of creation for this envelope
	 * @param utc The UTCTime representing the creation time of the envelope
	 */
	public void setDate( UTCTime utc )
	{
		setSlotValue( FIPACONSTANTS.DATE, utc );
	}

	
	/** Gets the creating date of this Envelope
	 * @return UTCTime representing the creation time
	 */
	public UTCTime getDate()
	{
		return (UTCTime) getSlotValue( FIPACONSTANTS.DATE );
	}
	
	
	/** Sets details of the encryption used on the content of the envelope
	 * @param enc Details of encryption applied
	 */
	public void setEncrypted( List enc )
	{
		setSlotValue( FIPACONSTANTS.ENCRYPTED, enc );
	}

	
	/** Gets details of the encryption used on the content of the envelope
	 * @return Details of the encryption used
	 */
	public List getEncrypted()
	{
		return (List) getSlotValue( FIPACONSTANTS.ENCRYPTED );
	}
	
	
	/** Sets details of when the message was received by an entity
	 * @param received Description of the time & entity
	 */
	public void setReceived( Received received )
	{
		setSlotValue( FIPACONSTANTS.RECEIVED, received );
	}

	
	/** Gets details of when the message was received by an entity
	 * @return Details of when & what entity received the message last
	 */
	public Received getReceived()
	{
		return (Received) getSlotValue( FIPACONSTANTS.RECEIVED );
	}
	
	
	/** Sets the sender of the message
	 * @param aid The AID belonging to the sending Agent
	 */
	public void setSender( AgentID aid )
	{
		setSlotValue( FIPACONSTANTS.SENDER, aid );
	}

	
	/** Gets the sender of the message
	 * @return The AID of the sending Agent
	 */
	public AgentID getSender()
	{
		return (AgentID) getSlotValue( FIPACONSTANTS.SENDER );
	}
	
	
	/** Sets the list of receivers for the message
	 * @param aids A List of AgentID's belonging to the recipients of the message
	 */
	public void setReceivers( List aids )
	{
		setSlotValue( FIPACONSTANTS.RECEIVER, aids );
	}

	
	/** Gets a List of the recipients' AID's
	 * @return List of the recipients' AID's
	 */
	public List getReceivers()
	{
		return (List) getSlotValue( FIPACONSTANTS.RECEIVER );
	}
	
	
	/** Sets the intended receivers for the message
	 * @param aids A List of AgentID's
	 */
	public void setIntendedReceiver( List aids )
	{
		setSlotValue( FIPACONSTANTS.INTENDED_RECEIVER, aids );
	}

	
	/** Gets the intended receivers for the message
	 * @return AgentID belonging to the intended recipient of the message
	 */
	public List getIntendedReceiver()
	{
		return (List) getSlotValue( FIPACONSTANTS.INTENDED_RECEIVER );
	}
	
	
	/** Sets the TransportBehviour parameter
	 * @param behaviour A description of the behaviour to use
	 */
	public void setTransportBehaviour( TransportBehaviour behaviour )
	{
		setSlotValue( FIPACONSTANTS.TRANSPORT_BEHAVIOUR, behaviour );
	}

	
	/** Gets the TransportBehviour parameter
	 * @return A description of the transport level behaviour
	 */
	public TransportBehaviour getTransportBehaviour()
	{
		return (TransportBehaviour) getSlotValue( FIPACONSTANTS.TRANSPORT_BEHAVIOUR );
	}
		

	/** <p>Set the internal error condition for this envelope
	 * <p> N.B. This is not part of the FIPA specification, but
	 * servers as a mechanism for "tagging" messages with "exception" codes
	 * at the transport layer
	 * @param code
	 */
	public void setErrorCode( int code )
	{
		_error_code = code;
	}
	
	
	/** Get the internal error condition for this envelope
	 * <p> N.B. This is not part of the FIPA specification, but
	 * servers as a mechanism for "tagging" messages with "exception" codes
	 * at the transport layer
	 * @return
	 */
	public int getErrorCode()
	{
		return _error_code;
	}
	
	
	/** Sets the flag that indicates if the message this encapsulates should be
	 * routed via the ACC
	 * @param value If true, indicates this message should be routed via the ACC
	 */
	public void setViaACC( boolean value )
	{
		_via_acc = value;
	}
	
	
	/** Indicates if this should be routed via the ACC
	 * @return If true, this should be routed via the ACC
	 */
	public boolean getViaACC()
	{
		return _via_acc;
	}
	
	
	
	/** Returns the transports that have been tried whilst attempting to send
	 * this message
	 * @return A List of transports tried (by protocol)
	 */
	public List getTransportsTried()
	{
		// Check if tried transports has been initialised
		if( _transports_tried == null )
		{
			_transports_tried = new LinkedList();
		}
		
		return _transports_tried;
	}
	
	
	/** Returns the URL's that have been tried, but an MTP implementation
	 * hasn't been found
	 * @return A List of transports tried (by protocol)
	 */
	public List getUnknownURLsTried()
	{
		// Check if tried transports has been initialised
		if( _urls_tried == null )
		{
			_urls_tried = new LinkedList();
		}
		
		return _urls_tried;
	}
	
	
	/** Generic method for adding a value to a slot within the envelope
	 * @param slot Slot name
	 * @param value Value
	 */
	public synchronized void setSlotValue( String slot, Object value )
	{
		Envelope target = null;
		
		// Check if we are dealing with a single envelope or multiple envelopes
		if( _parent == null )
		{
			_current_envelope.setSlotValue( slot, value );
		}
		else
		{
			// Add value to this Envelope
			_slots.put( slot, value );
		}
	}
	
		
	/** Generic method for retreiving the value assigned to a slot within the envelope
	 * @param slot Slot name
	 * @return Most recent value associated with a slot
	 */
	public synchronized Object getSlotValue( String slot )
	{
		Object result = null;
		
		// Check if this is a sub-envelope
		if( _parent != null )
		{
			// Return the value contained within this
			result = _slots.get( slot );
		}
		else
		{
			// Iterate through (in reverse order) the sub-envelopes
			// and get the value
			Iterator i = _sub_envelopes.iterator();
			while( i.hasNext() )
			{
				// Get previous envelope
				Envelope env = (Envelope) i.next();
				
				// Check for value of slot
				Object value = env.getSlotValue( slot );
				if( value != null )
				{
					result = value;
				}
			}
		}
		
		return result;
	}
	
	
	/** Returns a list of the names of non-standard slots contained within this
	 * Envelope
	 * @return A Set of user-defined slot names
	 */
	public synchronized Set getUserSlots()
	{
		Set slots = null;
		
		// Check if this is a sub-envelope or not
		if( _parent != null )
		{
			// Return user-slots for this envelope only
			
			// Create list of standard slots
			List std = Arrays.asList( STANDARD_SLOTS );
			
			// Get a list of slots & remove those we know about
			slots = new TreeSet( fipaos.util.Comparators.STRING_COMPARE );
			slots.addAll( _slots.keySet() );
			slots.removeAll( std );
		}
		else
		{
			// Create a set of all user slots in sub-envelopes
			
			// Create set to store info within
			slots = new TreeSet( fipaos.util.Comparators.STRING_COMPARE );
			
			// Iterate through sub-envelopes and amalgamate results
			Iterator i = _sub_envelopes.iterator();
			while( i.hasNext() )
			{
				// Get the next envelope, and its user slots
				Envelope env = (Envelope) i.next();
				slots.addAll( env.getUserSlots() );
			}
		}
		
		// Return result
		return slots;
	}
	
	
	/** Overrides default toString() method - displays contents of Envelope
	 * @return Stringified representation of this
	 */
	public String toString()
	{
		// Convert appropriate collection to a String to represent this
		return ( _parent == null ? _sub_envelopes.toString() : _slots.toString() );
	}
	
	
	/**
	 * Test-harness
	 */
	public static void main( String args[] )
	{
		System.out.println( "Envelope Test-Harness" );
		System.out.println( "=====================\n" );
		
		// Attempt to create new Envelope
		Envelope env = new Envelope();
		System.out.println( "env = " + env );
		
		// Add items to a random slot
		env.setSlotValue( FIPACONSTANTS.SENDER, "abc" );
		System.out.println( "env = " + env );
		env.newSubEnvelope();
		env.setSlotValue( FIPACONSTANTS.SENDER, "mnp" );
		System.out.println( "env = " + env );
		env.newSubEnvelope();
		env.setSlotValue( ":X-SomeOther", "xyz" );
		System.out.println( "env = " + env );
		
		System.out.println( "\n\n" + env.getUserSlots() );
		
		// Get the slot value
		if( env.getSlotValue( FIPACONSTANTS.SENDER ).equals( "mnp" ) &&
				env.getSlotValue( ":X-SomeOther" ).equals( "xyz" ) &&
				env.getUserSlots().size() == 1 &&
				env.getUserSlots().contains( ":X-SomeOther" ) )
		{
			System.out.println( "Test passed OK" );
		}
		else
		{
			System.out.println( "Test failed" );
		}
	}
}