//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.parser.acl;

// ---------------------------------------------------------------------------

import java.util.Enumeration;
import java.util.Vector;
import java.util.StringTokenizer;
import java.util.List;
import java.util.LinkedList;
import java.util.Collections;

import fipaos.parser.acl.string.ACLParser;
import fipaos.parser.acl.string.ParseException;
import fipaos.ont.fipa.fipaman.AgentID;
import fipaos.ont.fipa.ACL;
import fipaos.ont.BASETOKENS;
import fipaos.util.UTCTime;
import fipaos.util.UTCDuration;
import fipaos.util.UTCContainer;

// ---------------------------------------------------------------------------

/**
 *	ACLMessage is the java object for an acl message string
 **/
public class ACLMessage extends ACL implements java.io.Serializable
{
	/**
	 * Constructor of ACLMessage
	 **/
	public ACLMessage()
	{
	}
	
	
	/**
	 * Duplicate the given ACL object
	 */
	public ACLMessage( ACL other )
	{
		super( other );
	}
	
	
	/**
	 *	Sets Message performative
	 *  @param n value to be set
	 * @deprecated Please use setPerformative() in preference to this
	 **/
	public void setMessageType(String n)
	{
		super.setPerformative( n );
	}
	
	
	/**
	 *	Sets Sender
	 *  @param n value to be set
	 */
	public void setSender(String n)
	{
		try
		{
			// Attempt to set the AID of the target Agent
			super.setSenderAID( fipaos.parser.aid.AIDParser.parse( n ) );
		}
		catch( Exception e )
		{
			e.printStackTrace();
		}
	}
	
	
	/**
	 *	Sets Receiver
	 *  @param n value to be set
	 **/
	public void setReceiver( String n )
	{
		if( n != null )
		{
			// Create AID List
			List aids = AgentID.getAgentIDs( n );
			super.setReceiverAIDs( aids );
		}
		else
		{
			super.setReceiverAIDs( null );
		}
	}
	
	
	/**
	 * Sets the content of the message
	 *
	 * @param the_content value to be set
	 * NB:: the_content must not be null.
	 */
	public void setContent( String the_content )
	{
		super.setContentObject( the_content );
	} // end setContent
	
	
	
	/**
	 * Sets the content, optionally encoding the content
	 * as the specified type (see FIPA specification)
	 *
	 * @param content The content of the message
	 * @param encode Indicates the type of encoding to use for the string
	 */
	public void setContent( String content, int encode )
	{
		super.setContentObject( content, encode );
	}
	
	
	/**
	 * Set the AID of the Agent to reply to
	 */
	public void setReplyTo(String n)
	{
		if( n != null )
		{
			// Create AID List
			List aids = AgentID.getAgentIDs( n );
			super.setReplyToAIDs( aids );
		}
		else
		{
			super.setReplyToAIDs( null );
		}
	}
	
	
	/**
	 * Sets the reply-by time of this
	 */
	public void setReplyBy( String reply_by )
	{
		UTCContainer utc = null;
		
		try
		{
			utc = new UTCTime( reply_by );
		}
		catch( Throwable t )
		{
			utc = new UTCDuration( reply_by );
		}
		
		super.setReplyByUTC( utc );
	}
	
	
	/**
	 *	Gets Message performative
	 *  @return type_
	 * @deprecated Please use getPerformative() in preference to this
	 **/
	public String getMessageType()
	{
		return super.getPerformative();
	}
	
	
	/**
	 *	Gets Sender AID
	 *  @return sender_
	 **/
	public String getSender()
	{
		return super.getSenderAID().toString();
	}
	
	
	/**
	 *	Gets Receiver
	 *  @return receiver_
	 **/
	public String getReceiver()
	{
		// Just return the first receivers' AID
		List l = super.getReceiverAIDs();
		
		String aid = null;
		
		// Check that there is at least one receiver specified
		if( l != null && l.size() > 0 )
		{
			aid = ( (AgentID) l.get( 0 ) ).toString();
		}
		
		return aid;
	}
	
	
	/**
	 * Gets the content, dencoding the content as if it were encoded using the
	 * given encoding type (see FIPA specification)
	 *
	 * @param incoding Indicates the type of encoding the content is in
	 * @return The content of the message
	 */
	public String getContent()
	{
		Object orig_content = super.getContentObject();
		
		// Check we're dealing with a compatible type
		if( ! ( orig_content instanceof String ) )
		{
			return null;
		}
		
		String content = (String) orig_content;
		
		return content;
	}
	
	
	/**
	 *	Gets Replyby
	 *  @return replyby_
	 **/
	public String getReplyBy()
	{
		return super.getReplyByUTC().toString();
	}
	
	
	/**
	 * Get the AID of the first Agent to reply to
	 */
	public String getReplyTo()
	{
		// Just return the first receivers' AID
		List l = super.getReplyToAIDs();
		
		String aid = null;
		
		// Check that there is at least one receiver specified
		if( l != null && l.size() > 0 )
		{
			aid = (String) l.get( 0 );
		}
		
		return aid;
	}
	
	
	/**
	 *	Gets Envelope
	 *  @return envelope_
	 * @deprecated This method is no longer supported - please use the Envelope
	 * class instead
	 */
	public Vector getEnvelope()
	{
/*		if( envelope_ == null || envelope_.isEmpty() )
			return null;
 
		Vector out = new Vector();
 
		java.util.Enumeration enum = envelope_.elements();
		while( enum.hasMoreElements() )
			out.addElement( enum.nextElement() );
 
		return out;
 */
		return null;
	}
	
	
	//----------------------------------------------------------------------
	// TEST HARNESS
	//----------------------------------------------------------------------
	
	public static void main( String args[] ) throws Exception
	{
		String aid = "(agent-identifier :name agent-a " +
		":addresses (sequence fipaos-rmi://localhost:1099/ iiop://somehost.com:8888/acc) " +
		":resolvers (sequence (agent-identifier :name ams) (agent-identifier :name acc) ) )";
		
		String bid = "(agent-identifier :name agent-b " +
		":addresses (sequence fipaos-rmi://localhost:1099/ iiop://somehost.com:8989/acc) " +
		":resolvers (sequence (agent-identifier :name ams) (agent-identifier :name acc) ) )";
		
		
		ACLMessage acl = new ACLMessage();
		acl.setPerformative( "inform" );
		acl.setOntology( "my-ontology" );
		acl.setLanguage( "my-language" );
		acl.setProtocol( "my-protocol" );
		acl.setSender( aid );
		acl.setReceiver( bid );
		acl.setContent( "( Test Message )" );
		acl.setConversationID( "test-conv-id-2348723648" );
		
		System.out.println( "acl ==  " + acl.toString() );
		
		ACL acl2 = new ACL( acl.toString() );
		
		System.out.println( "acl2 == " + acl2.toString() );
	}
}