//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved.
//
//******************************************************************************

package fipaos.util;

import java.io.*;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Date;

import fipaos.util.Sink;

/** Object which can be used to handles all status and error messages.
 * <br> All methods are static - this class does not need to be instantiated
 */
public class DIAGNOSTICS
{
	/** minimum diagnostic level: produces the maximum of information */
	public static final int LEVEL_MIN = 1;
	/** diagnostic level */
	public static final int LEVEL_2 = 2;
	/** diagnostic level */
	public static final int LEVEL_3 = 3;
	/** diagnostic level */
	public static final int LEVEL_4 = 4;
	/** maximum diagnostic level: produces the minimum of information */
	public static final int LEVEL_MAX = 5;
	
	/** is used to 'Tokenize' error messages stored in file */
	private static final String DELIMITER = " | ";
	
	/** the current diagnostics level being used */
	private static int __current_level = LEVEL_MIN;
	private static int __memory_level;
	private static int __file_level;

	/**
	 * Reference to a single DIAGNOSTICS object
	 */
	private static DIAGNOSTICS __diag_obj;
	
	/**
	 * Indicates if DIAGNOSTICS are off
	 */
	private static boolean __diagnostics_off = System.getProperty( "debug.off", "false" ).equals( "true" );
	
	/**
	 * Name for the DIAGNOSITCS file this produces
	 */
	private String _filename = "DIAGNOSTICS at " + new UTCTime() + ".txt";
	
	/**
	 * Output file
	 */
	private BufferedWriter _out;
	
	/**
	 * Queue of DIAGNOSTICS messages to write to memory
	 */
	private Queue _memory_queue = new FIFOQueue();
	
	/**
	 * Queue of DIAGNOSTICS messages to write to file
	 */
	private Queue _file_queue = new FIFOQueue();
	
	/**
	 * Monitor of the memory message queue
	 */
	private Monitor _monitor_memory;
	
	/**
	 * Monitor of the file message queue
	 */
	private Monitor _monitor_file;
	
	/**
	 * Lock to prevent multiple threads accessing the buffers
	 */
	private Object _lock = new Object();

	static
	{
		// Attempt to get a new debug memory level, if specified
		try
		{
			__memory_level = ( new Integer( System.getProperty( "debug.level.memory", "3" ) ) ).intValue();
		}
		catch( Throwable t )
		{
			System.err.println( "Error parsing DIAGNOSTICS memory debug level!" );
		}
		
		
		// Attempt to get a new debug memory level, if specified
		try
		{
			__file_level = ( new Integer( System.getProperty( "debug.level.file", "2" ) ) ).intValue();
		}
		catch( Throwable t )
		{
			System.err.println( "Error parsing DIAGNOSTICS file debug level!" );
		}
		
		// Create DIAGNOSTICS object
		__diag_obj = new DIAGNOSTICS();
	}
	
	/**
	 * Turn DIAGNOSTICS on or off
	 */
	public static void setEnabled( boolean on )
	{
		__diagnostics_off = !on;
	}


	/**
	 * Flush the contents of the buffers
	 */
	public static void flush()
	{
		if( __diagnostics_off )
		{
			return;
		}

		synchronized( __diag_obj._lock )
		{
			// Kill monitors
			__diag_obj._monitor_memory.halt();
			__diag_obj._monitor_file.halt();

			// Do the flush
			__diag_obj.flushFile();
			
			// Create new DIAGNOSTICS
			DIAGNOSTICS diag = __diag_obj;
			__diag_obj = new DIAGNOSTICS();
			__diag_obj._filename = diag._filename;
		}
	}
	
	/**
	 * Flush the buffers and close current output file
	 */
	public static void close()
	{
		if( __diagnostics_off )
		{
			return;
		}

		synchronized( __diag_obj._lock )
		{
			// Kill monitors
			__diag_obj._monitor_memory.halt();
			__diag_obj._monitor_file.halt();

			// Do the flush
			__diag_obj.flushFile();
			__diag_obj.closeFile();
			
			// Create new DIAGNOSTICS
			__diag_obj = new DIAGNOSTICS();
		}		
	}
		
	/**
	 * Flush the buffers and close current output file - when a new output file
	 * is created use the given name
	 */
	public static void close( String name )
	{
		if( __diagnostics_off )
		{
			return;
		}

		synchronized( __diag_obj._lock )
		{
			// Kill monitors
			__diag_obj._monitor_memory.halt();
			__diag_obj._monitor_file.halt();

			// Do the flush
			__diag_obj.flushFile();
			__diag_obj.closeFile();
			
			// Create new DIAGNOSTICS
			DIAGNOSTICS diag = new DIAGNOSTICS();
			diag._filename = name;
			__diag_obj = diag;
		}		
	}
		
	/**
	 * This does nothing except provide a breakpoint location for debugging
	 * Creation date: (9/27/99 10:37:10 AM)
	 * @param condition boolean
	 */
	public static void assert(boolean condition)
	{
		if( __diagnostics_off )
		{
			return;
		}

		boolean s =  false;
		
		if (condition)
		{
			s = true;
		}
	}
	
	/**
	 * It outputs message if condition is true
	 * This provides a breakpoint location for debugging
	 * Creation date: (9/27/99 10:37:10 AM)
	 * @param condition boolean
	 */
	public static void assert(boolean condition, String msg)
	{
		if( __diagnostics_off )
		{
			return;
		}

		if (condition)
		{
			println("Assert failed",msg);
		}
	}

	/**
	 * get the level which filters whether messages are saved to disk
	 * Creation date: (3/3/00 3:01:08 AM)
	 * @return int
	 */
	public static int get_file_level()
	{
		return __file_level;
	}
	
	/**
	 * Insert the method's description here.
	 * Creation date: (3/3/00 3:00:47 AM)
	 * @return int
	 */
	public static int get_memory_level()
	{
		return __memory_level;
	}
	

	/*
	 * Deprecated empty lines are ignored
	 */
	public static void println()
	{
	}
	
	
	/**
	 * Receive data including exception & pass it out
	 */
	public static void println( Object data, Throwable exception )
	{
		if( __diagnostics_off )
		{
			return;
		}

		println(data,exception,(Object)null,__current_level);
	}
	
	
	/*
	 * Receive data  and output it to any file set and to
	 * any process [data sink] that is attached
	 * @param data Object the data received
	 */
	public static void println( Object data )
	{
		if( __diagnostics_off )
		{
			return;
		}

		if( __current_level >= __file_level || __current_level >= __memory_level )
		{
			println(data, (Object)null,__current_level);
		}
	}
	
	
	/**
	 * Receive data including exception & pass it out
	 */
	public static void println( Object data, Throwable exception, int level )
	{
		if( __diagnostics_off )
		{
			return;
		}

		println( data, exception, (Object)null, level );
	}
	
	/*
	 * Receive data  and output it to any file set and to
	 * any process [data sink] that is attached
	 * @param data Object the data received
	 * @param source Object the sender of the data
	 */
	public static void println( Object data, int alevel)
	{
		if( __diagnostics_off )
		{
			return;
		}

		println(data, (Object)null,alevel);
	}
	
	
	/**
	 * Receive data including exception & pass it out
	 */
	public static void println( Object data, Throwable exception, Object source )
	{
		if( __diagnostics_off )
		{
			return;
		}

		if( __current_level >= __file_level || __current_level >= __memory_level )
		{
			println(data,exception,source,__current_level);
		}
	}
	
	/*
	 * Receive data  and output it to any file set and to
	 * any process [data sink] that is attached
	 * @param data Object the data received
	 * @param source Object the sender of the data
	 */
	public static void println( Object data, Object source)
	{
		if( __diagnostics_off )
		{
			return;
		}

		if( __current_level >= __file_level || __current_level >= __memory_level )
		{
			println(data,source,__current_level);
		}
	}
	
	/**
	 * Receive data including exception & pass it out
	 */
	public static void println( Object data, Throwable exception, Object source, int level )
	{
		if( __diagnostics_off )
		{
			return;
		}

		if( exception != null && data != null )
		{
			// Convert exception to a String
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			PrintWriter pw = new PrintWriter( baos );
			exception.printStackTrace( pw );
			pw.close();
			
			String ex = new String( baos.toByteArray() );
			
			// Print
			println( data.toString() + "\n" + ex, source, level );
		}
		else if( data != null )
		{
			// Print
			println( data + "\nNo exception given", source, level );
		}
	}
	
	/**
	 * Receive data  and output it to any file set and to
	 * any process [data sink] that is attached
	 * @param data Object the data received
	 * @param source Object the sender of the data
	 * @param level int  the diagnostics level for this data
	 */
	public static void println(Object data, Object source, int newLevel)
	{
		if( __diagnostics_off )
		{
			return;
		}

		UTCTime time = new UTCTime();
		
		if( data != null )
		{
			Object in[] = new Object[5];
			in[0] = data;
			in[1] = source;
			in[2] = new Integer( newLevel );
			if( Thread.currentThread() instanceof ThreadPool.TPThread )
			{
				in[3] = ( (ThreadPool.TPThread) Thread.currentThread() ).getTPThreadGroup().getName();
			}
			else
			{
				in[3] = Thread.currentThread().getThreadGroup().getName();
			}
			in[4] = time.toString();
			
			synchronized( __diag_obj._lock )
			{
				if( newLevel >= __memory_level )
				{
					__diag_obj._memory_queue.addItem( (Object)in );
				}

				if( newLevel >= __file_level )
				{
					__diag_obj._file_queue.addItem( (Object)in );
				}
			}
		}
	}
	
	
	/** output all diagnostic messages for name
	 * @param name String the instance or class name
	 * @param msg String the message to be output
	 */
	public static void printStackTrace( String name, String msg)
	{
		if( __diagnostics_off )
		{
			return;
		}

		println(name+": "+msg);
		//TO BE DONE: read all messages from file at current time, one line at a time
		//if message begins with classname, store in memory
		//At EOF, display all messages
		
	}
	
	/**
	 * set the levelfor following messages
	 * Creation date: (3/3/00 4:08:27 AM)
	 * @param NewLevel int
	 */
	public void set_current_level(int NewLevel)
	{
		__current_level = NewLevel;
	}
	
	/**
	 * set a delimiter character for seperating fields in output messages
	 * @param delim String
	 */
	public static  void setDelimiter(String delim)
	{
	}
	
	/**
	 * set a new level to filter which messages are stored on disk
	 * Creation date: (3/3/00 3:01:08 AM)
	 * @param aLevel int
	 */
	public static void setFileLevel(int alevel)
	{
		if (alevel >= LEVEL_MIN && alevel <= LEVEL_MAX)
		{
			__file_level = alevel;
		}
	}
	
	/**
	 * set a new level to filter which messages are sent to any attached sinks
	 * level must be in the range LEVEL_MIN to LEVEL_MAX otherwise it is ignored
	 * Creation date: (3/3/00 3:00:47 AM)
	 * @param alevel int
	 */
	public static void setMemoryLevel(int alevel)
	{
		if (alevel >= LEVEL_MIN && alevel <= LEVEL_MAX)
		{
			__memory_level = alevel;
		}
	}
	
	/** Notify user of method call which is not yet implemented - it's To Be Done
	 * Deprecate this method?
	 * @param msg String the message to ouput
	 * @param level int
	 */
	public static void TBD( String msg)
	{
		if( __diagnostics_off )
		{
			return;
		}

		println("TO BE DONE: ",msg);
	}
	
//------------------------------------------------------------------------------
// INSTANCE METHODS
//------------------------------------------------------------------------------
	
	/**
	 * Print the given information to the screen
	 */
	public synchronized void monitorUpdateMemory( Object o )
	{
		Object in[] = (Object[]) o;
		Object data = in[0];
		Object source = in[1];
		int newLevel = ((Integer) in[2]).intValue();
		
		String sourceType = "N/A";
		if (source != null)
		{
			sourceType = source.getClass().getName();
		}
		
		StringBuffer buff = new StringBuffer();

		// \r added for UNIX system output
		buff.append( "\r\n" );
		buff.append( in[3] );
		buff.append( DELIMITER );
		buff.append( sourceType );
		buff.append( DELIMITER );
		buff.append( newLevel );
		buff.append( ":\t" );
		buff.append( data.toString() );
		
		// if no attached process , send it to stdout
		System.out.println( buff.toString() );
		
		Thread.yield();
	}
	
	
	/**
	 * Print the given information to a file
	 */
	public synchronized void monitorUpdateFile( Object o )
	{
		Object in[] = (Object[]) o;
		Object data = in[0];
		Object source = in[1];
		int newLevel = ((Integer) in[2]).intValue();
		
		String sourceType = "N/A";
		if (source != null)
		{
			sourceType = source.getClass().getName();
		}
		
		StringBuffer buff = new StringBuffer();
		buff.append( in[4] );
		buff.append( DELIMITER );
		buff.append( in[3] );
		buff.append( DELIMITER );
		buff.append( sourceType );
		buff.append( DELIMITER );
		buff.append( newLevel );
		buff.append( ":\n" );
		buff.append( data.toString() );
		buff.append( "\n\n\n" );
		
		try
		{
			if (_out == null)
			{
				openFile( _filename );
			}
			
			if (_out == null)
			{
				// file open may have failed
				return;
			}
			
			_out.write( buff.toString() );
		}
		catch (IOException e)
		{
			System.err.println( "Error when opening DIAGNOSTICS file for output - " + e );
			_out = null;
		}
		
		Thread.yield();
	}	
	
	/**
	 * Flush the contents of the buffers
	 */
	private void flushFile()
	{
		// Pull everything _out of the buffers
		Object next = null;
		
		do
		{
			next = _file_queue.getItemWithoutBlocking();
			if( next != null )
			{
				monitorUpdateFile( next );
			}
		}
		while( next != null );
		
		do
		{
			next = _memory_queue.getItemWithoutBlocking();
			if( next != null )
			{
				monitorUpdateMemory( next );
			}
		}
		while( next != null );
		
		// Flush file
		try
		{
			_out.flush();
		}
		catch (IOException e)
		{
		}
	}
	
	/**
	 * Set a file (only if it's not already been set) to dump the diagnostic output into
	 * N.B. File must be closed at the end by called DIAGNOSTICS.closeFile();
	 * @param file java.lang.String file name
	 */
	public synchronized void openFile(String afile)
	{
		if( afile == null )
		{
			return;
		}
		
		_filename = afile;
		
		if( _out == null )
		{
			try
			{
				_out = new BufferedWriter(new FileWriter(afile));
			}
			catch (IOException e)
			{
				System.err.println( "Error when opening DIAGNOSTICS file for output - " + e );
				_out = null;
			}
		}
		// else dump file already set
	}
	
	/**
	 * Close the dump file.
	 * File must be closed before it's contents can be viewed
	 * This object can continue to be used if closeFile is called, Diagnostic ouput appears
	 * on the console by default.
	 */
	private synchronized void closeFile()
	{
		if ( _out != null )
		{
			try
			{
				_out.flush();
				_out.close();
			}
			catch (IOException e)
			{
				// switch to stdout
			}
		}
		_out = null;
	}
	
	/**
	 * PRIVATE CONSTRUCTOR
	 */
	private DIAGNOSTICS()
	{
		// Create monitors for memory and file queues
		_monitor_memory = new Monitor( _memory_queue, new MonitorListener()
		{
			public void monitorUpdate( Object o )
			{
				monitorUpdateMemory( o );
			}
		}, Thread.MIN_PRIORITY );
		
		_monitor_file = new Monitor( _file_queue, new MonitorListener()
		{
			public void monitorUpdate( Object o )
			{
				monitorUpdateFile( o );
			}
		}, Thread.MIN_PRIORITY );
	}	
} // end class