//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.util;

//----------------------------------------------------------------------

import java.lang.reflect.*;
import java.util.*;

//----------------------------------------------------------------------

/**
 * This class provides dynamic access to methods of classes which may not be
 * available at compile time & run time.
 */
public class DynamicClass
{
	
	/**
	 * Indicates name of method to invoke in order to get hints for a particular
	 * property
	 */
	public static final String GET_HINTS_METHOD = "getDynamicClassHints";
	
//-----------------------------------------------------------------
// INSTANCE VARIABLES
//-----------------------------------------------------------------

	/**
	 * Object to wrap and dynamically invoke methods on
	 */
	private Object _object;
	
	/**
	 * Class representing the wrapped object
	 */
	private Class _class;

	
//-----------------------------------------------------------------
// CONSTRUCTORS
//-----------------------------------------------------------------
	
	/**
	 * Creates an instance of the given class, using the
	 * constructor with the given arguments.  NOTE: Primitive
	 * types (e.g. int, char, double) should be wrapped in their
	 * java.lang wrapper classes (e.g. Integer, Double etc...)
	 *
	 * @param class_name Name of the class to wrap
	 * @param params Argument signature for the constructor to use
	 * @param args Array containing parameters for constructor of class to wrap
	 */
	public DynamicClass( String class_name, Class params[], Object args[] )
		throws ClassNotFoundException, NoSuchMethodException, 
		InstantiationException, DynamicClassException
	{
		// Try to dynamically load the Class definition for the given
		// class
		_class = Class.forName( class_name );	

		// Got the class definition OK
		
		// Try to locate an appropriate constructor
		Constructor constr = null;
		
		try
		{
			constr = _class.getConstructor( params );
		}
		// Also throws NoSuchMethodException!
		catch( SecurityException e )
		{
			throw new NoSuchMethodException( e.toString() );
		}
			
		try
		{
			// Invoke the constructor
			_object = constr.newInstance( args );
		}
		catch( IllegalAccessException e )
		{
			throw new InstantiationException( e.toString() );
		}
		catch( ExceptionInInitializerError e )
		{
			throw new InstantiationException( e.toString() );
		}
		catch( SecurityException e )
		{
			throw new InstantiationException( e.toString() );
		}
		catch( InvocationTargetException e )
		{
			throw new DynamicClassException( e.getTargetException() );	
		}
		catch( Throwable e )
		{
			// Exception from underlying class, not from reflection API's
			throw new DynamicClassException( e );
		}
	}
	
	
	/**
	 * Creates an instance of the given class, using the
	 * default constructor.
	 *
	 * @param class_name Name of the class to wrap
	 * @param args Array containing parameters for constructor of class to wrap
	 */
	public DynamicClass( String class_name )
		throws ClassNotFoundException, NoSuchMethodException, 
		InstantiationException,Exception
	{
		this( class_name, new Class[0], new Object[0] );
	}
	
	/**
	 * Creates a DynamicClass to wrap the given Object
	 *
	 * @param obj Wrap the given object
	 */
	public DynamicClass( Object obj )
	{
		_class = obj.getClass();
		_object = obj;
	}

//-----------------------------------------------------------------
// PUBLIC METHODS
//-----------------------------------------------------------------

	/**
	 * Invokes a method on the wrapper object with the given arguments
	 *
	 * @param method The name of the method to invoke
	 * @param params Classes of arguments making up signature of method to invoke
	 * @param args An array of object to be used as arguments to the
	 *        method being invoked (primitive types should be wrapped
	 *        in appropriate Java classes (e.g. double => Double ).
	 *
	 * @return The result of the operation, null if the method was void
	 *         or an object wrapping the result if the return type is a primitive.
	 */
	public Object invoke( String method, Class params[], Object args[] )
		throws DynamicClassException, NoSuchMethodException
	{
		DIAGNOSTICS.println( "Attempting to invoke " + method + "(), with arguments " +
					Arrays.asList( params ) + " on object of type " + 
					_class.getName(), this, DIAGNOSTICS.LEVEL_2 );
		Object result = null;
		
		// Lookup the method we're after
		Method mthd = _class.getMethod( method, params );
		
		try
		{
			// Invoke the method
			result = mthd.invoke( _object, args );
		}
		catch( IllegalAccessException e )
		{
			throw new NoSuchMethodException( e.toString() );
		}
		catch( IllegalArgumentException e )
		{
			throw new NoSuchMethodException( e.toString() );
		}
		catch( InvocationTargetException e )
		{
			throw new DynamicClassException( e.getTargetException() );
		}
		catch( SecurityException e )
		{
			throw new NoSuchMethodException( e.toString() );
		}
		catch( Throwable e )
		{
			// Exception from underlying class, not from reflection API's
			throw new DynamicClassException( e );
		}
		
		// Return the result]
		return result;
	}
	
	
	/**
	 * Invokes a method on the wrapper object without any arguments
	 *
	 * @param method The name of the method to invoke
	 *
	 * @return The result of the operation, null if the method was void
	 *         or an object wrapping the result if the return type is a primitive.
	 */
	public Object invoke( String method )
		throws DynamicClassException, NoSuchMethodException
	{
		return invoke( method, new Class[0], new Object[0] );
	}
	
		
	/**
	 * Invokes a method on the given object with the given arguments
	 *
	 * @param method The name of the method to invoke
	 * @param args An array of object to be used as arguments to the
	 *        method being invoked (primitive types should be wrapped
	 *        in appropriate Java classes (e.g. double => Double ).
	 * @param params The classes of the arguments of the method to invoke
	 *
	 * @return The result of the operation, null if the method was void
	 *         or an object wrapping the result if the return type is a primitive.
	 */
	public static Object invokeStatic( String class_name, String method, 
		Class params[], Object args[] )
		throws DynamicClassException, NoSuchMethodException, ClassNotFoundException
	{
		DIAGNOSTICS.println( "Attempting to invoke " + method + "(), with arguments " +
					Arrays.asList( params ) + " on object of type " + class_name,
					DIAGNOSTICS.LEVEL_2 );
		
		Object result = null;
		
		// Lookup the method we're after
		Method mthd = Class.forName( class_name ).getMethod( method, params );
		
		try
		{
			// Invoke the method
			result = mthd.invoke( null, args );
		}
		catch( IllegalAccessException e )
		{
			throw new NoSuchMethodException( e.toString() );
		}
		catch( IllegalArgumentException e )
		{
			throw new NoSuchMethodException( e.toString() );
		}
		catch( InvocationTargetException e )
		{
			throw new DynamicClassException( e.getTargetException() );
		}
		catch( SecurityException e )
		{
			throw new NoSuchMethodException( e.toString() );
		}
		catch( Throwable e )
		{
			// Exception from underlying class, not from reflection API's
			throw new DynamicClassException( e );
		}
		
		// Return the result]
		return result;
	}


	/**
	 * Invokes a method on the given object with no arguments
	 *
	 * @param method The name of the method to invoke
	 *
	 * @return The result of the operation, null if the method was void
	 *         or an object wrapping the result if the return type is a primitive.
	 */
	public static Object invokeStatic( String class_name, String method )
		throws DynamicClassException, NoSuchMethodException, ClassNotFoundException
	{
		return invokeStatic( class_name, method, new Class[0], new Object[0] );
	}


	public String toString()
	{
		return _object.toString();	
	}
	
	
	public Object getObject()
	{
		return _object;
	}

	
	/** Instantiate an instance of a particular class
	 * @param class_name Fully qualified name of the class to instantitate
	 * @return An instance of the given class, or null if some error occured whilst
	 * instantiating an instance
	 */
	public static Object createInstance( String class_name )
	{
		Object inst = null;
		
		try
		{
			DynamicClass dc = new DynamicClass( class_name );
			inst = dc.getObject();
		}
		catch( Throwable t )
		{
		}
		
		return inst;
	}
	
	
	/** Instatiates a class, using a constructor with the given parameters
	 * @param class_name Name of the class to instantiate
	 * @param param_types An array of the types of the parameters of the constructor to use
	 * @param params The actual parameters to the constructor
	 * @return The instantiated object, or null if something went wrong
	 */
	public static Object createInstance( String class_name, Class param_types[], Object params[] )
	{
		Object inst = null;
		
		try
		{
			DynamicClass dc = new DynamicClass( class_name, param_types, params );
			inst = dc.getObject();
		}
		catch( Throwable t )
		{
		}
		
		return inst;
	}
	
	
	/**
	 * Attempt to duplicate the object this DynamicClass wrapps using the following
	 * mechanisms in the given order:
	 * <ul>
	 * <li> Object.clone() </li>
	 * <li> Serialization </li>
	 * </ul>
	 * @return A duplicate of the wrapped object, or null if this couldn't be achieved
	 */
	public Object duplicate()
	{
		try
		{
			// Check if the object is cloneable
			if( _object instanceof Cloneable )
			{
				return invoke( "clone" );
			}
		}
		catch( Throwable t )
		{
		}
		
		try
		{
			// Check if the object is serializable
			if( _object instanceof java.io.Serializable )
			{
				return Serialize.duplicate( _object );
			}
		}
		catch( Throwable t )
		{
		}
		
		// Tried everything we know how to do, just return null
		return null;
	}
	
	
	/**
	 * Gets the properties for this object (i.e. matching get/set pairs)
	 *
	 * <p>The List returned contains Property instances, or null if the
	 * Object is immutable
	 */
	public List getProperties()
	{
		// Check that we're not dealing with an immutable object
		PropertyHint obj_hint = getHint( "" );
		if( obj_hint != null && obj_hint instanceof ImmutableHint )
		{
			return null;
		}
		
		// List all methods
		Method methods[] = _class.getMethods();
		
		// Match get/set pairs
		List results = new LinkedList();
		
		// Find get methods, then find matching set methods
		for( int get = 0; get < methods.length; get++ )
		{
			String name = methods[ get ].getName();
			String property = name.substring( 3 );
			Class arg = methods[ get ].getReturnType();
			
			// Check if current method is a "get", has a return type and
			// takes no arguments
			if( name.startsWith( "get" ) && 
					arg != null &&
					methods[ get ].getParameterTypes().length == 0 )
			{
				name = "set" + property;
				
				// Check for matching set method
				try
				{
					Method set_method = _class.getMethod( name, new Class[] { arg } );
					
					// If we reach here, set method exists - add to property list
					
					// See if there is a hint associated with the property
					PropertyHint ph = getHint( property );
					
					// Create Property object to encapsulate info.
					results.add( new Property( property, arg, ph, this ) );
				}
				catch( Throwable t )
				{
					// Didn't find matching set
				}
			}
		}
		
		return results;
	}
	
	/**
	 * Get the hint (if available) for the given property - returns null
	 * if no hint available
	 */
	public PropertyHint getHint( String property )
	{
		Class args[] = new Class[] { String.class };
		Object params[] = new Object[] { property };
		
		try
		{
			return (PropertyHint) invoke( GET_HINTS_METHOD, args, params );
		}
		catch( Throwable t )
		{
			return null;
		}
	}
	
//-----------------------------------------------------------------
// PRIVATE METHODS
//-----------------------------------------------------------------

	/**
	 * Method to return an array containing the Class objects
	 * of the classes within the given Object array
	 */
	private static Class[] getClasses( Object objects[] )
	{
		// Create the Class[]
		Class class_array[] = new Class[ objects.length ];
		
		// Popuate the Class[]
		for( int i = 0; i < objects.length; i++ )
		{
			// Get the class of the current Object, and add it to the Class[]
			class_array[ i ] = objects[ i ].getClass();
		}

		return class_array;
	}

	
//-----------------------------------------------------------------
// INNER-CLASSES
//-----------------------------------------------------------------

	/**
	 * Class to represent a property type
	 */
	public static class Property
	{
		private PropertyHint _hint;
		private String _name;
		private Class _type;
		private DynamicClass _parent;
		
		private Property( String name, Class type, PropertyHint hint, DynamicClass parent )
		{
			_name = name;
			_type = type;
			_hint = hint;
			_parent = parent;
		}
		
		/**
		 * Returns the name of this property
		 */
		public String getName()
		{
			return _name;
		}
		
		/**
		 * The type that this property uses
		 */
		public Class getType()
		{
			return _type;
		}
		
		/**
		 * Returns a Hint object giving hints about the values
		 * this property can take
		 */
		public PropertyHint getHint()
		{
			return _hint;
		}
		
		/**
		 * Gets the value of this property from the object from which
		 * this was generated
		 */
		public Object getValue()
		{
			try
			{
				return _parent.invoke( "get" + _name );
			}
			catch( Throwable t )
			{
				return null;
			}
		}
		
		/**
		 * Sets the value of this property to the object from which
		 * this was generated
		 */
		public void setValue( Object o )
		{
			try
			{
				Class arg_types[] = { _type };
				Object params[] = { o };
				_parent.invoke( "set" + _name, arg_types, params );
			}
			catch( Throwable t )
			{
			}
		}
		
		public String toString()
		{
			return "[Property=" + _name + ",Type=" + _type.getName() + ",Hint=" + _hint + "]";
		}
	}
	
	
	/**
	 * Umbrella class PropertyHint that all property-hint types extend
	 */
	public static class PropertyHint
	{
	}
	
	/**
	 * Hint for a collection type - details Classes than it can contain
	 */
	public static class CollectionHint extends PropertyHint
	{
		public Class _types[];
		
		public CollectionHint( Class types[] )
		{
			_types = types;
		}
		
		public String toString()
		{
			String out = "Possible types = " + Arrays.asList( _types );
			return out;
		}
	}
	
	/**
	 * Hint for a property that should take only a constrained set of values
	 */
	public static class ConstrainedSelectionHint extends PropertyHint
	{
		public Object _selections[];
		public String _selection_texts[];
		
		public ConstrainedSelectionHint( Object selections[], String strings[] )
		{
			_selections = selections;
			_selection_texts = strings;
		}
		
		public String toString()
		{
			String out = "Constrained values = " + Arrays.asList( _selection_texts );
			return out;
		}
	}
	
	/**
	 * Hint for a property that COULD take a predefined set of values, or
	 * something else of the specified type
	 */
	public static class PossibleSelectionHint extends PropertyHint
	{
		public Class _types[];
		public Object _selections[];
		
		public PossibleSelectionHint( Class types[], Object selections[] )
		{
			_types = types;
			_selections = selections;
		}
		
		public String toString()
		{
			String out = "Possible values = " + Arrays.asList( _selections );
			out += "; Possible types = " + Arrays.asList( _types );
			return out;
		}
	}
	
	/**
	 * Hint that a CLASS is immutable (i.e. use constructor with String as an argument
	 * and toString() to get/set values)
	 */
	public static class ImmutableHint extends PropertyHint
	{
		public String toString()
		{
			String out = "Immutable";
			return out;
		}
	}
		
//-----------------------------------------------------------------
// TEST HARNESS
//-----------------------------------------------------------------

	public static void main( String args[] ) throws Throwable
	{
		Object params[] = { "20020101T1200Z" };
		Class param_types[] = { String.class };
		
		// Try Dynamically using the UTCTime class
		DynamicClass utctime1 = new DynamicClass( "fipaos.util.UTCTime" );
		DynamicClass utctime2 = new DynamicClass( "fipaos.util.UTCTime", param_types, params );
		
		
		// Invoke the toString() method
		System.out.println("utctime1 = " + utctime1 );
		System.out.println("utctime2 = " + utctime2 );
		
		
		// Invoke other methods
		System.out.println("utctime1.invoke(\"getTimeInMillis\") = " +
			utctime1.invoke( "getTimeInMillis" ) );
		System.out.println("utctime2.invoke(\"getTimeInMillis\") = " +
			utctime2.invoke( "getTimeInMillis" ) );
		
		
		// Indicate properties for AgentID
		DynamicClass aid = new DynamicClass( "fipaos.ont.fipa.ACL" );
		List properties = aid.getProperties();
		for( int i = 0; i < properties.size(); i++ )
		{
			Property prop = (Property) properties.get( i );
			System.out.println( prop );
		}
	}
}