//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.util;

// -------------------------------------------------------------------------

import java.util.*;
import java.io.Serializable;

// -------------------------------------------------------------------------

/**
 * FlagHandler provides an easy method of using bitwise flags to store state, minimising
 * the use of multiple booleans for state management. FlagHandler is thread safe - no
 * race conditions can occur when accessing the same FlagHandler from multiple threads.
 *
 *
 **/
public class FlagHandler implements Serializable
{

//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------	
	
	private BitSet _bits;
	private List _names;
	private Hashtable _mx;


//----------------------------------------------------------------------
// CONSTRUCTOR(S)
//----------------------------------------------------------------------
	
	/**
	 * Constructs an empty FlagHandler - there are intially no flags set or known about.
	 * Add new flags using addFlag(), and then set, clear and test them using the appropriate methods.
	 **/
	public FlagHandler() 
	{
		
	  _names = new ArrayList();
	  _mx = new Hashtable();
	  _bits = new BitSet();
		
	}


//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------
	
	/**
	 * Add a new flag to the handler. This will not set the flag, it will simply make the handler
	 * aware of the flag, and assign the flag a bit position. Once added, a flag cannot be removed.
	 *
	 * @param flag The flag name <b>as a String</b>. Flags are always referenced by string to ease
	 *             programmer effort. 
	 * @return True if the flag was added, false if the flag name is already in use or an error occurs
	 **/
	public synchronized boolean addFlag( String flag )
	{
		
		if ( _names.contains( flag ) )
		{
			
			return false;
			
		}
		else
		{
			
			_names.add( flag );
			
		}

	  return true;
		
	}
	
//----------------------------------------------------------------------
	
	/**
	 * Add a new flag to the handler. This will not set the flag, it will simply make the handler
	 * aware of the flag, and assign the flag a bit position. Once added, a flag cannot be removed.
	 * <br>This method also sets the flag to be mutually exclusive with the second flag specified.
	 * This means that if this flag is set, then the mutually exclusive flag will be cleared
	 * automatically, and vice versa. More mutual exclusions can be set using the setMutualExclusions()
	 * method. The semantics of the mutual exclusion mean that clearing a flag will not set any mutually
	 * exclusive flags, even though setting a flag clears the mutually exclusive flags.
	 *
	 * @param flag The flag name <b>as a String</b>. Flags are always referenced by string to ease
	 *             programmer effort. 
	 * @param mxflag Another flag that this flag is mutually exclusive with
	 * @return True if the flag was added, false if the flag name is already in use or an error occurs
	 **/
	public synchronized boolean addFlag( String flag, String mxflag )
	{
		
		if ( _names.contains( flag ) )
		{
			
			return false;
			
		}
		else
		{
			
			MXFlags mx = new MXFlags();
			mx.addMXFlag( mxflag );
			_mx.put( flag, mx );
			_names.add( flag );
			
		}

	  return true;		
		
	}				
	
//----------------------------------------------------------------------
	
	/**
	 * Add a new flag to the handler. This will not set the flag, it will simply make the handler
	 * aware of the flag, and assign the flag a bit position. Once added, a flag cannot be removed.
	 * <br>This method also sets the flag to be mutually exclusive with the array of flags specified.
	 * This means that if this flag is set, then the mutually exclusive flags will be cleared
	 * automatically, and vice versa. More mutual exclusions can be set using the setMutualExclusions()
	 * method. The semantics of the mutual exclusion mean that clearing a flag will not set any mutually
	 * exclusive flags, even though setting a flag clears the mutually exclusive flags.
	 *
	 * @param flag The flag name <b>as a String</b>. Flags are always referenced by string to ease
	 *             programmer effort. 
	 * @param mxflags An array of mutually exclusive flags
	 * @return True if the flag was added, false if the flag name is already in use or an error occurs
	 **/
	public synchronized boolean addFlag( String flag, String[] mxflags )
	{
		
		if ( _names.contains( flag ) )
		{
			
			return false;
			
		}
		else
		{
			
			MXFlags mx = new MXFlags();
			
		  for ( int i=0; i<mxflags.length; i++ )
			{
	  
	  		mx.addMXFlag( mxflags[ i ] );
	  	
			}
			
			_mx.put( flag, mx );
			_names.add( flag );
			
		}

	  return true;	
		
	}			

//----------------------------------------------------------------------
	
	/**
	 * Sets a flag to true, and sets any mutually exclusive flags to false.
	 *
	 * @param flag The flag name <b>as a String</b>. Flags are always referenced by string to ease
	 *             programmer effort. 
	 **/	
	public synchronized void setFlag( String flag )
	{
		
		if ( _names.contains( flag ) )
		{
		
	    int index = _names.indexOf( flag );
	    _bits.set( index );
	       
	    // clear mutually exclusive flags if needed
	    if ( _mx.containsKey( flag ) )
	    {
	    	
		    String[] mxflags = ( (MXFlags)_mx.get( flag ) ).getMXFlags();
		    
		    for ( int i=0; i<mxflags.length; i++ )
				{
					
					clearFlag( mxflags[ i ] );	
					
				}
				
	    }
		  
		}
		
	}

//----------------------------------------------------------------------
	
	/**
	 * Clears a flag (set it to false).
	 *
	 * @param flag The flag name <b>as a String</b>. Flags are always referenced by string to ease
	 *             programmer effort. 
	 **/	
	public synchronized void clearFlag( String flag )
	{
		
		if ( _names.contains( flag ) )
		{
		
		  int index = _names.indexOf( flag );
		  _bits.clear( index );
		  
		}
		
	}

//----------------------------------------------------------------------
	
	/**
	 * Tests the value of the flag.
	 *
	 * @param flag The flag name <b>as a String</b>. Flags are always referenced by string to ease
	 *             programmer effort. 
	 * @return True is the flag is set, false if it not or is not present
	 **/	
	public synchronized boolean testFlag( String flag )
	{
		
		if ( _names.contains( flag ) )
		{
		
		  int index = _names.indexOf( flag );
		  
		  return _bits.get( index );
		  
		}
		
		return false;
		
	}		
	
//----------------------------------------------------------------------
	
	/**
	 * Toggles the value of the flag, clearing any mutually exclusive flags if this flag is set.
	 * <br><b>Note:</b> If this flag is cleared then the mutually exclusive flags are not set.
	 *
	 * @param flag The flag name <b>as a String</b>. Flags are always referenced by string to ease
	 *             programmer effort. 
	 **/	
	public synchronized void toggleFlag( String flag )
	{
		
		if ( _names.contains( flag ) )
		{
			
			if ( testFlag( flag ) )
			{
				
			  clearFlag( flag );	
				
			}
			else
			{
				
				setFlag( flag );
				
			}
			
		}
		
	}		
	
//----------------------------------------------------------------------
	
	/**
	 * Checks to see if this handler contains the specified flag.
	 *
	 * @param flag The flag name <b>as a String</b>. Flags are always referenced by string to ease
	 *             programmer effort. 
	 * @return True if the flag is in this handler, false otherwise
	 **/	
	public boolean containsFlag( String flag )
	{
		
		return _names.contains( flag );
		
	}						
	
//----------------------------------------------------------------------
	
	/**
	 * Sets all of the flags to false (clears every flag);
	 **/	
	public synchronized void clearAllFlags()
	{
		
		Iterator flags = _names.iterator();
		
		while ( flags.hasNext() )
		{
		
			clearFlag( (String)flags.next() );  	
			
		}
		
	}		
	
//----------------------------------------------------------------------
	
	/**
	 * Returns the number of flags currently being handled. Useful if you want to calculate the 
	 * number of flags that can still be put into this handler.
	 *
	 * @return The number of flags currently in use by this handler
	 **/	
	public int getNumberOfFlags()
	{
		
		return _names.size();
		
	}	
	
//----------------------------------------------------------------------
	
	/**
	 * Adds a set of mutually exclusive flag markers to the specified flag. These will automatically
	 * be cleared when this flag is set. If this flag is cleared then these flags are <b>not</b>
	 * automatically set.
	 *
	 * @param flag The flag to add to 
	 * @param mxflags The mutually exclusive flag names
	 **/	
	public synchronized void setMutualExclusions( String flag, String[] mxflags )
	{
		
		if ( _names.contains( flag ) )
		{
		
			MXFlags mx = (MXFlags)_mx.get( flag );
			
			for ( int i=0; i<mxflags.length; i++ )
			{
				
			  mx.addMXFlag( mxflags[ i ] );		  	
				
			}
			
		}
		
	}	
		
//----------------------------------------------------------------------

	/**
	 * Returns a string based representation of the status of every flag in the handler.
	 * <br><b>Note:</b> The list of flags returned is <b>not</b> sorted.
	 *
	 * @return A string representation of the status of the flags
	 **/	
	public String toString()
	{
		
		Iterator flags = _names.iterator();
		StringBuffer output = new StringBuffer();
		String flag;
		
		while ( flags.hasNext() )
		{
		
			flag = (String)flags.next();
		
			output.append( "Flag " );
			output.append( flag );
			output.append( " is " );
			
			if ( testFlag( flag ) )
			{
				
				output.append( "TRUE\n" );
					
			}
			else
			{
				
				output.append( "FALSE\n" );
				
			}
			
		}
		
		return output.toString();
		
	}	
	

//----------------------------------------------------------------------
// INNER CLASSES
//----------------------------------------------------------------------

  /**
   * Inner class used to store mutual exclusion information for each flag.
   **/
  private class MXFlags implements Serializable
	{
		
		// The mutual exclusion store
		private List _mx;
		
//----------------------------------------------------------------------		
		
		/**
		 * Constructs a new flag mutual exclusion list.
		 */
	  private MXFlags()
		{
			
			_mx = new ArrayList();
			
		}	
				
//----------------------------------------------------------------------		
		
		/**
		 * Adds a mutual exclusive flag to this flag
		 *
		 * @param mxflag The mutually exclusive flag name
		 **/
		private void addMXFlag( String mxflag )
		{
			
			if ( ! _mx.contains( mxflag ) )
			{
			
			  _mx.add( mxflag );	
				
			}
			
		}
		
//----------------------------------------------------------------------		
		
		/**
		 * Gets all of the mutually exclusive flags.
		 *
		 * @return An array of flag names
		 **/		
		private String[] getMXFlags()
		{
						
			return (String[])_mx.toArray( new String[ 0 ] );
			
		}
		
		
	} // End inner class
	
	
//----------------------------------------------------------------------
// TEST HARNESS
//----------------------------------------------------------------------	
			
	/* public static void main( String[] args )
	{
		
		FlagHandler fh = new FlagHandler();


		
		System.out.println( fh.toString() );
		
	} */
	

//----------------------------------------------------------------------
// TEST HARNESS
//----------------------------------------------------------------------	
			
	/* public static void main( String[] args )
	{
		
		FlagHandler fh = new FlagHandler();


		
		System.out.println( fh.toString() );
		
	} */
	

} // End FlagHandler