//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.util;

//----------------------------------------------------------------------

import java.util.Vector;
import java.util.Enumeration;
import fipaos.util.Queue;

/**
 * This class provides a generic mechanism for forwarding objects
 * from one thread to another in order to decouple parts of a system.
 * A Queue must be passed to a Monitor upon construction, and it monitors
 * this for objects to forward.  When there is an object in the queue
 * it is passed to the MonitorListener registered with the Monitor
 * <br>
 * <br> 
 **/
public class Monitor extends Thread
{
	/**
	 * Indicates if we have been asked to stop
	 */
	private boolean _stop = false;
	
	/**
	 * The queue being used to store the objects
	 */
	private Queue _queue;

	/**
	 * The target MonitorListener
	 */
	protected MonitorListener _target;

	/**
	 * Constructor - starts threads running
	 *
	 * @param queue The queue that should be used for buffering waiting
	 *              objects
	 * @param ml The MonitorListener to forward objects to
	 */	
	public Monitor( Queue queue, MonitorListener ml )
	{
		this( queue, ml, Thread.NORM_PRIORITY );
	}
	
	/**
	 * Constructor - starts threads running
	 *
	 * @param queue The queue that should be used for buffering waiting
	 *              objects
	 * @param ml The MonitorListener to forward objects to
	 */	
	public Monitor( Queue queue, MonitorListener ml, int priority )
	{
		super( "fipaos.util.Monitor" );
		_queue = queue;
		_target = ml;

		setPriority( priority );
		start();
	}


	/**
	 * Run method - continuously checks for
	 * incomming messages, and passes them on to
	 * the MonitorListener implementations associated 
	 * with this
	 */
	public synchronized void run()
	{
		int current = 0;
		
		while( !_stop )
		{
			// Get the next object in the queue
			Object next = _queue.getItem();
			
			// If something has been returned, pass it to all
			// registered objects
			if( next != null )
			{
				// Forward the object
				_target.monitorUpdate( next );
				
				// Remove reference to object
				next = null;
				
				// Yield to allow other threads to execute
				yield();
			}
		}
	}
	
	
	/**
	 * Invoked when this should exit cleanly
	 */
	public void halt()
	{
		_stop = true;
		_queue.releaseBlock();
		interrupt();
	}


	/**
	 * Test harness for Monitor class
	 */
	public static void main( String args[] )
	{
		int messages = 50;
		
		// Display test parameters
		System.out.println("Test parameters\n===============");
		System.out.println("Messages to be sent = " + messages);
		
		// Setup the test
		
		// Create the queue & monitor
		Queue q = new FIFOQueue();
		Monitor monitor = new Monitor( q, new MonitorListenerTest( 0 ) );

		// Display progress
		System.out.println("Initialised test, starting now");
		
		// Send the messages
		for( int i = 0; i < messages; i++ )
		{
			// Send the message
			String m = "Msg. No. " + i;
			System.out.println("Sending \"" + m.toString() + "\"");
			q.addItem( m );
		}

		// Display progress
		System.out.println("Test done");
	}


	// Inner class for testing
	protected static class MonitorListenerTest implements MonitorListener
	{
		int n;
		
		public MonitorListenerTest( int n )
		{
			this.n = n;	
		}
		
		public void monitorUpdate( Object o )
		{
			System.out.println( "MonitorListener " + n + " - " + o.toString() );	
		}
	}
}