//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.util;

//----------------------------------------------------------------------

import java.util.Vector;
import java.util.Enumeration;
import fipaos.util.Queue;

/**
 * This is an extension of the Monitor object.  When an object is placed into the queue
 * it is passed to all of the MonitorListener registered with this MultiThreadedMonitor.
 *
 * WARNING:  There is no guarantee that objects will arrive at the MonitorListener's in
 * the order they are placed into the Queue this MultiThreadedMonitor is observing. It is
 * possible that more than one thread will invoke the monitorUpdate() method at a time,
 * so the implementation should be thread-safe.
 *
 * @see Monitor fipaos.util.Monitor
 * <br>
 * <br> 
 **/
public class MultiThreadedMonitor extends Monitor implements MonitorListener
{
	/**
	 * The Vector containing the targets of the objects
	 */
	private Vector _receivers = new Vector();

	/**
	 * Array of Monitors that belong to this MultiThreadedMonitor
	 */
	private Monitor _monitors[];

	/**
	 * Array of queues associated with Monitors
	 */
	private Queue _queues[];

	/**
	 * Next thread to receive job
	 */
	private int _next = 0;


	/**
	 * Constructor - starts threads running
	 *
	 * @param queue The queue that should be used for buffering waiting
	 *              objects
	 * @param threads The number of threads to be used (NOTE: These may
	 *                push objects to the MonitorListener's registered
	 *                concurrenty.
	 */	
	public MultiThreadedMonitor( Queue queue, int threads )
	{
		super( queue, null );
		_target = this;

		// Create the appropriate arrays
		_monitors = new Monitor[ threads ];
		_queues = new Queue[ threads ];
		
		// Create the Monitor to deal with our requests
		for( int i = 0; i < threads; i++ )
		{
			// Create target queue
			Queue q = new FIFOQueue();
			_queues[ i ] = q;
			
			// Add anonymous MonitorListener to deal with our objects
			_monitors[ i ] = new Monitor( q, new MonitorListener()
			{
				public void monitorUpdate( Object o )
				{
					Object items[] = (Object []) o;
					
					// Pull out the object to pass
					Object item = items[0];

					// Pull out the target
					MonitorListener target = (MonitorListener) items[1];
					
					// Push the item
					target.monitorUpdate( item );
				}	
			} );
		}
	}


	/**
	 * Deals with incoming objects by dispatching them to our Monitors
	 */
	public synchronized void monitorUpdate( Object o )
	{
		// Iterate through all of the receivers
		Enumeration enum = _receivers.elements();
		while( enum.hasMoreElements() )
		{
			// Pack the object and dispatch it	
			Object items[] = { o, enum.nextElement() };
			_queues[ _next ].addItem( items );
			
			// Increment the next count
			_next = ( _next + 1 ) % _queues.length;
		}
	}
		
		
	/**
	 * Adds a MonitorListener to this Monitor
	 *
	 * @param The MonitorListener to add
	 */
	public synchronized void addMonitorListener( MonitorListener ml )
	{
		_receivers.addElement( ml );
	}
	

	/**
	 * Removes a MonitorListener from this Monitor
	 *
	 * @param The MonitorListener to remove
	 */
	public synchronized void removeMonitorListener( MonitorListener ml )
	{
		_receivers.remove( ml );
	}


	/**
	 * Invoked when this should exit cleanly
	 */
	public void halt()
	{
		super.halt();
		
		// Kill the Monitor's
		for( int i = 0; i < _monitors.length; i++ )
		{
			_monitors[ i ].halt();
		}
	}


	/**
	 * Test harness for Monitor class
	 */
	public static void main( String args[] )
	{
		int threads = 10;
		int messages = 50;
		int listeners = 5;
		
		// Display test parameters
		System.out.println("Test parameters\n===============");
		System.out.println("MonitorThread's in use = " + threads);
		System.out.println("Messages to be sent = " + messages);
		System.out.println("MonitorListener's registered = " + listeners);
		
		// Setup the test
		
		// Create the queue & monitor
		Queue q = new FIFOQueue();
		MultiThreadedMonitor monitor = new MultiThreadedMonitor( q, threads );

		// Create and add the MonitorListener's
		for( int i = 0; i < listeners; i++ )
		{
			monitor.addMonitorListener( new MonitorListenerTest(i) );
		}
		
		// Display progress
		System.out.println("Initialised test, starting now");
		
		// Send the messages
		for( int i = 0; i < messages; i++ )
		{
			// Send the message
			String m = "Msg. No. " + i;
			System.out.println("Sending \"" + m.toString() + "\"");
			q.addItem( m );
		}

		// Display progress
		System.out.println("Test done");
	}
}