//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.util;

//----------------------------------------------------------------------

import java.util.Vector;
import java.lang.reflect.Constructor;

/**
 * This is a simple class which defines methods which all queues
 * should implement
 * <br>
 * <br> 
 **/

public abstract class Queue
{
	/**
	 * Vector containing the message queue
	 */
	protected Vector _queue = new Vector();
	

	/**
	 * Maximum size the queue should reach
	 */
	private int _max;


	/**
	 * Indicates that threads should not block in the getItem() method
	 */
	private boolean _block = true;
	
	
	/**
	 * Stores information such as maximum size of the queue
	 *
	 * @param max_size The most elements that can be contained within the
	 *                 queue at once, or -1 for no limit
	 */
	protected Queue( int max_size )
	{
		_max = max_size;	
	}
	
	
	/**
	*  Returns the number of items currently in the queue.
	*/
	public int queueLength()
	{
	   return _queue.size();
	}


	/**
	 * Adds an object to the queue
	 *
	 * @param obj The object to add, given that the queue is not full
	 */
	public synchronized void addItem( Object obj )
	{
		if( !isFull() )
		{
			_queue.addElement( obj);
			notify();
		}
	}
	

	/**
	 * Get the next message from the queue.  If
	 * the queue is empty, the thread will block
	 * until a message is added.
	 *
	 * @return The next message in the queue, or null
	 *					if the thread is interrupted
	 */
	public synchronized Object getItem()
	{
		while( _queue.isEmpty() && _block )
		{
			try
			{
				wait();	
			}
			catch( InterruptedException e )
			{
			}
		}
		
		return getNextItem();
	}


	/**
	 * Get the next message from the queue.  If
	 * the queue is empty, return null immediately.
	 *
	 * @return The next message in the queue or null
	 */
	public synchronized Object getItemWithoutBlocking()
	{
		return getNextItem();
	}
	
	
	/**
	 * Method to determine if the queue is full
	 *
	 * @return True if the queue is at full capacity
	 */
	public synchronized boolean isFull()
	{
		return ( _max != -1 && _queue.size() >= _max );
	}


	/**
	 * Method to determine if the queue is empty
	 *
	 * @return True if the queue is empty
	 */
	public synchronized boolean isEmpty()
	{
		return _queue.isEmpty();
	}


	/**
	 * Method to flush the contents of the queue
	 */
	public synchronized void flush()
	{
		_queue.clear();	
	}

	
	/**
	 * Method to release all blocked threads within the queue - once
	 * invoked, getItem() will never block again
	 */
	public synchronized void releaseBlock()
	{
	  _block = false;
	  notifyAll();
	}
	

	/**
	 * Method to get the next element from the queue
	 * It is down to the queue implementation to determine
	 * which item in the queue to retreive
	 *
	 * @return The next item from the queue
	 */
	protected abstract Object getNextItem();
	
	
	/**
	 * Test harness for the various Queue implementations
	 */
	public static void main( String args[] ) throws Throwable
	{
		// Classes to test
		String classes[] = { "fipaos.util.FIFOQueue", "fipaos.util.LIFOQueue" };
		
		// Constructor to test
		Class param_classes[] = { Integer.TYPE };
		
		// ....and parameters
		Object params[] = { new Integer( 10 ) };

		// Display the parameters to be used
		System.out.print("Using these parameters: ");
		for( int i = 0; i < params.length; i++ )
		{
			System.out.print( params[ i ] + " " );
		}
		System.out.print("\n");
		
		// Iterate through each class to test
		for( int i = 0; i < classes.length; i++ )
		{
			// Get the current class
			Class curr = Class.forName( classes[ i ] );
			
			// Find the relevant constructor	
			Constructor c = curr.getConstructor( param_classes );
			
			// Get a new instance
			Queue q = (Queue) c.newInstance( params );
			
			
			// Now for the tests
			System.out.println("Testing " + classes[ i ]);
			System.out.print("Input to queue = ");
			
			// Fill up the queue with the numbers 1 to 20
			for( int x = 1; x <= 20; x++ )
			{
				System.out.print( x + "; " );
				q.addItem( new Integer( x ) );	
			}
			
			// Display the contents of the queue
			System.out.print("\n\nOutput of queue = ");
			
			while( !q.isEmpty() )
			{
				System.out.print( q.getItem() + "; " );
			}

			System.out.print("\n\n");
		}
	}
}