//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.util;

//----------------------------------------------------------------------

import java.io.*;

//----------------------------------------------------------------------

/**
 * This class provides some simple Serialization primitives
 * <br>
 * <br>
 */
public class Serialize
{
	 
//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------	

	/** Serialises the given object into a byte[]
	 * @param o Object to serialise
	 * @return Byte[] containing the serialised object, or null
	 * if the object couldn't be serialised
	 */
	public static byte[] toByteArray( Object o )
	{
		byte out[] = null;
		
		try
		{
			// Create BAOS to collect stream data
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			
			// Create OOS to serialize the object
			ObjectOutputStream oos = new ObjectOutputStream( baos );
			
			// Write the object
			oos.writeObject( o );
			
			// Close & flush streams
			oos.flush();
			baos.flush();
			oos.close();
			baos.close();
			
			// Get byte[] from BAOS
			out = baos.toByteArray();
		}
		catch( Throwable t )
		{
			// Exception occured - ignore (result will be null)
		}
		
		return out;
	}
	
	
	/** Reconstructs a serialised object from a byte[] to an Object
	 * @param in The byte[] containing a serialised object
	 * @return The Object contained in the byte array, or null
	 * if it couldn't be de-serialized
	 */
	public static Object toObject( byte in[] )
	{
		Object out = null;
		
		try
		{
			// Create BAIS to provide stream data
			ByteArrayInputStream bais = new ByteArrayInputStream( in );
			
			// Create OIS to de-serialize the object
			ObjectInputStream ois = new ObjectInputStream( bais );
			
			// Read the object
			out = ois.readObject();
			
			// Close streams
			ois.close();
			bais.close();
		}
		catch( Throwable t )
		{
			// Exception occured - ignore (result will be null)
		}
		
		return out;
	}
	
	
	/** <p>Duplicates the given object using serialization - completely decouples
	 * both copies.  
	 * <p>WARNING: This is potentially an order of magnitude slower than manually
	 * cloning an object
	 * @param in Object to duplicate
	 * @return Duplicate object, or null if an error occured
	 */
	public static Object duplicate( Object in )
	{
		return toObject( toByteArray( in ) );
	}
	
	
	/**
	 * <p>Attempts to save the given Serializable object to the given file
	 */
	public static void objectToFile( Serializable in, File out ) throws IOException
	{
		// Create appropriate readers
		FileOutputStream fos = new FileOutputStream( out );
		ObjectOutputStream oos = new ObjectOutputStream( fos );
		
		oos.writeObject( in );
		
		oos.flush();
		oos.close();
		fos.flush();
		fos.close();
	}
	
	/**
	 * Attempts to read an Object from the given file (assume Object is serializable)
	 */
	public static Object fileToObject( File file ) throws IOException, ClassNotFoundException
	{
		// Create appropriate readers
		FileInputStream fis = new FileInputStream( file );
		ObjectInputStream ois = new ObjectInputStream( fis );
		
		Object out = ois.readObject();
		
		ois.close();
		fis.close();
		
		return out;
	}
	
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------

	public static void main( String args[] )
	{
		final int TIMES = 10000;
		
		String test = "Hello World";
		
		long str_copy = System.currentTimeMillis();
		for( int i = 0; i < TIMES; i++ )
		{
			String other = new String( test );
		}
		str_copy = System.currentTimeMillis() - str_copy;
		
		long ser_copy = System.currentTimeMillis();
		for( int i = 0; i < TIMES; i++ )
		{
			String other = (String) duplicate( test );
		}
		ser_copy = System.currentTimeMillis() - ser_copy;
		
		System.out.println( "Time to new String() " + TIMES + " Strings = " + str_copy );
		System.out.println( "Time to duplicate() " + TIMES + " Strings  = " + ser_copy );
	}
} // End class