//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files ("the
//     Programs") are Copyright (c) 1999 - 2000 Nortel Networks Corporation, 
//     8200 Dixie Road, Suite 100, Brampton, Ontario, Canada L6R 5P6.  
//     All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out in
//     the Nortel Networks FIPA-OS Public License, a copy of which can be found
//     in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version can
//     also be found at http://fipa-os.sourceforge.net/. By downloading the
//     files containing the Programs you accept the terms and  conditions of the
//     Public License. You do not have to accept these terms and conditions, but
//     unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.util;

//----------------------------------------------------------------------

import fipaos.ont.fipa.*;
import fipaos.ont.*;
import java.util.*;

//----------------------------------------------------------------------

/**
 * Provides methods to parse strings for ACL message fields. Useful for when you have an ACL message
 * (or bit of one) as a string from which you want one or two fields - saves the overhead of a whole
 * ACL parser.
 *
 */
public class StringParser {

	
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------	
	
	public static final String ACTION_TYPE		= ":action-type";
	public static final String PERFORMATIVE		= ":performative";
	
	
//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------

  /**
  * Parses a string for the given ACL field (any field can be given except :receiver and :envelope)<br>
  * Note: Giving ":performative" as the field name will return the message type.<br>
  * Giving ":action-type" as the field to look for will return the next keyword after an "(action ..."
  * statement in the string. This is useful for ACL content strings. If no "(action ..." is found, null
  * is returned.<br>
  * If any other string is given then the corresponding word or bracketed expression is given depending on the
  * first character found (e.g. searching for :language in ":language SL0" returns "SL0", but searching for
  * :language in "@language (SL0 SL1) will return "(SL0 SL1)" ). This method will not deal with multiple
  * brackets, e.g. searching for :language in ":language (SL0 (SL1 SL2) SL3)" will return "(SL0 (SL1 SL2)".
  * To get the whole bracketed expression call parseForIndex() with :language as the string to look for
  * and then call getBracketedExpression() with the returned index as the argument.<br>
  * If an error occurs or the field cannot be found null is returned.
  *
  * @param look_for ACL field to look for
  * @param to_parse String to look in
  **/
  public static String parseFor( String look_for, String to_parse )
  {
	
	int start, end = 0, alt_end = 0;
	
	if ( look_for.equals( PERFORMATIVE ) )
	{
	  
	  end = to_parse.indexOf( BASETOKENS.SPACE, 1 );
	  start = to_parse.indexOf( BASETOKENS.OPENBRACKET );
	  if ( start < 0 || start > end )
		start = -1;
	  
	}
	else if ( look_for.equals( FIPACONSTANTS.RECEIVER ) 
		     || look_for.equals( FIPACONSTANTS.ENVELOPE ) )
	{
	  
	  return null;
	  
	}
	else if ( look_for.equals( FIPACONSTANTS.CONTENT ) )
	{
	 
	  start = to_parse.indexOf( FIPACONSTANTS.CONTENT );   
	  start = to_parse.indexOf( BASETOKENS.OPENBRACKET, start );
	  int current = start + 1;
	  int bracket_count = 1;
	  String temp = new String();
	  boolean done = false;
		
	  while ( ! done )
	  {
		  
		try 
		{
			  
		  temp = to_parse.substring( current, current + 1 );
			  
		}
		catch ( StringIndexOutOfBoundsException sioobe )
		{
			 
		  done = true;
		  return null;
			  
		}
			
		if ( temp.equals( BASETOKENS.OPENBRACKET ) )
		{
			  
		  bracket_count++;
			  
		}
		else if ( temp.equals( BASETOKENS.CLOSEBRACKET ) )
		{
			  
		  bracket_count--;
			  
		  if ( bracket_count == 0 )
		  {
			   
			end = current;                   
			end++;
			start--;           
			done = true; 
				
		  }
			  
		}
		
		current++;
			
	  } // while
	  
	  temp = null;
	  
	}
	else if ( look_for.equals( ACTION_TYPE ) )
	{
	  
	  start = to_parse.indexOf( BASETOKENS.ACTION );
	  
	  if ( start < 0 )
	  {
	  	
		return null;
		
	  }
	  
	  start = to_parse.indexOf( BASETOKENS.OPENBRACKET, start + 1 );
	  end = to_parse.indexOf( BASETOKENS.SPACE, start + 1 );
	  
	}
	else
	{
	  start = to_parse.indexOf( look_for );

			// If we can't find the string, return null.
			if( start == -1 )
			{
				
				return null;
				
			}

	  start = to_parse.indexOf( BASETOKENS.SPACE, start );
	  end = to_parse.indexOf( BASETOKENS.SPACE, start + 1 );
	  alt_end = to_parse.indexOf( BASETOKENS.CLOSEBRACKET, start + 1 );
	  
	  if ( alt_end < end )
	  {
	  	
		end = alt_end;
		
	  }
			
	  if ( to_parse.substring( start + 1, start + 2 ).equals( BASETOKENS.OPENBRACKET ) )
	  {
	  	
		end = -1;
		
	  }
	  
	  if ( end < 0 ) 
	  {
		
		end = to_parse.indexOf( BASETOKENS.CLOSEBRACKET, start + 1 );
		end++;
		
		if ( end <= 0 ) 
		{
			
		  end = to_parse.length();
		  
		}
		  
	  }
	  else if ( end == start + 1 )
	  {
		
		// ignore leading blanks
		while ( end == start + 1 )
		{
		  
		  start++;
		  end = to_parse.indexOf( BASETOKENS.SPACE, start + 1 );
		  
		}
		
	  }
	  
	}
	
	try 
	{
		
	  return to_parse.substring( start + 1, end );
	  
	}
	catch ( StringIndexOutOfBoundsException sioobe )
	{
	  
	  return null;
	  
	}
	
  } // parseFor()  
  
//----------------------------------------------------------------------  
  
  /**
  * Parses a string for the ACL envelope or receiver field and returns the contents in a vector.
  * Assumes that the fields have bracketed sub-expressions listing the items to retrieve.
  *
  * @param look_for the ACL field to find
  * @param to_parse the string to parse
  **/
  public static Vector parseForVector( String look_for, String to_parse )
  {
	
	int start, end;
	boolean done = false;
	Vector output = new Vector( 1, 1 );
	
	start = to_parse.indexOf( look_for );
	start = to_parse.indexOf( BASETOKENS.OPENBRACKET, start );
	end = to_parse.indexOf( BASETOKENS.CLOSEBRACKET, start + 1 );
	
	if ( look_for.equals( FIPACONSTANTS.ENVELOPE ) )
	{
	
	  // AW specific envelope parsing here
	  try 
	  {
		  
		output.addElement( to_parse.substring( start + 1, end ) );
		
	  }
	  catch ( StringIndexOutOfBoundsException sioobe )
	  {
		
		return null;
		
	  }
	  
	}
	else if ( look_for.equals( FIPACONSTANTS.RECEIVER ) )
	{
	  
	  String temp = to_parse.substring( start + 1, end ); 
	  start = 0;
	  
	  while ( ! done )
	  {
	  
		end = temp.indexOf( BASETOKENS.SPACE, start );
			  
		if ( end > 0  )
		{
				   
		  try
		  {
		  
			output.addElement( temp.substring( start, end ) );
		  
		  }
		  catch ( StringIndexOutOfBoundsException sioobe )
		  {
		  
			return null;
		  
		  }
		  
		}
		else
		{
		  
		  end = temp.length();
		  
		  try
		  {
		  
			output.addElement( temp.substring( start, end ) );
		  
		  }
		  catch ( StringIndexOutOfBoundsException sioobe )
		  {
		  
			return null;
		  
		  }
		  
		  done = true;
		  
		}
		  
		
		start = end + 1;  
	  
	  }
	  
	  temp = null;
	  
	}
	else
	{
		
	  return null;
	  
	}
	
	// remove blank entries  
	for ( int i = 0; i < output.size(); i++ )
	{
	 
	  if ( output.elementAt( i ).equals( BASETOKENS.BLANK ) )
	  {
	  	
		output.removeElementAt( i );
		
	  }
	  
	}
	
	return output;
	
  }           
  
//----------------------------------------------------------------------  
  
  
  
  /** given a list of bracketed expressions as a string this function will return the list with the first expression removed
   * @param str list of bracketed expressions as a list
   * @return the list of bracketed expressions wilt the first one removed
   */
  public static String removeExpressionFromList(String str)
	{
		// Count up and down brackets until we either: a) reach 0 again, or b)
		// reach the end of the String
		StringTokenizer st = new StringTokenizer( str, "()", true );
		
		int level = 0;
		boolean start = true;
		while( st.hasMoreTokens() && ( level != 0 || start ) )
		{
			String next = st.nextToken();
			if( next.equals( BASETOKENS.OPENBRACKET ) )
			{
				level++;
				start = false;
			}
			else if( next.equals( BASETOKENS.CLOSEBRACKET ) )
			{
				level--;
			}
		}
		
		if( st.hasMoreTokens() )
		{
			return st.nextToken( "" );
		}
		else
		{
			return "";
		}
	}
  
  /** When given a list of bracketed expressions as a string and a tag within the expressions, this method will instanses of the expression are in the list
   * @param expressions list of bracketed expressions as a list
   * @param tag Tag for which to search
   * @return number of expressions in the list string
   */
  public static int getNoOfBracketedExpressions( String expressions ) {
	int count = 0;
	String expression = null;
	if( expressions != null && !expressions.equals( "" ) )
	{
	  do
	  {
		expression = removeExpressionFromList( expressions );
		count++;
	  }while( !expression.equals( "" ) );
	}
	
	return count;
  }  
  
  /** Gets the first bracketed expression from the String list
   * @param str String List of bracketed expressions
   * @return the first expression in the list
   */
  public static String getNextBracketedExpression(String str)
	{
	// Count up and down brackets until we either: a) reach 0 again, or b)
	// reach the end of the String
	StringTokenizer st = new StringTokenizer( str, "()", true );

	int level = 0;
	int end = 0;
	String next = null;
	boolean start = true;
	while( st.hasMoreTokens() && ( level != 0 || start ) )
	{
	   next = st.nextToken();
	  if( next.equals( BASETOKENS.OPENBRACKET ) )
	  {
		level++;
		start = false;
	  }
	  else if( next.equals( BASETOKENS.CLOSEBRACKET ) )
	  {
		level--;
	  }
	  end = end + next.length();
	}
	

	return str.substring( 0, end);

  }  
	
	
	/**
	 * Method to provide conversion between a document containing ACL messages
	 * and a List of ACL objects - returns null if anything goes wrong
	 */
	public static List parseACL( String messages )
	{
		List out = new LinkedList();
		
		try
		{
			// Loop until there is nothing left
			while( !messages.trim().equals( "" ) )
			{
				// Get next ACL message
				String msg = getNextBracketedExpression( messages );
				
				// Try to parse
				ACL acl = new ACL( msg );
				
				// Add to list
				out.add( acl );
				
				// Get rid of first ACl message
				messages = removeExpressionFromList( messages );
			}
		}
		catch( Throwable t )
		{
			// Exception occured - just return null
			return null;
		}
		
		return out;
	}
	
	
  /**
  * Parses the given string for the string specified, and returns the index of the character immediately
  * following the end of that string ( e.g. looking for :language in the string ":language SL0" will return
  * the integer 9 - marking the space after the word :language )
  *
  * @param look_for The string to parse for
  * @param to_parse The string to look in
  *
  * @see getBracketedExpression
  **/
  public static int parseForIndex( String look_for, String to_parse )
  {
	
	int start = to_parse.indexOf( look_for );
	start = to_parse.indexOf( BASETOKENS.SPACE, start + 1 );
	
	return start;
	
  }  
  
//----------------------------------------------------------------------  
  
  /**
  * Returns an entire bracketed expression from the given string starting from the index given.
  * Should be used with parseForIndex which will locate the index to the start of the expression.
  *
  * @param start_from The index to start from
  * @param to_parse The string to look in
  * @return The bracketed expression that is marked by index
  **/
  public static String getBracketedExpression( int start_from, String to_parse )
  {
  	
  	return getBracketedExpression( start_from, to_parse, true );	
  	
  }  
  
//----------------------------------------------------------------------  
  
  /**
  * Returns an entire bracketed expression from the given string starting from the index given.
  * Should be used with parseForIndex which will locate the index to the start of the expression.
  *
  * @param start_from The index to start from
  * @param to_parse The string to look in
  * @param keep_brackets True if the output should still have the enclosing brackets, false otherwise
  * @return The bracketed expression that is marked by index
  **/
  public static String getBracketedExpression( int start_from, String to_parse,
  																						 boolean keep_brackets )
  {
	
	int start = to_parse.indexOf( BASETOKENS.OPENBRACKET, start_from );
	int end = 0;
	int current = start + 1;
	int bracket_count = 1;
	String temp = new String();
	boolean done = false;
		
	while ( ! done )
	{
		  
	  try
	  {
			  
		temp = to_parse.substring( current, current + 1 );
			  
	  }
	  catch ( StringIndexOutOfBoundsException sioobe )
	  {
			 
		done = true;
		return null;
			  
	  }
			
	  if ( temp.equals( BASETOKENS.OPENBRACKET ) )
	  {
			  
		bracket_count++;
			  
	  }
	  else if ( temp.equals( BASETOKENS.CLOSEBRACKET ) )
	  {
			  
		bracket_count--;
			  
		if ( bracket_count == 0 )
		{
			   
		  end = current;                   
		  end++;
		  start--;           
		  done = true; 
				
		}
			  
	  }
		
	  current++;
			
	} // while 
	
	try
	{
		
	  temp = to_parse.substring( start + 1, end );
	  
	}
	catch ( StringIndexOutOfBoundsException sioobe )
	{
	  
	  return null;
	  
	}
	
	if ( ! keep_brackets )
	{
		
		start = temp.indexOf( BASETOKENS.OPENBRACKET );
		end = temp.lastIndexOf( BASETOKENS.CLOSEBRACKET );
		temp = temp.substring( start + 1, end );
		temp.trim();
		
	}
	
	return temp;
	
  }  
  
//----------------------------------------------------------------------  
  
  /**
   * Indicates whether the word to search for is actually present in the string to be parsed.
   *
   * @param look_for The word to check for
   * @param to_parse The string to check for keyword in
   * @return True if the word to search for is present, false otherwise
   **/
  public static boolean contains( String look_for, String to_parse )
	{
		
	  return ( to_parse.indexOf( look_for ) >= 0 );
		
	}
  
//----------------------------------------------------------------------	
	
  /**
   * Parses the supplied string for a list of space seperated values marked by the look-for
   * parameter. For example, given a list of addresses in the following format:
   * <br><code> (:address agent1@iiop://47.47.47.47:9000/acc agent2@wap://123456789/acc) </code>
   * <br>you can call this method with the look_for parameter set to ":address" and it will
   * return the two addresses in a vector. The value list must be space seperated and can end with 
   * " )" or ")".
   *
   * @param look_for The field marker to extract values for
   * @param to_parse The content to look in
   * @return A vector of values for the specified marker
   **/
  public static Vector parseForList( String look_for, String to_parse )
	{
		
		int start, space, bracket;
	boolean done = false;
	Vector output = new Vector( 1, 1 );
	
	start = to_parse.indexOf( look_for );
	start = to_parse.indexOf( BASETOKENS.SPACE, start ) + 1;
	
	while ( ! done )
		{
	
	    space = to_parse.indexOf( BASETOKENS.SPACE, start + 1 );	
	    bracket = to_parse.indexOf( BASETOKENS.CLOSEBRACKET, start + 1 );
	    
	    if ( space < bracket - 1 ) // -1 needed to ensure that it isn't " )"
			{
			
			  try 
			  {
				
					output.addElement( to_parse.substring( start, space ) );
					start = space + 1;
					
			  }
			  catch ( StringIndexOutOfBoundsException sioobe )
			  {
			  
			  	done = true;
			  	
			  }
				
			}
			else
			{
				
				try 
				{
				
					output.addElement( to_parse.substring( start, bracket ) );
					done = true;
				
			  }
			  catch ( StringIndexOutOfBoundsException sioobe )
			  {
			  
			  	done = true;	
			  	
			  }				
				
			}
			
		}
		
		return output;
		
	}
	
//----------------------------------------------------------------------	
	
	/**
	 * Strips the outside pair of brackets from a string. If only one odd bracket is present at
	 * the start or the end of the string then this will be removed as well. The string
	 * returned will also have any whitespace at the start and end removed.
	 *
	 * @param content The string to remove the brackets from
	 * @return The string with the brackets removed
	 **/
	public static String trimBrackets( String content )
	{
		
		// Possible inputs:
		// ( content ) <- remove both
		// content ) <- remove odd one at the end
		// ( content <- remove one at front
		// content <- do nothing
		
		content = content.trim(); // Remove spaces at the start and the end
		
		int openindex = content.indexOf( BASETOKENS.OPENBRACKET );
		int closeindex = content.lastIndexOf( BASETOKENS.CLOSEBRACKET );
		
		if ( openindex == 0 && closeindex == ( content.length() - 1 ) )
		{
			
			// brackets at both ends
			content = content.substring( openindex + 1, closeindex );
			
		}
		else if ( openindex == 0 )
		{
			
			// bracket at start
			content = content.substring( openindex + 1 );	
			
		}
		else if ( closeindex == ( content.length() -1 ) )
		{
			
			// bracket at end
			content = content.substring( 0, closeindex );
			
		}
		
		return content.trim();
		
	}
		/**
	 * Strips the outside pair of brackets from a string. If only one odd bracket is present at
	 * the start or the end of the string then this will be removed as well. The string
	 * returned will also have any whitespace at the start and end removed.
	 *
	 * @param content The string to remove the brackets from
	 * @return The string with the brackets removed
	 **/
	public static String trimSquareBrackets( String content )
	{
		
		// Possible inputs:
		// ( content ) <- remove both
		// content ) <- remove odd one at the end
		// ( content <- remove one at front
		// content <- do nothing
		
		content = content.trim(); // Remove spaces at the start and the end
		
		int openindex = content.indexOf( BASETOKENS.OPENSQUAREBRACKET );
		int closeindex = content.lastIndexOf( BASETOKENS.CLOSESQUAREBRACKET );
		
		if ( openindex == 0 && closeindex == ( content.length() - 1 ) )
		{
			
			// brackets at both ends
			content = content.substring( openindex + 1, closeindex );
			
		}
		else if ( openindex == 0 )
		{
			
			// bracket at start
			content = content.substring( openindex + 1 );	
			
		}
		else if ( closeindex == ( content.length() -1 ) )
		{
			
			// bracket at end
			content = content.substring( 0, closeindex );
			
		}
		
		return content.trim();
		
	}
  

//----------------------------------------------------------------------
// TEST HARNESS
//---------------------------------------------------------------------- 
  
  /* public static void main( String args[] ) {
	
	String test = "     this should not have ((()) (( ( ))))) (any )brackets around it at all    ";
	System.out.println( "'" + trimBrackets( test ) + "'" );
	
  } */
  
  
//----------------------------------------------------------------------
// TEST HARNESS
//---------------------------------------------------------------------- 
  
  /* public static void main( String args[] ) {
	
	String test = "     this should not have ((()) (( ( ))))) (any )brackets around it at all    ";
	System.out.println( "'" + trimBrackets( test ) + "'" );
	
  } */
  
  
} // End StringParser