//******************************************************************************
//
//  Open Source Copyright Notice and License: FIPA-OS
//
//  The contents of this file are subject to the Emorphia Public License (the 
//  "License"); you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at http://www.emorphia.com/EPL/
//
//  Software distributed under the License is distributed on an "AS IS" basis, 
//  WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License 
//  for the specific language governing rights and limitations under the
//  License.
//
//  The Original Code is FIPA-OS.
//
//  The Initial Developer of the Original Code is Emorphia.
//
//  Portions created by Emorphia are Copyright (C) 2001 Emorphia Limited
//  licensed under the Emorphia Public License, available at
//  http://www.emorphia.com/EPL/.
//
//  All Rights Reserved.
//
//  Contributor(s):
//  (add contributor names here)
//
//******************************************************************************

package fipaos.util;

import java.util.*;

/**
 * Implements a pool of Threads - enables multiple threads to
 * be used within an apllication without the associated overhead of
 * starting new Threads, since Threads are re-used
 */
public class ThreadPool implements Runnable
{
	/**
	 * Global thread-pool
	 */
	private static ThreadPool __global_thread_pool;
	
	/**
	 * Set of all ThreadPool objects
	 */
	private static Set __all_thread_pools = Collections.synchronizedSet( new HashSet() );
	
	/**
	 * List of Threads in this ThreadPool
	 */
	private List _threads;
	
	/**
	 * Queue containing jobs to be executed
	 */
	private Queue _jobs;
	
	/**
	 * Indicates if Threads should exit
	 */
	private boolean _shutdown;
	
	/**
	 * Count of active threads
	 */
	private int _active = 0;
	
	/**
	 * Cumulative time spent executing
	 */
	private long _total_active = 0;
	
	/**
	 * Cumulative time spent not executing
	 */
	private long _total_inactive = 0;
	
	/**
	 * Name of this thread pool
	 */
	private String _name;
	
	/**
	 * Create a thread pool with the specified number of threads, and a name
	 * to use as a prefix for the thread names
	 */
	public ThreadPool( int number_of_threads, String name )
	{
		// Save name
		_name = name;
		
		// Create list to contain all threads
		_threads = new LinkedList();
		
		// Create queue for jobs
		_jobs = new FIFOQueue();
		
		// Create required number of threads & start
		for( int i = 0; i < number_of_threads; i++ )
		{
			Thread t = new TPThread( this, name + "-" + i );
			_threads.add( t );
			t.start();
		}
		
		// Add this to set of thread-pools
		__all_thread_pools.add( this );
	}
	
	/**
	 * Invoked by each thread in the ThreadPool
	 */
	public void run()
	{
		long last_stop = System.currentTimeMillis();
		long last_start = last_stop;
		
		while( !_shutdown )
		{
			// Wait for next event
			Object obj[] = (Object[]) _jobs.getItem();
			last_start = System.currentTimeMillis();
			
			// Increment active thread count
			synchronized( this )
			{
				_active++;
				_total_inactive += last_start - last_stop;
			}
	
			// Check we actually got a job
			if( obj != null )
			{
				Job job = (Job) obj[0];
				ThreadGroup tg = (ThreadGroup) obj[1];
				( (TPThread) Thread.currentThread() )._parent_thread_group = tg;
				
				try
				{
					job.go();
				}
				catch( Throwable t )
				{
					DIAGNOSTICS.println( "Exception occured during Job execution", t, this, DIAGNOSTICS.LEVEL_MAX );
				}
				
				// Release references to aid garbage collection
				job = null;
				obj[0] = null;
				obj[1] = null;
				obj = null;
			}
			else
			{
				DIAGNOSTICS.println( "job == null", this );
			}
			
			last_stop = System.currentTimeMillis();
			
			// Decrement active thread count
			synchronized( this )
			{
				_active--;
				_total_active += last_stop - last_start;
			}
		}
	}
	
	/**
	 * Add a Job to be executed by the ThreadPool
	 */
	public synchronized void addJob( Job job )
	{
		if( !_shutdown )
		{
			_jobs.addItem( new Object[] { job, Thread.currentThread().getThreadGroup() } ); 
		}
	}
	
	/**
	 * Terminate all threads & stop executing pending jobs - doesn't block
	 */
	public synchronized void shutdown()
	{
		if( !_shutdown )
		{
			// Indicate that this is shutdown
			_shutdown = true;
			
			// Release all waiting threads
			_jobs.releaseBlock();
			
			// Remove this from the set of ThreadPools
			__all_thread_pools.remove( this );
		}
	}
	
	/**
	 * Get the name of the thread pool
	 */
	public String getName()
	{
		return _name;
	}
	
	/**
	 * Get the number of active threads
	 */
	public synchronized int getActiveThreadCount()
	{
		return _active;
	}
	
	/**
	 * Get the number of threads
	 */
	public int getThreadCount()
	{
		return _threads.size();
	}
	
	/**
	 * Get the number of active threads
	 */
	public synchronized long getTotalActiveTime()
	{
		return _total_active;
	}
	
	/**
	 * Get the number of threads
	 */
	public synchronized long getTotalInactiveTime()
	{
		return _total_inactive;
	}
	
	/**
	 * Get the number of waiting Jobs
	 */
	public int getWaitingJobCount()
	{
		return _jobs.queueLength();
	}
	
	/**
	 * Returns name of ThreadPool
	 */
	public String toString()
	{
		return _name;
	}
	
	/**
	 * Get the global thread-pool
	 */
	public synchronized static ThreadPool getGlobalThreadPool()
	{
		// Check if a global thread pool exists
		if( __global_thread_pool == null )
		{
			__global_thread_pool = new ThreadPool( 10, "GlobalThreadPool" );
		}
		
		return __global_thread_pool;
	}
	
	/**
	 * Get a set of all thread-pools
	 */
	public static Set getThreadPools()
	{
		return new HashSet( __all_thread_pools );
	}
	

	/**
	 * Interface to be implemented by objects to be executed by threads within a ThreadPool
	 */
	public interface Job
	{
		/**
		 * Invoked by a free-Thread when one becomes available
		 */
		public void go();
	}
	
	/**
	 * Extends Thread to provide the ThreadGroup of the thread that requested the
	 * Job be carried out
	 */
	public class TPThread extends Thread
	{
		public ThreadGroup _parent_thread_group;
		
		public TPThread( Runnable r, String name )
		{
			super( r, name );
		}
		
		public ThreadGroup getTPThreadGroup()
		{
			if( _parent_thread_group != null )
			{
				return _parent_thread_group;
			}
			else
			{
				return super.getThreadGroup();
			}
		}
	}
	
	
	/**
	 * Test harness
	 */
	public static void main( String args[] )
	{
		// Start 10 Threads, get them to execute our Jobs, and then shutdown
		int threads = 10;
		final ThreadPool tp = new ThreadPool( threads, "test" );
		Job b = new Job()
		{
			public synchronized void go()
			{
				System.out.println( Thread.currentThread().getName() );
			}
		};
		
		// Give threads * 3 jobs
		for( int i = 0; i < threads * 3; i++ )
		{
			tp.addJob( b );
		}
		
		// Wait 5 seconds & exit
		try
		{
			Thread.sleep( 5000 );
		}
		catch( Throwable t )
		{
		}
		
		// Terminate TP
		tp.shutdown();
	}
}