//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.util;

//----------------------------------------------------------------------

import java.io.Serializable;
import java.util.*;

//----------------------------------------------------------------------

/**
 * Provides a generalised URL class which can handle and protocol type (unlike
 * java.net.URL)
 * <br>
 * <br>
 */
public class URL implements Serializable, Cloneable
{
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	private String _protocol;

	private String _user;
	
	private String _pass;

	private String _host;
	
	private int _port = -1;
	
	private String _target;

	private boolean	_ior_mailto = false;
	
//----------------------------------------------------------------------
// CONSTRUCTOR(S)
//----------------------------------------------------------------------

	/**
	 * Allows an empty URL to be created
	 */
	public URL()
	{
	}
	
	/** Create a URL based upon the given URL
	 * @param url A URL to based the new URL object on
	 */
	public URL( String url )
	{
		try
		{
			// Get the protocol
			int protocol_sep = url.indexOf( "://" );
			int protocol_sep2 = url.indexOf( ":" );

			if( protocol_sep != -1 )
			{
				_protocol = url.substring( 0, protocol_sep );
				protocol_sep += 3;
			}
			else if( protocol_sep2 != -1 )
			{
				_protocol = url.substring( 0, protocol_sep2 );
				protocol_sep = protocol_sep2 + 1;
				_ior_mailto = true;
			}


			// Check if mailto or IOR style
			if( _ior_mailto )
			{
				if( _protocol.equalsIgnoreCase( "IOR" ) )
				{
					_host = url.substring( protocol_sep );
				}
				else
				{
					int at_sep = url.indexOf( "@", protocol_sep );

					// Extract username, and hostname 
					if( at_sep != -1 )
					{
						int user_pass_sep = url.indexOf( ":", protocol_sep );

						// Check if ":" found, and is before @
						if( user_pass_sep != -1 && user_pass_sep < at_sep )
						{
							_user = url.substring( protocol_sep, user_pass_sep );
							_pass = url.substring( user_pass_sep + 1, at_sep );
						}
						else
						{
							_user = url.substring( protocol_sep, at_sep );					
						}
						
						_host = url.substring( at_sep + 1 );
					}
					else
					{
						_host = url.substring( protocol_sep );
					}
				}
			}

			// Standard "well-known" URL type
			else
			{

				// Get [<user>[:<password>]@]<host>[:<port>]

				int target_start = url.indexOf( "/", protocol_sep );

				// If start of target not found, point target_start at end of String
				if( target_start == -1 )
				{
					target_start = url.length();
				}

				// Extract the substring we're after
				String middle_part = url.substring( protocol_sep, target_start );


				// Extract username, password, host and port info

				int at_sep = middle_part.indexOf( "@" );

				// Extract username, and/or password if present
				if( at_sep != -1 )
				{
					int user_pass_sep = middle_part.indexOf( ":" );

					// Check if ":" found, and is before @
					if( user_pass_sep != -1 && user_pass_sep < at_sep )
					{
						_user = middle_part.substring( 0, user_pass_sep );
						_pass = middle_part.substring( user_pass_sep + 1, at_sep );
					}
					else
					{
						_user = middle_part.substring( 0, at_sep );					
					}

					at_sep++;
				}
				else
				{
					at_sep = 0;
				}


				// Extract hostname and port (if present)

				int host_port_sep = middle_part.indexOf( ":", at_sep );

				// Check port found
				if( host_port_sep != -1 )
				{
					_host = middle_part.substring( at_sep, host_port_sep );
					_port = Integer.parseInt( middle_part.substring( host_port_sep + 1 ) );
				}
				else
				{
					_host = middle_part.substring( at_sep );
				}
				
				
				// Extract target (if present)
				if( target_start + 1 < url.length() )
				{
					_target = url.substring( target_start + 1 );
				}
			}
		}
		catch( Throwable t )
		{
			// TBD: Throw ParserException (eventually)
			DIAGNOSTICS.println( "Exception when parsing URL", t, this, DIAGNOSTICS.LEVEL_MAX );
		}
	}
	
	/**
	 * Duplicate the given URL
	 *
	 * @deprecated Please use clone() instead
	 */
	public URL( URL other )
	{
		// Duplicate the content of the given URL
		_port = other._port;
		_host = other._host;
		_pass = other._pass;
		_protocol = other._protocol;
		_target = other._target;
		_user = other._user;
		_ior_mailto = other._ior_mailto;
	}
	
	public Object clone()
	{
		try
		{
			return super.clone();
		}
		catch( CloneNotSupportedException e )
		{
			return null;
		}		
	}
	
//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------	
	
	/** Gets the protocol associated with the URL
	 * @return The protocol (e.g. iiop, rmi, http etc...)
	 */
	public String getProtocol()
	{
		return _protocol;
	}

	/** Gets the host/IP address contained within the URL
	 * @return Hostname/IP address
	 */
	public String getHost()
	{
		return _host;
	}

	/** Gets the port associated with the URL
	 * @return Port number
	 */
	public int getPort()
	{
		return _port;
	}

	/** Gets the "target" of the URL (or "path")
	 * @return Target of the URL
	 */
	public String getTarget()
	{
		return _target;
	}

	/** Username associated with the URL (i.e. as in ftp://user:pass@somehost.com/)
	 * @return The username
	 */
	public String getUser()
	{
		return _user;
	}

	/** Password associated with the URL (i.e. as in ftp://user:pass@somehost.com/)
	 * @return Password
	 */
	public String getPass()
	{
		return _pass;
	}

	/** Sets the protocol of the URL
	 * @param s The protocol (e.g. iiop, rmi, http etc...)
	 */
	public void setProtocol( String s )
	{
		_protocol = s;
	}

	/** Sets the host associated with the URL
	 * @param s A hostname/IP address
	 */
	public void setHost( String s )
	{
		_host = s;
	}

	/** Sets the port associated with the URL
	 * @param s The port-number, or -1 for none
	 */
	public void setPort( int s )
	{
		_port = s;
	}

	/** Sets the "target" (or "path") or the URL
	 * @param s The path (without leading "/")
	 */
	public void setTarget( String s )
	{
		_target = s;
	}

	/** Sets the username associated with the URL (i.e. as in ftp://user:pass@somehost.com/)
	 * @param s Username
	 */
	public void setUser( String s )
	{
		_user = s;
	}

	/** Sets the password associated with the URL (i.e. as in ftp://user:pass@somehost.com/)
	 * @param s Password
	 */
	public void setPass( String s )
	{
		_pass = s;
	}

	/** Converts back to a Stringified URL format
	 * @return String representation of this
	 */
	public String toString()
	{
		// Produce StringBuffer to handle output efficiently
		StringBuffer out = new StringBuffer();
		
		// Add the protocol
		out.append( _protocol );
		
		// Add appropriate seperator
		if( !_ior_mailto )
		{
			out.append( "://" );
		}
		else
		{
			out.append( ":" );
		}
		
		// Add username and/or password
		if( _user != null )
		{
			out.append( _user );
			
			if( _pass != null )
			{
				out.append( ':' );
				out.append( _pass );
			}
			
			out.append( '@' );
		}
		
		// Add the hostname
		out.append( _host );
		
		// Add port-number
		if( _port != -1 )
		{
			out.append( ':' );
			out.append( _port );
		}	
		
		// Add final "slash" and target
		if( _target != null )
		{
			if( !_ior_mailto )
			{
				out.append( '/' );
			}
			out.append( _target );
		}
		
		return out.toString();
	}
	
	/**
	 * Compares this with the given object
	 */ 
	public boolean equals( Object o )
	{
		boolean result = false;
		
		if( o instanceof URL )
		{
			URL url = (URL) o;
			
			result = _ior_mailto == url._ior_mailto;
			result &= _port == url._port;
			result &= compare( _host, url._host );
			result &= compare( _pass, url._pass );
			result &= compare( _protocol, url._protocol );
			result &= compare( _target, url._target );
			result &= compare( _user, url._user );
		}
		
		return result;
	}
	
	/**
	 * Compares two "Strings"
	 */
	public boolean compare( String a, String b )
	{
		if( a == b )
		{
			return true;
		}
		
		return ( a != null ) && ( b != null ) && a.equals( b );
	}
	
	/**
	 * Generates hashcodes
	 */
	public int hashCode()
	{
		return _protocol.hashCode() - _port;
	}
	
	/**
	 * Method to provide property hints to DynamicClass
	 */
	public static DynamicClass.PropertyHint getDynamicClassHints( String prop )
	{
		if( prop.equals( "" ) )
		{
			return new DynamicClass.ImmutableHint();
		}
		else
		{
			return null;
		}
	}
	
//----------------------------------------------------------------------
// MAIN/TESTING METHOD
//----------------------------------------------------------------------
  
	public static void main( String args[] )
	{
		List urls = new LinkedList();
		urls.add( "rmi-iiop://localhost:8080/myagent" );
		urls.add( "IOR:239472497203942936593247592374021374" );
		urls.add( "corbaname::localhost:2000/NameService#hello/world" );
		urls.add( "mailto:treadway@nortelnetworks.com" );
		urls.add( "http://localhost/myfile" );
		urls.add( "ftp://me@localhost/" );
		urls.add( "ftp://me:pass@localhost:8080" );
		urls.add( "wap://localhost" );
		urls.add( "fipaos-socket://host:8500/ams@localhost" );
		
		Iterator i = urls.iterator();
		while( i.hasNext() )
		{
			String next = (String) i.next();
			System.out.println( "\n         URL = " + next );
			System.out.println( "  Parsed URL = " + new URL( next ) );
			System.out.println( "Reparsed URL = " + new URL( new URL( next ).toString() ) );
		}
	}
	
} // End class