//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.util;

//----------------------------------------------------------------------

import java.io.Serializable;

//----------------------------------------------------------------------

  /**
   * Simple container for UTC time data
   */

public class UTCContainer implements Serializable, Cloneable
{
	 
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

	/**
	 * Constant to represent the year
	 */
	public static final int YEAR=0;

	/**
	 * Constant to represent the month
	 */
	public static final int MONTH=1;

	/**
	 * Constant to represent the date
	 */
	public static final int DATE=2;

	/**
	 * Constant to represent the hour of day (24 hour format)
	 */
	public static final int HOUR=3;

	/**
	 * Constant to represent the minute of the hour
	 */
	public static final int MINUTE=4;
		
	/**
	 * Constant to represent the second of the minute
	 */
	public static final int SECOND=5;		
	
	/**
	 * Constant to represent the millisecond of the second
	 */
	public static final int MILLISECOND=6;		

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------
		
//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	/**
	 * Array to contain the date/time information
	 */
	protected int _date_data[];

	/**
	 * Boolean to denote if seconds should be used
	 */
	protected boolean _use_seconds;

	/**
	 * Boolean to denote if milliseconds should be used
	 */
	protected boolean _use_millis;
	
	
//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

	/**
	 * Constructs a UTCContainer object from the given parameters
	 *
	 * @param year The year
	 * @param month The month (1 to 12)
	 * @param date The day of the month (1 to (28 thru 31))
	 * @param hour The hour of the day (0 to 23)
	 * @param minute The minute of the hour (0 to 59)
	 * @param second The second of the minute (0 to 59, or -1 for minute precision)
	 * @param millis The millisecond of the second (0 to 999, or -1 for second precision)
	 */
	public UTCContainer(int year, int month, int date, int hour,
									int minute, int second, int millis)
	{
		// Initialise date storage array
		_date_data = new int[7];
		
		_use_seconds = true;
		_use_millis = true;
		
		// Store the values given
		_date_data[YEAR] = year;
		_date_data[MONTH] = month;
		_date_data[DATE] = date;
		_date_data[HOUR] = hour;
		_date_data[MINUTE] = minute;
		
		if( second == -1 )
		{
			_use_seconds = false;
			_use_millis = false;
		}
		else
		{
			_date_data[SECOND] = second;
		}

		if( millis == -1 )
		{
			_use_millis = false;
		}
		else if( _use_millis )
		{
			_date_data[MILLISECOND] = millis;
		}
	}

	/**
	 * Constructs a "blank" UTCContainer
	 */
	protected UTCContainer()
	{
		// Initialise date storage array
		_date_data = new int[7];
	}


//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------

	/**
	 * Method to allow retreival of individual
	 * fields from the time/date
	 *
	 * @param field The field which is to be retreived
	 * @return The value of the given field
	 */
	public int get(int field)
	{
		// Check the field given is valid
		if(field>=YEAR && field<=MILLISECOND)
			return _date_data[field];
		else
			throw new ArrayIndexOutOfBoundsException("Unknown field of UTCContainer specified");
	}		


	/**
	 * Method to allow manual setting of date/time fields
	 *
	 * @param field The field which is to be changed
	 * @param value The new value to set the field to
	 */
	public void set(int field,int value)
	{
		// Check the field given is valid
		if(field>=YEAR && field<=MILLISECOND)
			_date_data[field] = value;
		else
			throw new ArrayIndexOutOfBoundsException("Unknown field of UTCContainer specified");
	}		


	/**
	 * Method to provide a way of determining if a UTCContainer object represents a time after
	 * the time this UTCContainer instance represents
	 *
	 * @param time A UTCContainer object to compare
	 */
	public boolean isBefore(UTCContainer otherTime)
	{
		// Check each field in descending significance
		for( int x=YEAR; x<=MILLISECOND; x++)
		{
			if(_date_data[x] < otherTime._date_data[x])
				return true;
			else if(_date_data[x] > otherTime._date_data[x])
				return false;
		}

		// if we've reached here, then the times must be equal!
		return false;			
	}
	

	/**
	 * Method to provide a way of determining if a UTCContainer object represents a time before
	 * the time this UTCContainer instance represents.
	 *
	 * @param time A UTCContainer object to compare
	 */
	public boolean isAfter(UTCContainer time)
	{
		return !(isBefore(time) || equals(time));	
	}


	/**
	 * Method to provide a mechanism for determining if two times are the same
	 *
	 * @param time The time to compare
	 */
	public boolean equals(UTCContainer otherTime)
	{
		// Test each field to see if they are equal
		for(int x=YEAR; x<=MILLISECOND; x++)
		{
			// Return false if a field is not equal
			if(_date_data[x] != otherTime._date_data[x])
				return false;
		}
		
		// If we reach here, then the times are equal
		return true;
	}


	/**
	 * Overrides default Object.toString()
	 */
	public String toString()
	{
		String UTCString = "";

		for(int i=YEAR;i<=MILLISECOND;i++)
		{
			if(i==YEAR)
			{
				// Calculate the number of leading "0"'s required
				String year = "" + _date_data[i];
				int leading = 4-year.length();
				
				// Add leading zero's
				for( int x=0;x<leading;x++)
					UTCString += "0";
				
				// Add year field
				UTCString += _date_data[i];
			}
			else
			{
				if(i==HOUR)
				{
					UTCString += "T";
				}
					

				// Check if we're using seconds or millis
				if( !_use_seconds && ( i == SECOND || i == MILLISECOND ) )
				{
					break;
				}
				else if( !_use_millis && i == MILLISECOND )
				{
					break;
				}
					
					
				UTCString += ( i == MILLISECOND ? ( _date_data[i]<100 ? "0" : "" ) : "" );
				UTCString += ( _date_data[i]<10 ? "0" : "");
				UTCString += _date_data[i];
			}
		}

		return UTCString;
	}

	/**
	 * Method to change accuracy of UTCTime object.  This
	 * can be changed to MINUTE, SECOND or MILLISECOND.
	 * NOTE: Internally the contents of unused fields
	 * will be reset to 0.
	 *
	 * @param accuracy One of the above constants
	 */
	public void setAccuracy(int accuracy)
	{
		if(accuracy == MINUTE)
		{
			// If the accuracy required is minutes,
			// don't use seconds or millis
			_use_seconds = false;
			_use_millis = false;
			
			_date_data[SECOND] = 0;
			_date_data[MILLISECOND] = 0;
		}
		else if(accuracy == SECOND)
		{
			// If the accuracy required is seconds,
			// don't use millis
			_use_seconds = true;
			_use_millis = false;

			_date_data[MILLISECOND] = 0;
		}
		else if(accuracy == MILLISECOND)
		{
			// If the accuracy required is millis,
			// use seconds and millis
			_use_seconds = true;
			_use_millis = true;
		}
		else
		{
			// An illegal accuracy has been specified! STOP!
			throw new RuntimeException("ERROR IN UTCContainer - CANNOT SET ACCURACY TO VALUE SPECIFIED");
		}
	}


	public Object clone()
	{
		try
		{
			UTCContainer utc = (UTCContainer) super.clone();
			utc._date_data = new int[ _date_data.length ];
			System.arraycopy( _date_data, 0, utc._date_data, 0, _date_data.length );
			return utc;
		}
		catch( CloneNotSupportedException e )
		{
			return null;
		}
	}
	
	
//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------


//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------


}