//******************************************************************************
//  Open Source Copyright Notice and License: FIPA-OS
//
//  1. The programs and other works made available to you in these files 
//     ("the Programs") are Copyright (c) 1999 - 2000 Nortel Networks
//     Corporation, 8200 Dixie Road, Suite 100, Brampton, Ontario, Canada 
//     L6R 5P6.  All rights reserved.
//
//  2. Your rights to copy, distribute and modify the Programs are as set out 
//     in the Nortel Networks FIPA-OS Public License, a copy of which can be
//     found in file "Nortel_FIPA_OS_Public_Licence.txt" and the latest version 
//     can also be found at http://fipa-os.sourceforge.net/. By  downloading
//     the files containing the Programs you accept the terms and conditions of
//     the Public License. You do not have to accept these terms and conditions,
//     but unless you do so you have no rights to use the Programs.
//
//  The Original Code is Nortel Networks' FIPA-OS (Foundation for Intelligent
//  Physical Agents - Open Source).
//
//  The Initial Developer of the Original Code is Nortel Networks Corporation.
//  Portions created by Nortel Networks Corporation or its subsidiaries are
//  Copyright (c) 1999 - 2000 Nortel Networks Corporation. All Rights Reserved.
//
//  Contributor(s): 
//    Emorphia agree to provide Modifications to Nortel Networks' FIPA-OS 
//    Covered Code under Nortel Networks' FIPA-OS Public License. All Emorphia's 
//    Modifications remain Copyright (c) 2001 Emorphia Limited. All Rights
//    Reserved. 
//
//******************************************************************************

package fipaos.util;
 
//----------------------------------------------------------------------

import java.io.*;
import java.util.*;

    /**
   * This is a class which can be used to directly manipulate UTC
   * time strings in a consitant manner.  Due to the unreliablility
   * of information about which timezone the user is in from Java,
   * details of the local timezone are picked up from a user.profile
   * file in the currect directory (this is not an ideal solution
   * since the information should really come from some form of roaming
   * profile).  An example of the user.profile file is given below
   *
   * User profile.  Includes user specific information about
	 * the timezone they are working in (at a future date this
	 * should be picked up from the user's roaming profile)
   *
	 * Local timezone offset from Z (GMT) excluding DST <br>
	 * TimeZoneOffset = 0
	 *
	 * <br> Offset from normal local timezone during DST
	 * If this is omitted or == 0, DST is not used <br>
	 * DSTOffset = 1.0
	 *
	 * <br> <br> Details about when DST should start <br>
	 *
	 * <br> DSTStartMonth indicates the starting month of DST, and should
	 * be in the range 1 to 12 <br>
	 *
	 * <br> DSTStartWeekOfMonth indicates which week of the month DST should
	 * start.  Negative values are relative to the end of the month. <br>
	 *
	 * <br> DSTStartDayOfWeek indicates which day of the week DST should start.
	 * This should be in the range 0 to 6, with 0 being Sunday and 6 being
	 * Saturday (e.g. 2 = Tuesday) <br>
	 *
	 * <br> DSTStartHourOfDay indicates at which hour of the day (24 hour format)
	 * the change should occur.  This should be in the range 0 to 23 <br>
	 *
	 * DSTStartMonth = 3 <br>
	 * DSTStartWeekOfMonth = -1 <br>
	 * DSTStartDayOfWeek = 0 <br>
	 * DSTStartHourOfDay = 2 <br>
	 *
	 * <br> Details about when DST should end.  Similar to DSTStart
	 * properties above  <br>
	 *
	 * DSTEndMonth = 10 <br>
	 * DSTEndWeekOfMonth = -1 <br>
	 * DSTEndDayOfWeek = 0 <br>
	 * DSTEndHourOfDay = 2 <br>
	 *
	 *<br> Use of set method<br>
	 * a problem can occur when setting the MONTH attibute <br>
	 * if the DATE attribute is after the last day of the month eg.30th when the <br>
	 * month is February this will have the effect of incrementing the month <br>
	 * this can be avoided bysetting the DATE atribute to below the last day of the MONTH <br>
	 * before setting the MONTH attribute <br>
   *
   * <br>
   * <br>
   */

public class UTCTime extends UTCContainer implements Cloneable, Serializable
{
	 
//----------------------------------------------------------------------
// CONSTANTS
//----------------------------------------------------------------------

	/**
	 * List of timezones
	 */
	private static final String __time_zones[] = { "Z", "A", "B", "C", "C*", "D", "D*", "E", "E*", "F",
																								 "F*", "G", "H", "I", "I*", "K", "K*", "L", "L*", "M",
																								 "M*", "N", "O", "P", "P*", "Q", "R", "S", "T", "U",
																								 "U*", "V", "V*", "W", "X", "Y" };

	/**
	 * List of timezone offsets (in half-hours)
	 */
	private static final int __time_zone_offsets[] = { 0, 2, 4, 6, 7, 8, 9, 10, 11, 12,
																										 13, 14, 16, 18, 19, 20, 21, 22, 23, 24,
																										 26, -2, -4, -6, -7, -8, -10, -12, -14, -16,
																										 -17, -18, -19, -20, -22, -24 };

//----------------------------------------------------------------------
// STATIC VARIABLES
//----------------------------------------------------------------------

	/**
	 * Hashtable of timezones to offsets
	 */
	private static Hashtable __timezone_to_offset;


	/**
	 * Variables containing local timezone information
	 * (offsets are in half hours)
	 */
	private static int __local_offset_from_GMT;
	private static int __local_dst_offset;
	private static int __local_dst_start[];
	private static int __local_dst_end[];


	/**
	 * Initialise all static variables
	 */
	static
	{
		// Initialise timezone offset lookup table
		__timezone_to_offset = new Hashtable();
		for(int i=0;i<__time_zones.length;i++)
		{
			__timezone_to_offset.put( __time_zones[i], new Integer( __time_zone_offsets[i] ) );
		}
		
		
		// Attempt to get info about current timezone
		try
		{
			// Open the users profile and retrieve info on current timezone and daylight saving
			// BufferedReader in = new BufferedReader(new FileReader( "user.profile" ));
			Properties user_properties = new Properties();
			user_properties.load(new FileInputStream( "user.profile" ));
			
			// Retrieve properties from file, or use default values
			String local_offset = user_properties.getProperty( "TimeZoneOffset", "0.0");
			String dst_offset = user_properties.getProperty( "DSTOffset" );
			
			__local_offset_from_GMT = (int) ( ( new Double( local_offset ) ).doubleValue() * 2 );
			
			// Check if daylight saving has been specified
			if( ! ( (dst_offset == null) || dst_offset.equals("") ) )
			{
				// If so, get the offset
				__local_dst_offset = (int) ( ( new Double( dst_offset ) ).doubleValue() * 2 );
				
				// Read in information about start and end times (or use specified defaults)
				String dst_start_month = user_properties.getProperty( "DSTStartMonth", "3" );
				String dst_start_day_of_week = user_properties.getProperty( "DSTStartDayOfWeek", "0" );
				String dst_start_week_of_month = user_properties.getProperty( "DSTStartWeekOfMonth", "-1" );
				String dst_start_hour_of_day = user_properties.getProperty( "DSTStartHourOfDay", "2" );
				String dst_end_month = user_properties.getProperty( "DSTEndMonth", "10" );
				String dst_end_day_of_week = user_properties.getProperty( "DSTEndDayOfWeek", "0" );
				String dst_end_week_of_month = user_properties.getProperty( "DSTEndWeekOfMonth", "-1" );
				String dst_end_hour_of_day = user_properties.getProperty( "DSTEndHourOfDay", "2" );
				
				// Parse the values obtained above
				__local_dst_start = new int[4];
				__local_dst_end = new int[4];
				
				__local_dst_start[0] = Integer.parseInt( dst_start_month );
				__local_dst_start[1] = Integer.parseInt( dst_start_week_of_month );
				__local_dst_start[2] = Integer.parseInt( dst_start_day_of_week );
				__local_dst_start[3] = Integer.parseInt( dst_start_hour_of_day );
				
				__local_dst_end[0] = Integer.parseInt( dst_end_month );
				__local_dst_end[1] = Integer.parseInt( dst_end_week_of_month );				
				__local_dst_end[2] = Integer.parseInt( dst_end_day_of_week );
				__local_dst_end[3] = Integer.parseInt( dst_end_hour_of_day );
			}
			else
			{
				// If not, indicate so in the __local_dst_offset variable
				__local_dst_offset = 0;
			}
		}
		catch (Exception e)
		{
			// If there is an error whilst trying to read the user.profile file,
			// use the default British DST parameters
			// fipaos.util.DIAGNOSTICS.println("Couldn't find user.profile for timezones, using default British setup for UTCTime");
			
			// Offset from GMT in half hours
			__local_offset_from_GMT = 0;
			
			// DST offset in half hours
			__local_dst_offset = 2;
			
			// Initialise arrays
			__local_dst_start = new int[4];
			__local_dst_end = new int[4];
				
			// Setup DST start to be last Sunday in March at 2am
			__local_dst_start[0] = 3;
			__local_dst_start[1] = -1;
			__local_dst_start[2] = 0;
			__local_dst_start[3] = 2;
			
			// Setup DST end to be last Sunday in October at 2am
			__local_dst_end[0] = 10;
			__local_dst_end[1] = -1;				
			__local_dst_end[2] = 0;
			__local_dst_end[3] = 2;
		}
	}


//----------------------------------------------------------------------
// INSTANCE VARIABLES
//----------------------------------------------------------------------

	/**
	 * String to denote which timezone this time represents
	 */
	private String _time_zone;

	/**
	 * Boolean to denote if seconds should be used
	 */
	private boolean _use_seconds;

	/**
	 * Boolean to denote if milliseconds should be used
	 */
	private boolean _use_millis;

//----------------------------------------------------------------------
// CONSTRUCTOR
//----------------------------------------------------------------------

	/**
	 * Constructs a UTCTime object given a UTC string of the
	 * form 19990427T1200Z or 19990427T120000Z.  If the time zone
	 * is ommited, the local time zone is assumed.  Seconds and
	 * milliseconds may not need to be specified.  If the string is
	 * prefixed with a + the time is taken to be relative to the
	 * present time.
	 *
	 * @param time A UTC time string of the form YYYYMMDDTHHMMZ or
	 *             YYYYMMDDTHHMMSSMMMZ where YYYY=year, MM=month, DD=date,
	 *             HH=hour, second MM = minutes, SS = seconds, MMM = millisecond
	 *             and Z=timezone
	 */
	public UTCTime(String time)
	{
		// Initialise date storage array
		_use_seconds = false;
		_use_millis = false;

		// Storage for the current time if necessary
		UTCTime current_time = new UTCTime();
		String unparsed_string = time;
		boolean relative_time = false;

		// Determine if the time string is a relative offset
		if( time.substring(0,1).equals("+") )
		{
			// If so, get the current time
			// and prune the "+"
			unparsed_string = time.substring(1);
			relative_time = true;
		}
		
		try
		{
			// Parse the string for the time/date/timezone
			_date_data[YEAR] = Integer.parseInt( unparsed_string.substring( 0, 4 ) );
			_date_data[MONTH] = Integer.parseInt( unparsed_string.substring( 4, 6 ) );
			_date_data[DATE] = Integer.parseInt( unparsed_string.substring( 6, 8 ) );
			_date_data[HOUR] = Integer.parseInt( unparsed_string.substring( 9, 11 ) );
			_date_data[MINUTE] = Integer.parseInt( unparsed_string.substring( 11, 13 ) );

			try
			{
				// Try to parse seconds
				_date_data[SECOND] = Integer.parseInt( unparsed_string.substring( 13, 15 ) );

				// If we've reached here, seconds are being used
				_use_seconds = true;
				
				// Try to parse milliseconds
				_date_data[MILLISECOND] = Integer.parseInt( unparsed_string.substring( 15, 18 ) );

				// If we've reached here, milliseconds are being used
				_use_millis = true;
			}
			catch (Exception e)
			{
				// Used since we don't necessarilly want to handle
				// seconds or millis.
			}

			// Only get the timezone if we are not dealing with a relative time offset
			if(!relative_time)
			{
				// Grab the timezone from the end of the string
				if(_use_seconds == false)
				{
					_time_zone = unparsed_string.substring( 13, time.length()).toUpperCase();
				}
				else if(_use_millis == false)
				{
					_time_zone = unparsed_string.substring( 15, time.length()).toUpperCase();
				}
				else
				{
					_time_zone = unparsed_string.substring( 18, time.length()).toUpperCase();	
				}

				// If a timezone is given, check that the timezone is valid
				if( !_time_zone.equals("") )
				{			
					// Check that the timezone given is in our hashtable
					if( !__timezone_to_offset.containsKey( _time_zone ) )
					{
						// If the timezone is unknown, throw an exception
						throw new RuntimeException( "ERROR PARSING UTC STRING - Unknown timezone \"" + _time_zone + "\"");
					}
				}
			}
			else
			{
				// We're dealing with a relative time, this is in local time as default
				_time_zone = "";
			}
		}
		catch (ArrayIndexOutOfBoundsException e)
		{
			// If an error occurs, throw our exception
			throw new RuntimeException( "ERROR PARSING UTC STRING - " + e.toString() );
		}

		// If we're dealing with an offset, add it to the current time
		if(relative_time)
		{
			// Update each field one at a time
			for( int x=MILLISECOND; x>=YEAR; x-- )
			{
				// Update and then re-validate
				_date_data[x] += current_time._date_data[x];
				validate();	
			}	
		}
		else
		{
			// Ensure that all fields are within correct ranges
			validate();
		}

	} // UTCTime()


	/**
	 * Constructs a UTCTime object from the given parameters
	 *
	 * @param year The year
	 * @param month The month (1 to 12)
	 * @param date The day of the month (1 to (28 thru 31))
	 * @param hour The hour of the day (0 to 23)
	 * @param minute The minute of the hour (0 to 59)
	 * @param second The second of the minute (0 to 59, or -1 to ignore minutes)
	 * @param millis The millisecond of the second (0 to 999, or -1 to ignore millis)
	 * @param time_zone Letter representing the timezone
	 *                  If this is null, the local time zone is assumed
	 */
	public UTCTime(int year, int month, int date, int hour,
									int minute, int second, int millis, String time_zone)
	{
		// Initialise date storage array
		_use_seconds = false;
		_use_millis = false;

		// Store the values given
		_date_data[YEAR] = year;
		_date_data[MONTH] = month;
		_date_data[DATE] = date;
		_date_data[HOUR] = hour;
		_date_data[MINUTE] = minute;
		
		// Check if we want to use seconds
		if(second!=-1)
		{
			_date_data[SECOND] = second;
			_use_seconds = true;
		}

		// Check if we want to use milliseconds
		if(millis!=-1)
		{
			_date_data[MILLISECOND] = millis;
			_use_millis = true;	
		}

		// Check to see if a time zone has been specified
		if(time_zone == null || time_zone.equals(""))
		{
			// If not, assume a local time
			_time_zone = "";
		}
		else
		{
			// Check that the timezone is valid
			if( !__timezone_to_offset.containsKey( time_zone ) )
			{
				// If the timezone is unknown, throw an exception
				throw new RuntimeException( "ERROR CREATING UTCTime OBJECT - Unknown timezone \"" + time_zone + "\"");
			}
			else
			{
				_time_zone = time_zone;
			}
		}

		// Ensure that all fields are within correct ranges
		validate();

	} // UTCTime()


	/**
	 * Constructs a UTCTime object to represent the current time
	 */
	public UTCTime()
	{
		// Get todays date/time
		Calendar today = (Calendar) new GregorianCalendar();

		// Initialise date storage array
		_use_seconds = true;
		_use_millis = true;

		// Put todays date/time into the storage array
		_date_data[YEAR] = today.get(Calendar.YEAR);
		_date_data[MONTH] = today.get(Calendar.MONTH) + 1;
		_date_data[DATE] = today.get(Calendar.DATE);
		_date_data[HOUR] = today.get(Calendar.HOUR_OF_DAY);
		_date_data[MINUTE] = today.get(Calendar.MINUTE);
		_date_data[SECOND] = today.get(Calendar.SECOND);
		_date_data[MILLISECOND] = today.get(Calendar.MILLISECOND);
				
		// May not be necessary on non-Windows machines?
		if(isInDaylightSavingTime())
			_date_data[HOUR] += __local_dst_offset/2;
				
		_time_zone = "";

		// Validate the date/time fields, just in case
		validate();

	} // UTCTime()

//----------------------------------------------------------------------
// PUBLIC METHODS
//----------------------------------------------------------------------

	/**
	 * Method to allow addition/subtraction from an element of the
	 * date
	 *
	 * @param field The field of the UTCTime object to change
	 * @param value The value to add to the element (use negative values
	 *        to subtract)
	 */
	public void add(int field,int value)
	{
		// Check the field given is valid
		if(field>=YEAR && field<=MILLISECOND)
			_date_data[field] += value;
		else
			throw new RuntimeException( "ERROR IN UTCTime OBJECT - UNKNOWN FIELD" );

		// Validate the new value
		validate();
	}


	/**
	 * Method to allow addition of a UTCDuration to this UTCTime object
	 * Fields are added from least significant to most significant field
	 *
	 * @param duration The duration to add to this UTCTime
	 */
	public void addDuration(UTCDuration duration)
	{
		// Iterate through each field
		for(int i=MILLISECOND;i>=YEAR;i--)
		{
			// Add the content of that field in duration
			// to the corresponding field in this UTCTime
			add( i, duration.get( i ) );
			
			// Re-validate the field content on each iteration
			validate();
		}
	}


	/**
	 * Method to allow manual setting of date/time fields
	 * All fields will be re-validated before this method
	 * returns
	 *
	 * @param field The field which is to be changed
	 * @param value The new value to set the field to
	 */
	public void set(int field,int value)
	{
		super.set( field, value );

		// Validate the new value
		validate();
	}		


	/**
	 * Method to provide a way of determining if a UTCTime object represents a time after
	 * the time this UTCTime instance represents
	 *
	 * @param time A UTCTime object to compare
	 */
	public boolean isBefore(UTCTime time)
	{
		// Convert the times to a common time zone
		UTCTime thisTime = ((UTCTime) clone());
		thisTime.changeTimeZone("Z");
		UTCTime otherTime	= ((UTCTime) time.clone());
		otherTime.changeTimeZone("Z");
		
		return ((UTCContainer) thisTime).isBefore((UTCContainer) otherTime);
	}
	

	/**
	 * Method to provide a way of determining if a UTCTime object represents a time before
	 * the time this UTCTime instance represents.
	 *
	 * @param time A UTCTime object to compare
	 */
	public boolean isAfter(UTCTime time)
	{
		return !(isBefore(time) || equals(time));	
	}


	/**
	 * Method to provide a mechanism for determining if two times are the same
	 *
	 * @param time The time to compare
	 */
	public boolean equals(UTCTime time)
	{
		// Convert the times to a common time zone
		UTCTime thisTime = ((UTCTime) clone());
		thisTime.changeTimeZone("Z");
		UTCTime otherTime	= ((UTCTime) time.clone());
		otherTime.changeTimeZone("Z");
		
		return ((UTCContainer) thisTime).equals((UTCContainer) otherTime);
	}


	/**
	 * Method to allow change of timezone
	 *
	 * @param timezone A string representing the timezone to change to,
	 *                 a letter in the range A to Z
	 */
	public void setTimeZone(String timezone)
	{
		// Check that the timezone given is a valid timezone
		if( !__timezone_to_offset.containsKey( timezone.toUpperCase() ) )
		{
			throw new RuntimeException( "ERROR PARSING UTC STRING - Unknown timezone \"" + _time_zone + "\"");
		}
		else
		{
			_time_zone = timezone.toUpperCase();
		}
	}


	/**
	 * Method to retreive the timezone of the UTCTime object
	 *
	 * @return A string representing the timezone of the UTCTime object,
	 *         a letter in the range A to Z
	 */
	public String getTimeZone()
	{
		return _time_zone;
	}


	/**
	 * Method to return a representation of this object in
	 * Z time
	 */
	public UTCTime getZTime()
	{
		// Clone this object
		UTCTime time = (UTCTime) this.clone();
		
		// Change its timezone to Z
		time.changeTimeZone( "Z" );
		
		// Pass back
		return time;
	}


	/**
	 * Change timezone from the current to the one specified.
	 * Also changes the time accordingly
	 *
	 * @param new_zone A string representing the new timezone
	 */
	public void changeTimeZone(String new_zone)
	{
		int old_offset, new_offset;
		
		// Check if this object is a local time
		if( _time_zone.equals("") )
		{
			// If we're in local time, get the appropriate offset
			old_offset = getLocalOffset();
		}
		else
		{
			old_offset = ((Integer) __timezone_to_offset.get( _time_zone )).intValue();
		}

		// Check if we're changing to local time
		if( (new_zone == null) || new_zone.equals("") )
		{
			// If so get the local offset
			new_offset = getLocalOffset();
		}
		else if( __timezone_to_offset.containsKey( new_zone.toUpperCase() ) )
		{
			// Otherwise try to lookup the offset
			new_offset = ((Integer) __timezone_to_offset.get( new_zone.toUpperCase() )).intValue();
		}
		else
		{
			// If it isn't recognised, STOP!
			throw new RuntimeException("ERROR IN UTCTime OBJECT - UNKNOWN TIMEZONE \""+new_zone+"\"");
		}
		
		// Calculate the time difference between old and new timezones
		int difference = (new_offset - old_offset) * 30;

		// Add the difference and change the time zone
		add(MINUTE,difference);
		_time_zone = new_zone.toUpperCase();			
	}


	/**
	 * Returns the number of milliseconds since 19700101T0000
	 * in the timezone of this UTCTime object
	 */
	public long getTimeInMillis()
	{
		// Get a copy of this time object
		UTCTime this_time = (UTCTime) clone();
		
		// Calculate number of days since 1970
		long number_of_leap_years = this_time.getLeapYearsSince1970();
		long number_of_days = 365 * (this_time._date_data[YEAR] - 1970) + number_of_leap_years;
		number_of_days += this_time.getDaysSinceYearStart();
		
		// Convert number of days and other fields into milliseconds
		long millis = number_of_days * 24 * 60 * 60 * 1000;
		if(_use_millis)
			millis += this_time._date_data[MILLISECOND];
		if(_use_seconds)
			millis += this_time._date_data[SECOND] * 1000;
		millis += this_time._date_data[MINUTE] * 60 * 1000;		
		millis += this_time._date_data[HOUR] * 60 * 60 * 1000;
		
		return millis;		
	}


	/**
	 * Overrides default Object.toString()
	 */
	public String toString()
	{
		String UTCString = "";

		for(int i=YEAR;i<=MILLISECOND;i++)
		{
			if(i==YEAR)
			{
				UTCString += _date_data[i];
			}
			else
			{
				if(i==SECOND && !_use_seconds)
					continue;
					
				if(i==MILLISECOND && !_use_millis)
					continue;

				if(i==HOUR)
					UTCString += "T";

				UTCString += ( i == MILLISECOND ? ( _date_data[i]<100 ? "0" : "" ) : "" );
				UTCString += ( _date_data[i]<10 ? "0" : "");
				UTCString += _date_data[i];
			}
		}

		UTCString += _time_zone;

		return UTCString;
	}
	
	
	/**
	 * returns the day of the week as an integer sunday - 1, monday - 2 .....saturday - 7	 
	 */
	public int getDayOfWeek()
	{
	
		int year = _date_data[YEAR];
		int month = _date_data[MONTH];
		int date = _date_data[DATE];
		
		//this code was taken from the internet I've adapted it but have not looked at what its
		//doing but it works OK
		
		if (month == 1)
		{
			month = 13;
			year = year - 1;
		}
		else if (month ==2)
		{
			month = 14;
				year = year - 1;
		}
		
		int result = 
		(int)((date + (month * 2) + Math.floor(((month + 1) * 3) / 5) + Math.floor(year / 4) + year - Math.floor(year / 100) + Math.floor(year / 400) + 2) - 
		(Math.floor((date + (month * 2) + Math.floor(((month + 1) * 3) / 5) + Math.floor(year / 4) + year - Math.floor(year / 100) + Math.floor(Math.floor(year / 400) + 2))/7)*7));
		
		//saturday gives a 0 but we need to have saturday as 7
		if (result == 0)result = 7;	
		return result;
	
	}


//----------------------------------------------------------------------
// PRIVATE METHODS
//----------------------------------------------------------------------

	/**
	 * Method to get the number of leap years since 1970
	 * (excludes current year)
	 */
	private int getLeapYearsSince1970()
	{
		// Initialise variants
		int year = 1972;
		int leaps = 0;
		
		while(year < _date_data[YEAR])
		{
			if( (year%4==0) && (year%100!=0 || year%400==0) )
			{
				leaps++;
			}
			year+=4;
		}
		
		return leaps;
	}


	private int getDaysSinceYearStart()
	{
		// Initialise
		int month = 1;
		int days = 0;
		
		while(month < _date_data[MONTH])
		{
			days += getDaysInMonth( _date_data[YEAR], month );
			month++;
		}
		
		days += _date_data[DATE] - 1;
		
		return days;
	}


	/**
	 * Method used to validate all date/time fields. Assumes
	 * least significant elements are "more" correct than
	 * higher significance elements - exception is date/month
	 * since detection of over/under-flow in date is dependant
	 * upon the current month 
	 */
	private void validate()	
	{
		// Validate milliseconds (easy)
		_date_data[SECOND] += (int) Math.floor( _date_data[MILLISECOND] / 1000.0 );
		_date_data[MILLISECOND] = modulus(_date_data[MILLISECOND], 1000);
		
		// Validate seconds (easy)
		_date_data[MINUTE] += (int) Math.floor( _date_data[SECOND] / 60.0 );
		_date_data[SECOND] = modulus(_date_data[SECOND], 60);

		// Validate minutes (easy)
		_date_data[HOUR] += (int) Math.floor( _date_data[MINUTE] / 60.0 );
		_date_data[MINUTE] = modulus(_date_data[MINUTE], 60);

		// Validate hours (easy)
		_date_data[DATE] += (int) Math.floor( _date_data[HOUR] / 24.0 );
		_date_data[HOUR] = modulus(_date_data[HOUR], 24); 

		// Validate months (easy)
		_date_data[YEAR] += (int) Math.floor( (_date_data[MONTH]-1) / 12.0 );
		_date_data[MONTH] = modulus(_date_data[MONTH]-1, 12) + 1;

		// Validate date (not so easy or efficient) and month/year again
		
		if(_date_data[DATE]>0)
		{
			// Adding months on each iteration

			int days_in_month = getDaysInMonth(_date_data[YEAR],_date_data[MONTH]);

			while( _date_data[DATE] > days_in_month )
			{
				// Increment the month
				_date_data[MONTH] ++;

				// Decrement the number of days
				_date_data[DATE] -= days_in_month;

				// Roll over months if necessary
				if(_date_data[MONTH]>12)
				{
					_date_data[YEAR]++;
					_date_data[MONTH] = 1;
				}

				// Get the number of days in the new current month
				days_in_month = getDaysInMonth(_date_data[YEAR],_date_data[MONTH]);
			}
		}
		else
		{
			// Subtracting months on each iteration

			while( _date_data[DATE] < 1 )
			{
				// decrement the month
				_date_data[MONTH]--;

				// Roll over months if necessary
				if(_date_data[MONTH]<1)
				{
					_date_data[YEAR]--;
					_date_data[MONTH] = 12;
				}

				// Get the number of days in the current month
				int days_in_month = getDaysInMonth(_date_data[YEAR],_date_data[MONTH]);

				// Increment the number of days
				_date_data[DATE] += days_in_month;
			}
		}
	}


	/**
	 * Returns the number of days in the given month
	 */
	private int getDaysInMonth(int year,int month)
	{
		// Initialise our list of normal days of month
		int days_of_month[] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

		// Handle leap years
		if((month == 2) && (year%4 == 0))
		{
			if(year%400 == 0)
			{
				return 29;
			}	
			else if(year%100 == 0)
			{
				return 28;
			}	
			else
			{
				return 29;
			}
		}
		else
		{
			// Return the normal amount
			return days_of_month[month-1];
		}		
	}


	/**
	 * Special modulus method - will produce the modulus of
	 * negative numbers as well as positive (unlike the % operator
	 * where (-10 % 4) = -10 - the desired result for this application
	 * would be (-10 % 4) = 2)
	 */
	private int modulus(int value,int divisor)
	{
		if(value>=0)
		{
			// If we're dealing with a positive number
			// just use the % operator
			return (value%divisor);
		}
		else
		{
			// Otherwise do something special!
			int mult = Math.abs(value) / divisor;
			int new_value = value + (mult * divisor);
			if(new_value == 0)
				return 0;
			else
				return (divisor + new_value);
		}
	}


	/**
	 * Method to calculate the current offset from GMT for this time,
	 * including daylight saving
	 */
	private int getLocalOffset()
	{
		// Calculate the offset, given we know the base offset
		// for the current timezone
		int offset = __local_offset_from_GMT;
		
		// Add the daylight saving compensation if we are within daylight saving
		// time
		offset += ( isInDaylightSavingTime() ? __local_dst_offset : 0 );
		
		return offset;
	}


	/**
	 * Method to indicate if this UTCTime is between the times designated
	 * as DST start and end dates
	 */
	private boolean isInDaylightSavingTime()
	{
		// Build a timezone, and use it to check if the date this UTCTime
		// object represents is is daylight saving time
		SimpleTimeZone time_zone = new SimpleTimeZone(0, "GMT");
		
		time_zone.setStartRule( __local_dst_start[0] - 1, __local_dst_start[1], __local_dst_start[2] + 1,
														__local_dst_start[3] * 3600000 );
		time_zone.setEndRule( __local_dst_end[0] - 1, __local_dst_end[1], __local_dst_end[2] + 1,
													__local_dst_end[3] * 3600000 );
		
		GregorianCalendar today = new GregorianCalendar(_date_data[YEAR],_date_data[MONTH] - 1,_date_data[DATE],_date_data[HOUR],_date_data[MINUTE]);
		today.setTimeZone(time_zone);
		
		return (today.get(Calendar.DST_OFFSET) != 0); 
	}


	/** 
	 * Method used to get the package name when supplied with String in
	 * the format package.classname e.g. fipaos.platform.DFRequest will
	 * return fipaos.platform.  To be used when filtering to DIAGNOSTIC 
	 * category files
	 */
	private String getpackage(String full_class_name)
	{
		int dot = full_class_name.lastIndexOf(".");
		String package_name = full_class_name.substring(0, dot);
		return package_name;
	}

  	

	public static void main(String args[])
	{
		UTCTime test_time,test_time2;
/*		
		test_time = new UTCTime(0,0,0,0,0,0,0,null);
		fipaos.util.DIAGNOSTICS.println(test_time.toString());

		test_time.set(test_time.YEAR, 2000);
		fipaos.util.DIAGNOSTICS.println(test_time.toString());

		test_time.set(test_time.MONTH, 2);
		fipaos.util.DIAGNOSTICS.println(test_time.toString());

		test_time.set(test_time.DATE, 12);
		
		fipaos.util.DIAGNOSTICS.println(test_time.toString());
*/

/*		if(args.length>=1)
			test_time = new UTCTime(args[0]);
		else
			test_time = new UTCTime("19990427T1200Z");

		fipaos.util.DIAGNOSTICS.println( "test_time = " + test_time );
		
		if(args.length>=2)
			test_time.changeTimeZone(args[1]);
		else
			test_time.changeTimeZone("");
*/
		
//		UTCDuration duration = new UTCDuration( 0, 0, 10, 1, 0, -1, 0 );
		UTCDuration duration = new UTCDuration( "+00000000T002000000" );

		if(args.length>=1)
			test_time = new UTCTime(args[0]);
		else
			test_time = new UTCTime();
		
		if(args.length>=2)
			test_time2 = new UTCTime(args[1]);
		else
			test_time2 = new UTCTime("19990520T1500Z");

		
		fipaos.util.DIAGNOSTICS.println( "duration   = " + duration );
		fipaos.util.DIAGNOSTICS.println( "test_time  = " + test_time );
		fipaos.util.DIAGNOSTICS.println( "test_time2 = " + test_time2 );
		test_time.addDuration( duration );
		fipaos.util.DIAGNOSTICS.println( "test_time.addDuration( duration ) = " + test_time);
		fipaos.util.DIAGNOSTICS.println( "test_time.getTimeInMillis() = " + test_time.getTimeInMillis() );
		fipaos.util.DIAGNOSTICS.println( "System.currentTimeMillis()  = " + System.currentTimeMillis() );
		
		fipaos.util.DIAGNOSTICS.println( "test_time.isAfter( test_time2 )  = " + test_time.isAfter( test_time2 ) );
		fipaos.util.DIAGNOSTICS.println( "test_time.isBefore( test_time2 ) = " + test_time.isBefore( test_time2 ) );
		fipaos.util.DIAGNOSTICS.println( "test_time.equals( test_time2 )   = " + test_time.equals( test_time2 ) );		
	}
}