/*
 * Copyright (C) The Apache Software Foundation. All rights reserved.
 *
 * This software is published under the terms of the Apache Software License
 * version 1.1, a copy of which has been included with this distribution in
 * the LICENSE file.
 */
package ranab.server.ftp;

import java.io.File;
import java.net.Socket;
import java.net.InetAddress;
import java.net.ServerSocket;
import ranab.server.BaseServer;
import ranab.server.ServerEngine;

/**
 * This is ftp server starting point. For each request it
 * instantiates a new <code>BaseFtpConnection</code> object.
 * 
 * @author <a href="mailto:rana_b@yahoo.com">Rana Bhattacharyya</a>
 */
public class FtpServer extends BaseServer {                             
    
    // server name
    public final static String NAME = "FTP";
        
    // configuration object
    private FtpConfig mConfig;
    
    // pasv server socket
    private ServerSocket mPasvSocket;
    
    private String configFile;
    //////////////////////////////////////////////
    /**
     * Constructor.
     * @param configFile this ftp server configuration file. 
     */
    public FtpServer(File configFile) throws Exception 
    {
        mConfig = new FtpConfig(configFile);
    }   
       
    
    public FtpServer(String configfile) throws Exception
    {
    	super();
    	this.configFile = configfile;
    	File file = new File(configfile);
        mConfig =  new FtpConfig(file);
    }
    
    
    public FtpServer() throws Exception
    {
    	super();
    	File file = new File(this.configFile);
        mConfig =  new FtpConfig(file);
    }
    
    /**
     * Get server port.
     */
    public int getServerPort() {
        return mConfig.getServerPort();
    }
     
    /**
     * Get server bind address.
     */
    public InetAddress getServerAddress() {
        return mConfig.getServerAddress();
    }
     
    /**
     * Get server name.
     */
    public String getServerName() {
        return NAME;
    }
     
    /**
     * Serves threaded request. 
     */
    public void serveRequest(Socket soc) {
        BaseFtpConnection ftpCon = new FtpConnection(mConfig, soc);
        new Thread(ftpCon).start();
    }
    
    /**
     * Stop all currently executing threads.
     */
    public void dispose() {
        if (mConfig != null) {
            mConfig.dispose();
            mConfig = null;
        }
    } 
    
    /**
     * Get the configuration object.
     */
    public FtpConfig getConfig() {
        return mConfig;
    }
        
    
    /////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////
    /**
     * Command line ftp server starting point.
     */
    public void run() 
    {        
        try {
        	
            BaseServer server = new FtpServer(this.configFile);
            final ServerEngine engine = new ServerEngine(server);            
            engine.startServer();
        } 
        catch (Exception ex) {
            ex.printStackTrace();
        }    
    }
}